MODULE sannel
USE rrestr, only: restraint_type
USE reflection_type_util, only: gref_type
USE progtype, only: opti_info_type, param_type
USE timer, only: timer_type
USE rigid_body, only: rigid_body_type
USE cfunction, only: cfunction_type
USE fileutil, only: file_handle

implicit none

integer, parameter :: THERMALF_TYPE = 1
integer, parameter :: PO_TYPE = 2
integer, parameter :: OTHERS_TYPE   = 3
integer, parameter :: EXT_TYPE = 4
integer, parameter :: INT_TYPE = 5

enum, bind(c) 
  enumerator :: SA_LOCAL=1,BBBC,SA_GLOBAL,HARMONYS
endenum

integer, parameter :: RWPC_COSTF=1,  &  ! CF compacted pattern
                      CHI2I=2,       &  ! CF intensita'
                      CHI2FACT=3,    &  ! CF fattori di struttura (tutti i riflessi)
                      CHI2I_NEW=4,   &  ! CF intensita' integrate
                      RWP_COSTF=5,   &  ! CF profilo 
                      CHI2GRU=6,     &  ! CF intensities groups
                      CHI2_COSTF=7,  &  ! CHI2 pattern
                      CHI2I_SD=8        ! CF fattori di struttura (sing e doppietti)
character(len=*), dimension(8), parameter :: COSTF_NAME = ['R weighted profile','R intensities     ','R structure factor',   &
                                                           'Chisquared 1      ','Rwp 1             ','R groups          ',   &
                                                           'Chisquared 3      ','Chisquared 4      ']
character(len=*), dimension(8), parameter :: COSTF_NAME_SHORT = ['Rwp  ','Rint ','RF   ',   &
                                                                 'Chi^2','Rwp  ','Rint ',   &
                                                                 'Chi^2','Chi^2']
integer, dimension(2), parameter :: XPD_COSTF = [RWPC_COSTF,CHI2I] ! CF for powder, XDP_COSTF(1) is the default
integer, dimension(2), parameter :: XSE_COSTF = [CHI2FACT,CHI2I]   ! CF for single crystal and electrons, XSE_COSTF(1) is the default

type(rigid_body_type), dimension(:), allocatable :: rigidb_copy
type(opti_info_type)                             :: ginfo              ! informazioni generali sulla procedura
type(opti_info_type)                             :: goptim             ! informazioni generali sulla procedura
type(param_type), dimension(:), allocatable      :: gpar               ! i parametri dell'algoritmo di SA
real                                             :: shiftatomval       ! shift applicato agli atomi
integer                                          :: numcyc_calc        ! calculated number of CF evaluation for run
real                                             :: rwden              ! denominatore funzione costo
real                                             :: sumycossb
type(cfunction_type), dimension(:), allocatable  :: cfuncsa
#if defined(DISTRIB)
logical, private                                 :: print_tb = .false. ! stampa info per tabella riepilogativa
#else
logical, private                                 :: print_tb = .true.  ! stampa info per tabella riepilogativa
#endif
integer                                          :: ncuold 
integer :: nco_new
real, parameter, private :: DISTDOC = 1.0   ! distanza limite per la d.o.c.
real, parameter, private :: DISTOCC = 0.9   ! distanza limite per atomi in posizione speciale
character(len=*), dimension(4), parameter :: algoname_long = (/'Simulated Annealing                             ',  &
                                                               !'Hybrid Big Bang - Big Crunch combined with GRASP',  &
                                                               'Big Bang - Big Crunch                           ',  &
                                                               'Simulated Annealing                             ',  &
                                                               'Harmony Search                                  '/)  
character(len=*), dimension(4), parameter :: algoname_short= (/'SA      ','GHBB-BC ','SA      ','HS      '/)
logical :: map_enabled = .true.
!
! restraints variables
type(restraint_type), dimension(:), allocatable  :: bumptab

real, parameter :: SASIGD = 0.2
real, parameter :: SAWEID = 100
real, parameter :: SHIFT_DEF = 0.5

integer, parameter :: SATIME_CF=1, SATIME_MOVES=2, SATIME_GRAPH=3, SATIME_RUN=4, SATIME_TOT=5
type(timer_type), dimension(5) :: timesa
character(len=*), private, parameter :: STRTAG='@@'
integer, parameter, private :: DEF_REFINE_TYPE = 1
integer :: refine_type = DEF_REFINE_TYPE ! type of refinement 1=LEBAIL_DECOMP, 2=PAWLEY_DECOMP
integer, parameter :: NINFOMODEL = 6
integer, parameter :: SINGLE_OUTPUT=1, MULTI_OUTPUT=2
integer, parameter :: output_type = SINGLE_OUTPUT
!integer, parameter :: output_type = MULTI_OUTPUT
integer :: go_unit
type(file_handle) :: go_file
CONTAINS

   subroutine sa_prelim(gopti,scond,bcond,read_dir,init_cond,saerr)
   USE variables, only:cryst, dataset
   !USE counts, only: dvalue
   USE General, only:lo,ifAutomatic,OUT_FILE,fname,neutro,is_anomalous
   USE simulated_anneal, only: sa_set_default, SimAnn_Conditions_type, print_unit
   USE RIETV
   USE view
   USE Conteggi, only:npunti
   USE ilmviewmod
   USE messagemod
   USE Molcom, only:jscreen
   USE bb_bc
!   USE plotstyle
   USE anti_bump
   USE errormod
   USE constraints
   USE commandsmod
   USE mpi_prog
   USE fragmentmod
   USE crystal_phase
   USE cryutil
   USE RefinecomRef, only:rfcond
!   USE rigid_body
   USE datasetmod
   USE strutil, only: i_to_s
   USE fileutil, only: file_basename
   type(opti_info_type), intent(inout)       :: gopti    
   type(SimAnn_Conditions_type), intent(out) :: scond                     
   type(bb_bc_condition), intent(out)        :: bcond                     
   logical, intent(in)                       :: read_dir
   logical, intent(in), optional             :: init_cond
   type(error_type), intent(out)             :: saerr
   logical                                   :: initc
   integer                                   :: j,nph
   real                                      :: resmaxg
   real                                      :: resmin
   real, parameter                           :: RESLIMIT = 2.0
   type(error_type)                          :: errdir
   integer, dimension(1)                     :: pfrag
   integer                                   :: numat
   real, dimension(:), allocatable           :: resv
   logical                                   :: update
!
   if (output_type == SINGLE_OUTPUT) then
       go_unit = lo
   else
       call go_file%fopen(file_basename(fname(OUT_FILE))//"_rank"//i_to_s(mpi_prog_rank)//".out","w")
       go_unit = go_file%handle()
   endif
   !if (mpi_prog_rank == 0) then
   !   write(go_unit,'(a,i0)')'GO_UNIT:',go_unit
   print_unit = go_unit
   !endif
!
   call saerr%reset()
!   
   initc = .true.
   if (present(init_cond)) initc = init_cond
!
   call init_rietveld(2,initc,.false.,rfcond)
   if (jscreen > 0 .and. dataset(1)%datatype /= SCRYS_DATA) then
!       call jav_set_modify(3)
       call set_plot_style(8)
       style(STYLE_CALC)%vis = 0
       style(STYLE_DIFF)%vis = 0
       call vedinew(8,1,npunti,lsetstyle=.false.,rescale=0)
   endif      
!
!  Inizializza
   if (initc) then ! if false the conditions are initialize externally
       call sa_set_default(scond)
       call bb_bc_initialize(bcond)
       gopti%cftype = set_default_cfunction(1)  ! set only cost function
   endif
!
!!!!!FIX THIS FOR MULTI
   numat = cryst(1)%natoms()
   call create_bump_table(cryst(1)%at,bumptab)
   shiftatomval = SHIFT_DEF
!corr   call clear_rigidb(rigidb)
!
   do nph=1,numphase(cryst)
      call cryst(nph)%init_conn()
!     preserve directives using keep
      call cryst(nph)%make_fragments(keep=.true.)  
      call cryst(nph)%make_rotations(keep=.true.)
      numat = cryst(nph)%natoms()
      forall (j=1:numat) cryst(nph)%at(j)%rcod = 0    ! reset refinement codes
      !call cryst(nph)%load_bvparam()  !for 'BVPAR' directive and BV restraints
   enddo
!
!  Leggi le direttive se e' attivo il comando %sanneal
   if (read_dir) then
!
       call sa_directives(gopti,scond,bcond,resmaxg,errdir,update)
       if (jscreen > 0 .and. update) then
           call ilmview_draw(cryst(active_phase)%at,cryst(active_phase)%bond,   &
                cryst(active_phase)%elem,cryst(active_phase)%cell,cryst(active_phase)%spg)
       endif
!
       if (errdir%signal) then
           if (jscreen == 0 .or. ifAutomatic == 1) then
               saerr = errdir
               return   ! stop procedure only if no GUI and automatic run
           else
               call errdir%print()
           endif
       endif
   else
       resmaxg = -1      ! 
   endif
!
   !call make_test(gopti,scond,1)
!
!  Align bump restraints from directive with graphic
   if (jscreen == 1 .and. ifAutomatic == 0) then
       if (cryst(1)%numresab() > 0) then
           call table_from_bump(bumptab,cryst(1)%abres,cryst(1)%at,cryst(1)%abset)
       endif
   endif
   do nph=1,numphase(cryst)
!
!     Porta i frammenti in cella
      call fragment_in_cell(cryst(nph)%frag,cryst(nph)%at)
!
!     Fix traslation in case of polar axis only for fragment with more atoms
!     uncomment only for test if (.not.spacegr%polar()) stop
      if (cryst(nph)%spg%polar()) then
          pfrag = maxloc(cryst(nph)%frag(:)%nat)
          call cryst(nph)%spg%code_for_polar(cryst(nph)%frag(pfrag(1))%rcod)
      endif
   enddo
!
!  Compute number of parameters for setting of number of moves and resolution
   call gopt_count_parameters(gopti)
!
!  Calcola valore iniziali per algoritmi di ottimizzazione
   if (gopti%autoalgo) call gopt_set_algo(gopti)   !!!!!!,cryst(1)%numrot(),cryst(1)%numfrag())
   if (scond%autont) call gopt_set_ncycle(gopti,scond%nt)  
   if (gopti%algo == BBBC) then
       call set_bb_bc_condition(gopti,bcond)
   endif
!
!  Memorizza tutti i riflessi
   do nph=1,numphase(cryst)
      !call copy_ref(cryst(nph)%refc,cryst(nph)%ref)
      call cryst(nph)%save_ref()
   enddo
   ncuold = 0
!
!  Definisci i limiti risoluzione dello spettro 
   if (resmaxg < 0) then
       !resmin = dvalue(dataset(1)%tmax,dataset(1)%wave(1))
       resmin = dataset(1)%resolution()
       if (resmin > RESLIMIT) then
           gopti%res = resmin
       else
           !gopti%res = RESLIMIT
           !call gopt_count_parameters(gopti)
           !call check_resp(cryst(1),gopti%res)
           allocate(resv(numphase(cryst)),source=RESLIMIT)
           do nph=1,numphase(cryst)
              call check_resp(cryst(nph),resv(nph))
              !write(0,*)'PH:',nph,resv(nph),RESLIMIT
           enddo
           gopti%res = minval(resv)
       endif
   else    
       gopti%res = resmaxg
   endif
!
   ginfo = gopti  ! save in ginfo for graphic and external function
!   
   if (jscreen > 0) then
       call SetRowCountMesg(4+count(cryst%is_known()),1)
       call write_message(trim(algoname_long(gopti%algo)), pos=1, kproc=0)
       call write_message('Number of refined parameters:',inum=gopti%npar,pos=2, kproc=0)
       call write_message('DOF(ext):',inum=gopti%dofe,pos=3, kproc=0)
       call write_message('DOF(int):',inum=gopti%dofi,pos=4, kproc=1)
       do nph=1,numphase(cryst)
          if (cryst(nph)%is_known()) call cryst(nph)%sfactor(neutro,is_anomalous)
       enddo
       call to_coord_cry(cryst,go_unit,-1,5)
   endif
!   
!  Print info
   if (is_go_print()) then
       write(go_unit,'(1x,79("+"),//,30x,"Global Optimization Procedure",//,1x,79("+"))')
       do nph=1,numphase(cryst)
          write(go_unit,'(/2x,a,i0,a)')'Phase #',nph,': '//trim(cryst(nph)%cr_name)
          call cryst(nph)%print_bonds(go_unit,'C')
          call cryst(nph)%print_molecules(go_unit)
          call cryst(nph)%print_bonds(go_unit,'R')
          if (cryst(nph)%numresbv() > 0) then
              call cryst(nph)%print_bvparam(go_unit)
          endif
       enddo
   endif
!   
   end subroutine sa_prelim     

!----------------------------------------------------------------------------------------------------      

   logical function is_go_print() 
   use mpi_prog
   if (output_type == SINGLE_OUTPUT) then
       if (mpi_prog_rank /= mpi_prog_master) then
           is_go_print = .false.
           return
       endif
   endif

   is_go_print = .true.
   end function is_go_print

!----------------------------------------------------------------------------------------------------      

   subroutine global_init(gopti)
   type(opti_info_type), intent(out)  :: gopti     
!
   gopti%cftype = set_default_cfunction(1)
   !!gopti%cftype = CHI2GRU
   !!gopti%cftype = RWP_COSTF
   !!gopti%cftype = CHI2I_NEW
!corr   gopti%doc = .false.
   gopti%seed = 1      ! If different from zero, holds the seed for random number generator
   gopti%nrun = 10     ! Number of runs
!only for test
!   gopti%nrun = 20     ! Number of runs
!only for test
   gopti%nrunc = 1     ! Contatore dei run
   !gopti%algo = SA_LOCAL  ! Tipo di algoritmo usato
   gopti%algo = BBBC   ! Tipo di algoritmo usato
   gopti%autoalgo = .true.
   gopti%popt = 'PMVESB'   ! Opzioni di output
!
   end subroutine global_init

!----------------------------------------------------------------------------------------------------      

   integer function set_default_cfunction(codedata)  result(cftype)
!
!  Set the default CF for the dataset codedata
!
   USE variables, only: dataset
   USE datasetmod
   integer, intent(in) :: codedata
!
   cftype = XPD_COSTF(1)
   if (ndataset(dataset) > 0) then
       if (dataset(codedata)%datatype == SCRYS_DATA) then
           cftype = XSE_COSTF(1)
       endif
   endif
!
   end function set_default_cfunction

!----------------------------------------------------------------------------------------------------      

   subroutine set_allowed_cfunction(codedata,cfvet)
!
!  Set allowed CF for the dataset codedata
!
   USE variables, only: dataset
   USE datasetmod
   integer, intent(in)                             :: codedata
   integer, dimension(:), allocatable, intent(out) :: cfvet
!
   if (ndataset(dataset) > 0) then
       if (dataset(codedata)%datatype == SCRYS_DATA) then
           allocate(cfvet(size(XSE_COSTF)),source=XSE_COSTF)
           return
       endif
   endif
   allocate(cfvet(size(XPD_COSTF)),source=XPD_COSTF)
!
   end subroutine set_allowed_cfunction

!----------------------------------------------------------------------------------------------------      
 
   subroutine sa_run(gopti,scond,bcond,extrac)
   USE mpi_prog
   USE RefinecomRef, only:rinfo,rfcond
   USE genlsq, only: ncytot, allocafuncls
   USE variables, only:cryst,dataset
   USE Molcom, only:ifBrekke,jscreen
   USE simulated_anneal, only: sa_set_temperature_new, SimAnn_Conditions_type
   USE compare_struct, only: to_coord_sir
   USE atom_type_util, only: print_atoms, volume_per_atom, baricentro, coord_in_newcell
   USE Conteggi, only:npunti
   USE messagemod, only: progress_bar_init, progress_bar_close
   USE bb_bc, only: bb_bc_condition
   USE connect_mod, only: copy_bonds, bond_type
   USE RIETV, only: printfin, init_fobs, count0, setRInt, pawley
   USE View, only: vedinew
   USE ilmviewmod, only: active_phase, ilmview_draw, ilmview_set_cell, ilmview_open, &
                         jav_set_grow_fragment, jav_set_polyhedra, &
                         jav_set_contacts, jav_set_symmetry
   USE enable_amb, only: abilita_tasti, RunSkipAction
   USE fragmentmod, only: set_atm_in_frag, connect_fragment
   USE datasetmod, only: POW_DATA
   USE timer, only: print_time
   USE crystal_phase, only: numphase
   use unit_cell, only: cell_type
   use plotstyle, only: style,STYLE_BACKP
   USE counts, only: dvalue, thvalue
   USE mapinfo, only: FOUR_FC, FOUR_FO, radmap_from_atoms
   USE fourier_mod, only: get_f000
   USE fouriermap   !, only: is_map_enabled, jav_show_map, map_set_rad
   use calculate_pdp, only: LEBAIL_DECOMP, RIET_REFINE, PAWLEY_DECOMP
   use progtype, only: ALLOCA
   use General, only: neutro, is_anomalous
   use plot2
   use enable_amb, only: STOP_EVENT
      !use rigid_body
   type array_temp_sav
        type(cell_type)                    :: cell 
        integer                            :: gcode
        real                               :: gpar
        integer, dimension(:), allocatable :: rcodb,rcodo       
   end type array_temp_sav
   type(opti_info_type), intent(inout)         :: gopti    
   type(SimAnn_Conditions_type), intent(inout) :: scond                     
   type(bb_bc_condition), intent(in)           :: bcond                     
   logical, intent(in)                         :: extrac
   integer :: exit_code
   real                                        :: resmaxg
   integer                                     :: i,nph
   real                                        :: res,resi,minres
   integer                                     :: ncc
   real                                        :: vol_per_atom
   type(bond_type), dimension(:), allocatable  :: legm
!   logical :: extrac
   integer :: nrefu
   type(array_temp_sav), dimension(:), allocatable :: cry_sav
   real                                            :: f000
!
!   extrac = .true.
!   if (present(extra)) extrac = extra
!!!!!!!!!!test
          !call set_bump_exclusions(cryst(1)%at,cryst(1)%bond,cryst(1)%frag,cryst(1)%rot,cryst(1)%abres)
!!!!!!!!!!test
!
!  Conta gradi di liberta
   call gopt_count_parameters(gopti)
!   
!  Assegna valore per la risoluzione
   resmaxg = gopti%res
   resi = resmaxg
!   
!  Controlla che la risoluzione assegnata non sia inferiore al valore minimo gestibile
   ncc = dataset(1)%npoints()
   if (npunti > 0) then
       minres = dvalue(dataset(1)%x0(ncc),dataset(1)%wave(1))
   else
!corr       minres = dvalue(cryst(1)%refc(cryst(1)%numref())%tthd(1),dataset(1)%wave(1))
       minres = cryst(1)%refc(cryst(1)%numref())%dval()
   endif
   res = max(resi,minres)
!   
   gopti%res = res
   gopti%ttmax = thvalue(res,dataset(1)%wave(1))
   if (npunti > 0) then
       gopti%ttmin = dataset(1)%x0(1)  
       call dataset(1)%set_limit(-1.0,gopti%ttmax)
       gopti%ttmax = dataset(1)%xmaxc()
   else
       gopti%ttmin = cryst(1)%refc(1)%tthd(1)
   endif
   do nph=1,numphase(cryst)
      !nrefu = count(cryst(nph)%refc%tthd(1) <= gopti%ttmax .and. cryst(nph)%refc%tthd(1) >= dataset(1)%x(1))
      !nrefu = count(cryst(nph)%refc%tthd(1) <= gopti%ttmax) 
      nrefu = count(cryst(nph)%refc%dval() >= gopti%res)
      call cryst(nph)%set_used_ref(nrefu)
   enddo
!corr   !call set_limit_ref(cryst(1)%ref,rfcond%lim,cryst(1)%refc,gopti%ttmax)
!corr   !call set_limit_ref(cryst(1)%ref,cryst(1)%refc,gopti%ttmax)
!corr       call dataset(1)%set_limit(-1.0,gopti%ttmax)
       !write(0,*)'NREF_SA=',nrefu,cryst(1)%numref(),dataset(1)%xmaxc(),cryst(1)%ref(cryst(1)%numref())%tthd(1),cryst(1)%ref(nrefu)%tthd(1)
       !write(0,*)'NCU=',dataset(1)%nc2
       !write(0,*)'TMAX=',gopti%ttmax,dataset(1)%xmaxc(),dataset(1)%x(dataset(1)%nc2-1)
!      
!!!!prova
!      write(0,*)'nc prima=',rfcond%lim%ncu
!      ncusav = rfcond%lim%ncu
!      rfcond%lim%ncu = min(rfcond%lim%ncu*1.1,1.0*rfcond%lim%nctot)
!      write(0,*)'nc dopo =',rfcond%lim%ncu
!!!!prova
!corr   gopti%nref = rfcond%lim%nrefu
   !gopti%nco = rfcond%lim%ncu
   gopti%nco = dataset(1)%nc2
!   
!corr   if(ncuold /= rfcond%lim%ncu .and. npunti > 0 .and. extrac .eqv. .true.) then
   if(ncuold /= dataset(1)%nc2 .and. npunti > 0 .and. extrac .eqv. .true.) then
       !ncuold = rfcond%lim%ncu
       ncuold = dataset(1)%nc2
       ncytot = 0            
       do nph=1,numphase(cryst)
          cryst(nph)%ref(:)%pk = 20
       enddo
!
       !call dataset(1)%set_limit(n2=rfcond%lim%ncu)
       call dataset(1)%make_background()
!
!      Disegna pattern
       if (jscreen /= 0) then
           call vedinew(8,rescale=0)
           style(STYLE_BACKP)%vis = 0
       endif      
!       
!      Affina il profilo
       !rfcond%raction = 1
       rfcond%raction = refine_type
!corr       rfcond%winmess = .false.
       rfcond%sprint = 'W'
       !rfcond%sprint = 'WSV'
            !rfcond%typefun = PVOIG
            !rfcond%typefun = TCHZ
                   !call init_function_param(pear,pvoi,tch)
       do nph=1,numphase(cryst)
          call setRInt(cryst(nph),rfcond)
       enddo
       call allocafuncls(ALLOCA,dataset(1)%nc2) ! rialloca prima della count0
       call setwei()
       call Count0(rfcond)
       if (index(rfcond%sprint,'S') > 0 .and. is_go_print()) call printfin(go_unit,'Starting Rietveld parameters',rfcond)
!
!      Save parameters modified by refinement
       allocate(cry_sav(numphase(cryst)))
       do nph=1,numphase(cryst)
          cry_sav(nph)%cell = cryst(nph)%cell
          cry_sav(nph)%gcode = cryst(nph)%gcode
          cry_sav(nph)%gpar = cryst(nph)%gpar
          allocate(cry_sav(nph)%rcodb(cryst(nph)%natoms()),source=cryst(nph)%at%rcod(4))  ! salva il codice dei b perche' la riet_new li ricopre
          allocate(cry_sav(nph)%rcodo(cryst(nph)%natoms()),source=cryst(nph)%at%rcod(5))  ! e per occ
       enddo
!
!qt       call clear_plot(grplot,1)
       if (jscreen > 0) call init_plot2(c_char_'cycles'//c_null_char,c_char_'Rwp'//c_null_char)
       if (rfcond%raction == LEBAIL_DECOMP) then
           rfcond%ncauto = 1
           call init_fobs()
           do i=1,3
              rfcond%auto_prof = .true.
              rfcond%auto_stru = .false.
              call riet_new(i,exit_code)
              if (exit_code == STOP_EVENT) return
           enddo
       else
           call pawley(1,rfcond)
           call init_fobs()
           do i=1,3
              rfcond%auto_prof = .true.
              rfcond%auto_stru = .false.
              call riet_new(i,exit_code)
              if (exit_code == STOP_EVENT) return
           enddo
           call pawley(0,rfcond)
       endif
!
       gopti%rwp_leb = rinfo%rwp  ! salva Rwp alla fine dell'affinamento del profilo
!
       do nph=1,numphase(cryst)
          cryst(nph)%gcode = cry_sav(nph)%gcode ! ripristina il codice
          cryst(nph)%gpar = cry_sav(nph)%gpar   ! tieni il valore di G affinato
          cryst(nph)%at%rcod(4) = cry_sav(nph)%rcodb(:) 
          cryst(nph)%at%rcod(5) = cry_sav(nph)%rcodo(:) 
          call coord_in_newcell(cryst(nph)%at,cry_sav(nph)%cell,cryst(nph)%cell)
       enddo
!
!      Riaggiorna le coordinate dei frammenti           
       do nph=1,numphase(cryst)
          call set_atm_in_frag(cryst(nph)%frag,cryst(nph)%at)
       enddo
!
       if (is_go_print()) then
           call printfin(go_unit,'Parameters after profile refinement',rfcond,opt='')
           !write(6,'(a10,3f10.3)')STRTAG//cryst(1)%cr_name,rinfo%rwp,rinfo%rp,Rvalue(cryst(1)%ref%fv,cryst(1)%ref%fo); return
       endif
       !if (rinfo%rpw1 > 50) return  !TEST
!
       rfcond%raction = RIET_REFINE 
   endif
!
!qt   fortran_work = .true.
   do nph=1,numphase(cryst)
!
!     Refined atoms are in general position
      do i=1,cryst(nph)%numfrag()
         if (any(cryst(nph)%frag(i)%rcod(:) > 0)) then
             cryst(nph)%at(cryst(nph)%frag(i)%pos)%ocry = 1.0
         endif
      enddo
!
!corr      call set_equivalent_for_restraints(cryst(nph)%res,cryst(nph)%bond)
   enddo
!
!  Attiva la gestione dei tempi   
   call timesa(SATIME_CF)%set_type('Cost function calculation')
   call timesa(SATIME_MOVES)%set_type('Moves')
   call timesa(SATIME_GRAPH)%set_type('Graphic')
   call timesa(SATIME_RUN)%set_type('Time/run')
   call timesa(SATIME_TOT)%set_type('Total time')          
!
   call gopt_init_costf(gopti)   ! inizializzazione per funzioni costo
!
!  Inizializza gpar, randomizza se richiesto
   call sa_prerefine(gopti,scond%rand)
!
   do nph=1,numphase(cryst)
      if (cryst(nph)%doc_active() .or. cryst(nph)%numresbv() > 0) then
!
!        Set size of grid according to the active procedures
         if (cryst(nph)%doc_active() .and. cryst(nph)%numresbv() > 0) then  
             !kgr = ceiling(maxval(cryst(nph)%bvpar(:,:)%rmax))
             !call init_grid(cryst(nph)%cell,cryst(nph)%spg,cryst(nph)%natoms(),1.0,kgr)  ! grid for doc + bv
             call cryst(nph)%allocate_grid(cutoff=1.0)      ! grid only for doc + bv
         else
             if (cryst(nph)%doc_active()) then
                 !call init_grid(cryst(nph)%cell,cryst(nph)%spg,cryst(nph)%natoms(),1.0,1)  ! grid only for doc
                 call cryst(nph)%allocate_grid(cutoff=1.0)      ! grid only for doc
             elseif (cryst(nph)%numresbv() > 0) then
                 !call init_grid(cryst(nph)%cell,cryst(nph)%spg,cryst(nph)%natoms(),maxval(cryst(nph)%bvpar(:,:)%rmax),1)  ! grid only for bv
                 !call cryst(nph)%allocate_grid(cutoff=maxval(cryst(nph)%bvpar(:,:)%rmax))  ! grid only for bv
                 call cryst(nph)%allocate_grid(cutoff=cryst(nph)%bvs_cutoff())  ! grid only for bv
             endif
         endif
      endif
   enddo
!
!  Stampa informazioni generali:restraints,range,num. di riflessi,dof,etc.   
   !if (mpi_prog_rank == mpi_prog_master) then
   if (is_go_print()) then
       do nph=1,numphase(cryst)
          call print_atoms(cryst(nph)%at,title='Starting structure',kpr=go_unit)
          if (cryst(nph)%is_known()) then
              call to_coord_sir(1,cryst(nph)%at,cryst(nph)%ref,size(cryst(nph)%ref),    &
                                cryst(nph)%cell,cryst(nph)%spg,go_unit,cryst(nph)%atpub)
          endif
!
!         Checking sul volume per atomo
          vol_per_atom = volume_per_atom(cryst(nph)%at, cryst(nph)%spg%nsymop, cryst(nph)%cell%volume())
          write(go_unit,'("   Volume per atom calculated by starting model = ",f0.3)') vol_per_atom
          if (vol_per_atom < 5 .or. vol_per_atom > 30)    &
          write(go_unit,'("   WARNING: volume per atom seems unresonable. Check the starting fragment")')
       enddo

       call gopt_print_info(gopti,scond,go_unit)
   endif
          !vol_per_atom = volume_per_atom(cryst(1)%at, cryst(1)%spg%nsymop, cryst(1)%cell%volume())
          !if (vol_per_atom < 13) return  !TEST
!
!  Operazioni per la grafica, imponi visualizzazione cella   
   if (jscreen > 0) then
       call ilmview_set_cell(1)
       call ilmview_open(cryst(active_phase)%cr_name)
       call ilmview_draw(cryst(active_phase)%at,cryst(active_phase)%bond,   &
                         cryst(active_phase)%elem,cryst(active_phase)%cell,cryst(active_phase)%spg)
       !call jav_set_style(5,1)          !!mr 221014
       !call jav_set_sym(0,0,1)
       call jav_set_symmetry(0)
       call jav_set_symmetry(0)
       call jav_set_polyhedra(0)
       call jav_set_contacts(0)
       call jav_set_grow_fragment(0)
   endif
!   
   call timesa(SATIME_TOT)%tic()
!
   !call abilita_tasti(RunSkipAction,1)
!
!  Calcolo automatico del numero di cicli per temperatura
   if (scond%autont) call gopt_set_ncycle(gopti,scond%nt)
!
!  Calcolo automatico della temperatura iniziale
   if (scond%autoti) call sa_set_temperature_new(scond,gopti,gpar,safunc_temp)
   !ginfo = gopti; if (sc%autoti) call sa_set_temperature(sc,gopti,gpar,tfunk,safunc1)
!
!  Setting della convergenza
   call sa_set_convergence(scond,gopti%npar)
!
!qt   call progress_bar_init('Starting simulated annealing')   !  Inizializza progress bar      
   call progress_bar_init()   !  Inizializza progress bar      
!
!  Definisce su che righe scrivere i vari messaggi
   if (.not.extrac) then
       call sa_set_messages(gopti%algo, 0, cryst(1)%numresab(),cryst(1)%numres())
   else
       call sa_set_messages(gopti%algo, npunti, cryst(1)%numresab(),cryst(1)%numres())
   endif
!
!  Initialize map visualization: compute radius and define sogmap
!old   call copy_atoms(atomcart,cryst(1)%at)
!old   if (cryst(1)%numfrag() > 1) then
!old       call regroup_fragment(cryst(1)%frag,atomcart,cryst(1)%cell,cryst(1)%spg)
!old          write(0,*)'REGROUP'
!old   endif
!old   call frac_to_cart(atomcart,cryst(1)%cell%get_ortom())
!old   call get_radius_molecule(atomcart,radmap)
!old     kmin = minloc(cryst(1)%spg%asulim,dim=1)
!old     write(0,*)'ASYM=',cryst(1)%spg%asulim,cryst(1)%spg%asulim(kmin)
!old     write(0,*)'RAD=',radmap, cryst(1)%cell%get_par(kmin)*cryst(1)%spg%asulim(kmin)
   do nph=1,numphase(cryst)
      cryst(nph)%mapinfo%sog = 0.9   !!!120
      cryst(nph)%mapinfo%radius = radmap_from_atoms(cryst(nph)%at,cryst(nph)%cell,cryst(nph)%frag)
      !write(0,*)'radmap=',cryst(1)%mapinfo%radius
   enddo
   call map_set_rad(cryst(active_phase)%mapinfo%radius)
!!!   call sogmap_init(FOUR_FO_FC,ref,atm,sasogmap)   ! non funziona ad es. con clomipra
!!!   sasogmap = sasogmap * 0.003
!!!      write(0,*)'sogmap=',sasogmap
!
   ifBrekke = 0      !azzera codice di interazione con la grafica
   call abilita_tasti(RunSkipAction,1)
   call simann(gopti,scond,bcond,extrac) 
   call abilita_tasti(RunSkipAction,0)
   !call ptempering(scond) 
!
   call gopt_close_costf()            ! dealloca info
   do nph=1,numphase(cryst)
      call set_atm_in_frag(cryst(1)%frag,cryst(1)%at)
   enddo
!
!  Chiudi gestione tempi   
   call timesa(SATIME_TOT)%toc()
   call timesa(SATIME_RUN)%set_print_ave()
   if (is_go_print()) then
       call print_time(timesa,go_unit)
   endif
!
   if (jscreen > 0) then
       call progress_bar_close(1)
!
!      disegna modello
       call copy_bonds(legm,cryst(1)%bond)
       if (cryst(1)%numfrag() > 1) then 
           call connect_fragment(cryst(1)%frag,cryst(1)%at,legm,cryst(1)%cell,.false.)
       endif
       call ilmview_draw(cryst(1)%at,legm,cryst(1)%elem,cryst(1)%cell,cryst(1)%spg)
!
!      disegna profilo
       call cryst(1)%sfactor(neutro,is_anomalous)
       if (dataset(1)%datatype == POW_DATA) then
           call sa_profile(square=.false.)
           call vedinew(8,1,gopti%nco,rescale=0)
       endif
!
       if (is_map_enabled(cryst(active_phase)%mapinfo)) then
           nph = active_phase
           call map_set_centre(baricentro(cryst(nph)%at))
           if (cryst(nph)%mapinfo%ftype == FOUR_FO .or. cryst(nph)%mapinfo%ftype == FOUR_FC) then
               f000 = get_f000(cryst(nph)%at,cryst(nph)%elem,cryst(nph)%spg,dataset(nph)%radtype)
           else
               f000 = 0
           endif
!!!!!FIXME-use dataset(1)%resolution() instead of dvalue
           call jav_show_map(cryst(nph)%mapinfo%ftype,dvalue(dataset(1)%tmax,dataset(1)%wave(1)),cryst(nph)%ref,   &
                             'SA Map',sogmap=cryst(nph)%mapinfo%sog,fctype=1,scalem=.true.,f000=f000,cell=cryst(nph)%cell)
       endif
!
   endif
!   
!qt   fortran_work = .false.
!
   end subroutine sa_run

!----------------------------------------------------------------------------------------------------      

   subroutine sa_set_messages(algo, data_type, nbump, nres)
!
!  Set order of messages for SA in the split window
!
   USE simulated_anneal
   USE molcom, only:jscreen
   USE messagemod
   integer, intent(in) :: algo
   integer, intent(in) :: data_type
   integer, intent(in) :: nbump, nres
   integer             :: nline
!
   if (jscreen == 0) return

   nline = 1
   if (algo == BBBC) then
       nline = nline + 1; vmess(LINE_RUN) = nline
                          vmess(LINE_TEMP) = -100   ! non mostrare la temperatura
       nline = nline + 1; vmess(LINE_MOVES) = nline
       nline = nline + 1; vmess(LINE_CF) = nline
       nline = nline + 1; vmess(LINE_CFMIN) = 5
   else
       nline = nline + 1; vmess(LINE_RUN) = nline
       nline = nline + 1; vmess(LINE_TEMP) = nline
       nline = nline + 1; vmess(LINE_MOVES) = nline
       nline = nline + 1; vmess(LINE_CF) = nline
       nline = nline + 1; vmess(LINE_CFMIN) = nline
   endif
   if (nres > 0) then
       nline = nline + 1; vmess(LINE_CFRES) = nline
   endif
   if (nbump > 0) then
       nline = nline + 1; vmess(LINE_CFBUMP) = nline
   endif
   if (data_type > 0) then
       nline = nline + 1; vmess(LINE_RATIO) = nline
       nline = nline + 1; vmess(LINE_TIME) = nline
   else
       vmess(LINE_RATIO) = -100
       nline = nline + 1; vmess(LINE_TIME) = nline + 1
   endif
   call SetRowCountMesg(nline,1)
!
   end subroutine sa_set_messages

!----------------------------------------------------------------------------------------------------      

   subroutine gopt_init_costf(gopti)
!
!  Inizializzazione della funzione costo
!   
   USE fastfcal
   USE genlsq
   USE variables, only:cryst,dataset
   USE reflection_type_util
   USE arrayutil
   USE variables, only: dataset
   USE cryutil
   USE crystal_phase
   type(opti_info_type), intent(inout) :: gopti    
   integer                             :: i,kal
   integer                             :: nnn
   integer                             :: maxlim
   integer, dimension(:), allocatable  :: ivetc,puntc
   real, dimension(:), allocatable     :: ycountonew,ycossb,weichi
   integer                             :: ncc,nph,maxf
!   
   select case (gopti%cftype)
      case (RWP_COSTF)     
        do nph=1,numphase(cryst)
           cryst(nph)%ref(:)%pk = 1.0
           call profile_init(dataset(1)%typefun(nph),gopti,cryst(nph),nph)
           call new_array(cryst(nph)%yc,gopti%nco)
        enddo
        call dataset(1)%set_ys(gopti%nco)
        call dataset(1)%set_wei(real(weilsq(:gopti%nco)))
        sumycossb = sum(dataset(1)%ys)

        rwden = sqrt(sum(dataset(1)%wei*dataset(1)%y(:gopti%nco)**2))/100.0
        gopti%Cost_function_name = COSTF_NAME(RWP_COSTF)
        gopti%Cost_function_name_short = 'Rwp'
               
      case (CHI2FACT)
        rwden = 0.0
        do nph=1,numphase(cryst)
           rwden = sqrt(sum(cryst(nph)%ref%fo**2))/100.0
        enddo
        gopti%Cost_function_name = COSTF_NAME(CHI2FACT)
        gopti%Cost_function_name_short = 'R-factor'
      
      case (CHI2I)
        if (allocated(cfuncsa)) deallocate(cfuncsa)
        allocate(cfuncsa(numphase(cryst)))
        rwden = 0.0
        do nph=1,numphase(cryst)
           call new_array(cfuncsa(nph)%intoss,cryst(nph)%numref())
           call new_array(cfuncsa(nph)%intcal,cryst(nph)%numref())
           call new_array(cfuncsa(nph)%wei,cryst(nph)%numref())
           !intoss = ref%m*ref%lp*ref%fo**2      
           !intoss = ref%m*ref%fo**2      
           !intoss = ref%lp*ref%fo**2      
           cfuncsa(nph)%intoss = cryst(nph)%ref%fo**2      
           cfuncsa(nph)%sumoss = sum(cfuncsa(nph)%intoss)
           rwden = rwden + sum(cfuncsa(nph)%intoss**2)
           cfuncsa(nph)%wei(:) = 1.0/(1.0+cfuncsa(nph)%intoss(:))
        enddo
        !write(6,*)'LIMITS=',minval(cfuncsa(1)%intoss),maxval(cfuncsa(1)%intoss)
        rwden = sqrt(rwden)/100.0
        gopti%Cost_function_name = COSTF_NAME(CHI2I)
        gopti%Cost_function_name_short = 'R-int'
        
      case (CHI2I_SD)
        if (allocated(cfuncsa)) deallocate(cfuncsa)
        allocate(cfuncsa(numphase(cryst)))
        rwden = 0.0
        do nph=1,numphase(cryst)
           call select_ref(cryst(nph)%ref,3)
           call new_array(cfuncsa(nph)%intoss,cryst(nph)%numref())
           call new_array(cfuncsa(nph)%intcal,cryst(nph)%numref())
           cfuncsa(nph)%intoss = cryst(nph)%ref%m*cryst(nph)%ref%lp(1)*cryst(nph)%ref%fo**2      
           cfuncsa(nph)%sumoss = sum(cfuncsa(nph)%intoss)
           rwden = rwden + sum(cfuncsa(nph)%intoss**2)
        enddo
        rwden = sqrt(rwden)/100.0
        gopti%Cost_function_name = COSTF_NAME(CHI2I_SD)
        gopti%Cost_function_name_short = 'R-Bragg'
        
      case (CHI2I_NEW)
        if (allocated(cfuncsa)) deallocate(cfuncsa)
        allocate(cfuncsa(numphase(cryst)))
        rwden = 0.0
        do nph=1,numphase(cryst)
           cryst(nph)%ref(:)%pk = 1.0
           call new_array(cfuncsa(nph)%intoss,cryst(nph)%numref())
           call new_array(cfuncsa(nph)%intcal,cryst(nph)%numref())
           call profile_init(dataset(1)%typefun(nph),gopti,cryst(nph),nph)
           cfuncsa(nph)%sumoss = 0
           kal = 1
           do i=1,size(cryst(nph)%ref)
              cfuncsa(nph)%intoss(i) = 0
              do nnn=cryst(nph)%profi(i,kal)%iniz,cryst(nph)%profi(i,kal)%ifin
                 cfuncsa(nph)%intoss(i) = cfuncsa(nph)%intoss(i) + dataset(1)%y(nnn) - dataset(1)%yb(nnn)
              enddo 
              cfuncsa(nph)%sumoss = cfuncsa(nph)%sumoss + cfuncsa(nph)%intoss(i)
           enddo
           rwden = rwden + sum(cfuncsa(nph)%intoss**2)
        enddo
        rwden = sqrt(rwden)/100.0
        gopti%Cost_function_name = COSTF_NAME(CHI2I_NEW)
        gopti%Cost_function_name_short = 'R-Bragg'

      case (RWPC_COSTF,CHI2_COSTF)     
        maxlim = 0
        do nph=1,numphase(cryst)
           cryst(nph)%ref(:)%pk = 1.0
           call profile_init(dataset(1)%typefun(nph),gopti,cryst(nph),nph)
           maxf = maxval(cryst(nph)%profi%ifin)
           if (maxf > maxlim) maxlim = maxf
        enddo
        call new_array(ivetc,maxlim)
        call new_array(puntc,maxlim)
        call new_array(ycountonew,maxlim)
        call new_array(weichi,maxlim)
        call new_array(ycossb,maxlim)
!corr        call new_array(ycountsa,maxlim)
        ivetc(:) = 0
        do kal=1,dataset(1)%nwave
           do nph=1,numphase(cryst)
              do i=1,size(cryst(nph)%ref)
                 do nnn=cryst(nph)%profi(i,kal)%iniz,cryst(nph)%profi(i,kal)%ifin
                    ivetc(nnn) = 1
                 enddo
                 !write(6,*)i,'refo=',profi(i)%iniz,profi(i)%ifin
                 !write(6,*)'yc=',dataset(1)%y(profi(i)%iniz:profi(i)%ifin)
              enddo
           enddo
        enddo
        ncc = 0
        do nnn=1,maxlim
           if (ivetc(nnn) == 1) then
               ncc = ncc + 1
               puntc(nnn) = ncc
               ycountonew(ncc) = dataset(1)%y(nnn)
               ycossb(ncc) = dataset(1)%y(nnn) - dataset(1)%yb(nnn)
               weichi(ncc) = real(weilsq(nnn))
           endif
        enddo
        call dataset(1)%set_ys(ycossb(:ncc))
        call dataset(1)%set_wei(weichi(:ncc))
        do kal=1,dataset(1)%nwave
           do nph=1,numphase(cryst)
              do i=1,size(cryst(nph)%ref)
                 cryst(nph)%profi(i,kal)%iniz = puntc(cryst(nph)%profi(i,kal)%iniz)
                 cryst(nph)%profi(i,kal)%ifin = puntc(cryst(nph)%profi(i,kal)%ifin)
              enddo
           enddo
        enddo
        nco_new = ncc
        do nph=1,numphase(cryst)
           call new_array(cryst(nph)%yc,ncc)
        enddo

        rwden = sqrt(sum(dataset(1)%wei*ycountonew(:ncc)**2))/100.0
        gopti%Cost_function_name = COSTF_NAME(gopti%cftype)
        gopti%Cost_function_name_short = 'Rwp'
        sumycossb = sum(dataset(1)%ys)

      case (CHI2GRU)
        rwden = 0.0
        do nph=1,numphase(cryst)
           call make_ref_groups(cryst(nph)%ref,cryst(nph)%gref,sogi=1.0)
           !call print_refg(cryst(nph)%gref)
           rwden = rwden + sum(cryst(nph)%gref(:)%Io**2)
        enddo
        rwden = sqrt(rwden)/100.0
        gopti%Cost_function_name = COSTF_NAME(CHI2GRU)
        gopti%Cost_function_name_short = 'Rg'
        
   end select
   do nph=1,numphase(cryst)
      call fcalc_init(cryst(nph)%ref,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%elem,dataset(1)%radtype,   &
                   cryst(nph)%sfact,cryst(nph)%sfactc,cryst(nph)%kpscat)
   enddo
!  
   end subroutine gopt_init_costf
        
!----------------------------------------------------------------------------------------------------         

   subroutine profile_init(typef,gopti,crystal,nph)
   USE profile_function
   USE variables, only: dataset
   USE cryutil
   USE crystal_phase
   type(opti_info_type), intent(in)     :: gopti    
   type(crystal_phase_t), intent(inout) :: crystal
   integer, intent(in)                  :: nph
   integer, intent(in)                  :: typef
   integer                              :: ier

   select case (typef)
     case (PEARSON7)
       call pearson_init(crystal,dataset(1)%x0(:gopti%nco),dataset(1)%pear(1),dataset(1)%nwave,dataset(1)%ratio,nph,ier)

     case (PVOIG)
       call psvoigt_init(crystal,dataset(1)%x0(:gopti%nco),dataset(1)%pvoi(1),dataset(1)%nwave,dataset(1)%ratio,nph)

     case (TCHZ)

   end select

   end subroutine profile_init

!----------------------------------------------------------------------------------------------------         

   subroutine gopt_close_costf()
   USE fastfcal
   USE variables, only: cryst
   USE cryutil
   USE crystal_phase
   integer :: nph
!
   do nph=1,numphase(cryst)
      call fcalc_close(cryst(nph)%sfact,cryst(nph)%sfactc)     ! dealloca info sui riflessi
!
      call close_profile(cryst(nph))   ! dealloca info sul profilo
   enddo
!   
   end subroutine gopt_close_costf
   
!----------------------------------------------------------------------------------------------------         

   subroutine sa_set_convergence(sc,npar)
!
!  Setting della convergenza in base al numero di parametri
!
   use simulated_anneal, only: SimAnn_Conditions_type
   type(SimAnn_Conditions_type), intent(inout) :: sc
   integer, intent(in)                         :: npar
!
   select case (npar)
      case (1:3)
        sc%eps = 0.01

      case (4:5)
        sc%eps = 0.1

      case (6:)
        sc%eps = 2.0

   end select
!
   end subroutine sa_set_convergence

!----------------------------------------------------------------------------------------------------         

   subroutine sa_directives(gopti,sc,bcond,resmaxg,errdir,update) 
!
!  Read directives for global optimization
!
   USE simulated_anneal                  
   USE commandsmod
   USE bb_bc
   USE errormod
   USE connect_mod
   USE variables, only: cryst,dataset
   USE rrestr
   USE unit_cell
   USE constraints
   USE rotationmod
   USE fragmentmod
   USE gen_frm
   USE strutil
   use rigid_body
   use calculate_pdp, only: PAWLEY_DECOMP
   type(opti_info_type), intent(inout)         :: gopti     
   type(SimAnn_Conditions_type), intent(inout) :: sc                     
   type(bb_bc_condition), intent(inout)        :: bcond                     
   real, intent(out)                           :: resmaxg
   type(error_type), intent(out)               :: errdir
   logical, intent(out)                        :: update
   logical                                     :: found
   integer                                     :: i
   integer                                     :: icpos
   type(error_type)                            :: errf
   logical                                     :: pawley
!
!  Inizializza parametri dell'algoritmo
   resmaxg = -1
   refine_type = DEF_REFINE_TYPE
!
!  Read directives
   icpos = command_position(commands,[SA_CMD,SA2_CMD])
   if (icpos > 0) then
       call reset_constraints(cryst(1)%constr)  ! delete old constraints and load in constr new constraints by read_dir_frag
!
!      Read directives specific for SA
       update = .false.
       do i=1,commands(icpos)%ndir
          call read_dir_sa(commands(icpos)%strdir(i),gopti,sc,bcond,resmaxg,pawley,errdir,found)
          if (errdir%signal) return
          if (pawley) refine_type = PAWLEY_DECOMP
          if (found) cycle
!
!         Read directives specific for molecular fragments
          if (word_is_contained(commands(icpos)%strdir(i),"deletehydro")) cycle  ! jump directive 'deletehydro'
          call read_dir_frag(commands(icpos)%strdir(i),cryst(1),dataset(1)%radtype,errdir,found,update)
          if (errdir%signal) return
          if (.not.found) then 
              call errf%set('Error reading directive "'//trim(commands(icpos)%strdir(i))//'"')
          endif
       enddo
       if (errf%signal) errdir = errf
   endif
!
   call errdir%as_severe()
!
!  Global optimization works on array sconstr 
   if (nconstraints(cryst(1)%constr) > 0) then
       call copy_constraints(cryst(1)%sconstr,cryst(1)%constr) 
       call refine_constraint(cryst(1)%sconstr,cryst(1)%at,1)  ! activate constraints
   endif
!
   end subroutine sa_directives

 !----------------------------------------------------------------------------------------------------

   subroutine read_dir_sa(strdir0,gopti,sc,bcond,resmaxg,pawley,errdir,found)
   USE strutil
   USE simulated_anneal                 
   USE bb_bc
   USE errormod
   USE mpi_prog
   character(len=*), intent(in)                :: strdir0
   type(opti_info_type), intent(inout)         :: gopti     
   type(SimAnn_Conditions_type), intent(inout) :: sc                     
   type(bb_bc_condition), intent(inout)        :: bcond                     
   real, intent(inout)                         :: resmaxg
   logical, intent(out)                        :: pawley
   type(error_type), intent(out)               :: errdir
   logical, intent(out)                        :: found
   character(len=len(strdir0))                 :: strdir,word
   real, dimension(100)                        :: vet
   integer, dimension(100)                     :: ivet
   integer, dimension(:), allocatable          :: iveta
   integer                                     :: iv
   integer                                     :: nlongs
   logical                                     :: errpar
!
   strdir = strdir0
   call Cutst(strdir,nlongs,word) 
!
   errpar = .false.
   found = .true.
   pawley = .false.
   if (word_is_contained(word,'TEMPER')) then ! temperature        
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv >= 1) then
           sc%t_ini = vet(1)
           sc%autoti = .false.
       else
           errpar = .true.
       endif

   elseif (word_is_contained(word,'ALGO')) then       ! tipo di algoritmo
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv > 0) then
           if (ivet(1) > 0 .and. ivet(1) < 5) then
               ! ignore this directive if PSA is activated
               if (sc%cooperate .and. mpi_prog_numprocs > 1 .and. ivet(1) /= SA_LOCAL) then   
                   call errdir%setw('Directive '//trim(strdir0)//' will be ignored')
               else
                   gopti%algo = ivet(1)
                   gopti%autoalgo = .false.
               endif
           endif
       endif

   elseif (word_is_contained(word,'COST_FUNCTION')) then ! cost function
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv > 0) then
           gopti%cftype = ivet(1)
       else
           errpar = .true.
       endif
             
   elseif (word_is_contained(word,'EXPORT')) then ! cost function
       call Getnum1(strdir(:nlongs),ivet=iveta,iv=iv)
       if (iv >= 1) then
           gopti%nexport = iveta(1)
       else
           errpar = .true.
       endif

   elseif (word_is_contained(word,'TFACTOR')) then ! temperature factor
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv >= 1) then
           sc%temp_fact = vet(1)
       else
           errpar = .true.
       endif
               
   elseif (word_is_contained(word,'NITER')) then ! number of iterations
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv >= 1) then
           sc%nt = ivet(1)
           sc%autont = .false.
       else
           errpar = .true.
       endif

   elseif (word_is_contained(word,'NRUN')) then ! number of runs
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv >= 1) then
           gopti%nrun = ivet(1)
!corr           call update_run_cooperate(gopti%nrun,sc%cooperate)
       else
           errpar = .true.
       endif

   elseif (word_is_contained(word,'NCYC')) then ! numero di cicli per bb-bc
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv >= 1) then
           bcond%niter = ivet(1)
           bcond%auto = .false.
       else
           errpar = .true.
       endif

   elseif (word_is_contained(word,'RESMAX')) then ! max resolution
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv > 0) then
           resmaxg = vet(1)
       else
           errpar = .true.
       endif

   elseif (word_is_contained(word,'RAND')) then ! randomize fragment
       call Getnum(strdir(:nlongs),vet,ivet,iv)
       if (iv == 0) then
           sc%rand = 0
       else
           sc%rand = ivet(1)
       endif
   elseif (word_is_contained(word,'COOP')) then ! cooperation in SA
       if (mpi_prog_numprocs == 1) then
           call errdir%setw('Directive '//trim(strdir0)//' will be ignored')
       else
           sc%cooperate = .true.
           gopti%algo = SA_LOCAL
           gopti%autoalgo = .false.
!corr           call update_run_cooperate(gopti%nrun,sc%cooperate)
       endif
   elseif (word_is_contained(word,'RARE')) then ! rare cooperation in SA
       if (mpi_prog_numprocs == 1) then
           call errdir%setw('Directive '//trim(strdir0)//' will be ignored')
       else
           sc%nrare = 10 !mpi_prog_numprocs / 5  !TOFIX
           if (sc%nrare == 0) sc%nrare = 1
       endif
   elseif (word_is_contained(word,'PAWLEY')) then ! type of refinement
       pawley = .true.
   elseif (word_is_contained(word,'SEED')) then   ! random seed
       call Getnum(strdir(:nlongs),vet,iv=iv)
       if (iv > 0) gopti%seed = vet(1)
   else
       found = .false.
   endif
!   
!  Segnale errore nell'inserimento del paramatro da associare alla stringa
   if (errpar) then
       call errdir%set('Error in directive "'//trim(strdir0)//'"')
   endif        
!
   end subroutine read_dir_sa
  
!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine update_run_cooperate(nrun,coop)
!corr!
!corr!  In case of cooperation the number of run is updated to avoid idle cpus
!corr!
!corr   use mpi_prog
!corr   integer, intent(inout) :: nrun
!corr   logical, intent(in)    :: coop
!corr!
!corr!!!!!TOFIX: remove this function
!corr   return
!corr!
!corr   if (mpi_prog_numprocs > 1 .and. coop) then
!corr       if (mod(nrun,mpi_prog_numprocs) /= 0) nrun = mpi_prog_numprocs * (nrun/mpi_prog_numprocs + 1)
!corr   endif
!corr!
!corr   end subroutine update_run_cooperate
!corr
!----------------------------------------------------------------------------------------------------

   subroutine gopt_count_parameters(info)   !!!!,atomo,frag,rotat,gpocode)
   USE constraints
   USE variables, only: cryst
   USE crystal_phase
   use rigid_body
   type(opti_info_type), intent(inout) :: info
   integer                             :: i,nph
!
   info%dofe = 0           
   info%npb = 0
   info%npshift = 0
   info%npar = 0
   info%nprot = 0
!   
   do nph=1,numphase(cryst)
!
!     Count external DOF
      info%dofe = info%dofe + cryst(nph)%edof()
!   
!     Count thermal factors B
      info%npb = info%npb + count(cryst(nph)%at%rcod(4) > 0)
      if (nconstraints(cryst(nph)%sconstr) > 0) info%npb = info%npb + count(cryst(nph)%sconstr%ref > 0)
!
!     Count rotations   
      info%nprot = info%nprot + cryst(nph)%idof()
!
!     Count shifts
      do i=1,cryst(1)%nrigid()
         info%npshift = info%npshift + count(cryst(1)%rigidb(i)%rcod > 0)
      enddo
!
!     Preferred orientation
      if (cryst(nph)%gcode > 0) then
          info%npar = info%npar + 1
      endif
   enddo
!
!  DOF interni   
   info%dofi = info%nprot
!
!  Numero totale di parametri   
   info%npar = info%npar + info%dofi + info%dofe + info%npb + info%npshift
!
   end subroutine gopt_count_parameters

!----------------------------------------------------------------------------------------------------

   subroutine sa_prerefine(gopti,randomod)
!
!  Conta DOF, definisce i parametri (gpar) e randomizza i parametri se richiesto
!   
   USE variables, only: cryst, dataset
   USE simulated_anneal
   USE compare_struct
   USE trig_constants, only: twopi,dtor
   USE atom_type_util
   USE errormod
   USE rand_mod
   USE connect_mod
   USE elements
   USE constraints
   USE gen_frm
   USE rotationmod
   USE fragmentmod
   USE mpi_prog
   USE crystal_phase
   USE rrestr
   USE arrayutil
   USE rigid_body
         use anti_bump
   type(opti_info_type), intent(inout)       :: gopti     
   integer, intent(in)                       :: randomod 
   integer                                   :: i,j,nph
   integer                                   :: nkref
   character(len=30)                         :: strat
   real, dimension(3)                        :: bar0
   character(len=*), dimension(5), parameter :: atomstring = (/'x   ','y   ','z   ','B   ','occ.'/)
   character(len=*), dimension(3), parameter :: string_rot =(/'theta','phi  ','z    '/)
   real, dimension(3), parameter             :: XROT0 = (/0.5,0.5,0.0/) 
   real, dimension(3)                        :: rangef
   logical, dimension(3)                     :: not_periodic, out_range
   real, dimension(3)                        :: newpos
   real                                      :: newang
   integer, dimension(:), allocatable        :: vres,vares
!
!  Randomize models if required
   if (randomod >= 0) then
                !write(0,*)'RAND=',randomod,mpi_prog_rank
                ! FIX problem with randmod=0 int on system clock, gfortran, see intel
       do nph=1,numphase(cryst)
          call Init_Ran(randomod)
          call randomize_molecule(cryst(nph)%at,cryst(nph)%bond,cryst(nph)%cell,cryst(nph)%frag,cryst(nph)%rot)
          call set_atm_in_frag(cryst(nph)%frag,cryst(nph)%at)                  ! update fragment info
          do i=1,cryst(nph)%numrot()                                         ! update torsion value
             cryst(nph)%rot(i)%torsval = torsion_value_for_rotat(cryst(nph)%rot(i),cryst(nph)%at,cryst(nph)%cell)
          enddo
!
!         Crea mol file dalle coordinate random
          if (mpi_prog_rank == mpi_prog_master) then
              call export_structure(trim(cryst(nph)%cr_name)//'_rand.mol',cryst(nph)%at,cryst(nph)%cell,cryst(nph)%spg,  &
                                    cryst(nph)%elem,MOL_FILE,cryst(nph)%cr_name,dataset(1)%wave(1),cryst(nph)%bond)
          endif
       enddo
   endif
!
!  Alloca xpar al numero di parametri       
   if (allocated(gpar)) deallocate(gpar)
   allocate(gpar(gopti%npar))
!
   gpar(:)%stp = 1.0
   nkref = 0      
   do nph=1,numphase(cryst)
!   
!     Parametri termici
      do j=1,size(cryst(nph)%at)
         write(strat,'(i0,a)')j,' ('//trim(cryst(nph)%at(j)%lab)//')'     
         if (cryst(nph)%at(j)%rcod(4) > 0) then      
            nkref = nkref + 1
            gpar(nkref)%str = trim(atomstring(4))//trim(strat)
            gpar(nkref)%val = cryst(nph)%at(j)%biso
            cryst(nph)%at(j)%rcod(4) = nkref
            gpar(nkref)%low = max(0.0,cryst(nph)%at(j)%biso-10)
            gpar(nkref)%high = cryst(nph)%at(j)%biso+10
            gpar(nkref)%bound = 0         
            allocate(gpar(nkref)%pat(1))
            gpar(nkref)%pat(1) = j         
            gpar(nkref)%typepar = THERMALF_TYPE
            gpar(nkref)%ph = nph
         endif
      enddo
!
      do i=1,nconstraints(cryst(nph)%sconstr)
         if (cryst(nph)%sconstr(i)%ref > 0) then
             select case (cryst(nph)%sconstr(i)%code)
               case (CONSTR_B)
                 nkref = nkref + 1
                 !gpar(nkref)%str = 'constraint on '//trim(specie_from_pxen(atm(sconstr(i)%vat(1))%ptab))
                 gpar(nkref)%str = 'constraint on B'
                 cryst(nph)%sconstr(i)%ref = nkref
                 gpar(nkref)%low = 0
                 gpar(nkref)%high = 20
                 gpar(nkref)%bound = 0         
                 gpar(nkref)%val = constr_aveb(cryst(nph)%sconstr(i),cryst(nph)%at,gpar(nkref)%low,gpar(nkref)%high)
             !write(0,*)'MED=',gpar(nkref)%val,atm(sconstr(i)%vat)%biso
                 allocate(gpar(nkref)%pat(cryst(nph)%sconstr(i)%npar))
                 gpar(nkref)%pat(:) = cryst(nph)%sconstr(i)%vat
                 gpar(nkref)%typepar = THERMALF_TYPE
                 gpar(nkref)%ph = nph
                case (4)
             end select
         endif
      enddo
!
!     Loop sui frammenti
      do i=1,cryst(nph)%numfrag()
!
!        Traslazioni del frammento
!        Calcola il baricentro del frammento
         bar0 = baricentro(cryst(nph)%at(cryst(nph)%frag(i)%pos)) 
!
!        calcola il range di movimento in ognuna delle 3 direzioni
         rangef(:) = cryst(nph)%frag(i)%thigh(:) - cryst(nph)%frag(i)%tlow(:)    
!
!        controlla se il range e' diverso da 1 e quindi non pu� essere periodico
         not_periodic(:) = abs(rangef(:) - 1.0) > 0.01
!
!        controlla se il frammento e' fouri range
         out_range(:) = any(bar0(:) > cryst(nph)%frag(i)%thigh) .or. any(bar0(:) < cryst(nph)%frag(i)%tlow)      
!
!        Traslo il frammento all'interno del range se si verificano contemporaneamente le 2 condizioni:
!        1) Il range non e' periodico in almeno una direzione
!        2) Il frammento e' fuori range in almeno una direzione
         if (any(not_periodic(:)) .and. (any(out_range(:))) ) then
!
!            Modifica le coordinate del frammento solo lungo le direzioni che non
!            soddisfano le 2 condizioni precedenti      
             where(not_periodic(:) .and. out_range) 
                   newpos(:) = rangef(:)/2
             elsewhere
                   newpos(:) = bar0(:)
             endwhere
!!!   !          write(6,*)'Frammento spostato in ',newpos,' da ',bar0
!
!            Ora sposta il frammento nella nuova posizione          
             call set_fragment_position(cryst(nph)%frag(i),newpos,cryst(nph)%at,0)
         endif       
!
         if (cryst(nph)%frag(i)%ref()) call get_modified_restraints(cryst(nph),cryst(nph)%frag(i)%pos,vres)
         if (cryst(nph)%frag(i)%ref()) call get_modified_abump(cryst(nph)%abres,cryst(nph)%frag(i)%pos,vares)
         do j=1,3      
            if (cryst(nph)%frag(i)%rcod(j) > 0) then             
                nkref = nkref + 1
                gpar(nkref)%str = atomstring(j)
                gpar(nkref)%val = 0  
                cryst(nph)%frag(i)%rcod(j) = nkref         
                if (not_periodic(j)) then     ! range non periodico
                    gpar(nkref)%low = cryst(nph)%frag(i)%tlow(j)/2.0
                    gpar(nkref)%high = cryst(nph)%frag(i)%thigh(j)/2.0
                    gpar(nkref)%bound = 0             
                else                          ! range periodico
                    gpar(nkref)%low = -bar0(j)
                    gpar(nkref)%high = 1-bar0(j)
                    gpar(nkref)%bound = 1             
                endif
                !no_H = fragment(i)%nat - number_of_hydrogens(atm(fragment(i)%pos))
                !write(0,*)'NO_H=',no_H
                !allocate(gpar(nkref)%pat(no_H))
                !gpar(nkref)%pat = pack(fragment(i)%pos, mask = atomic_number(atm(fragment(i)%pos)) /= H_at)
                !      write(0,*)'PAR=',gpar(nkref)%pat
                allocate(gpar(nkref)%pat(cryst(nph)%frag(i)%nat))
                gpar(nkref)%pat = cryst(nph)%frag(i)%pos   
                gpar(nkref)%typepar = EXT_TYPE
                gpar(nkref)%ph = nph
                call copy_array(gpar(nkref)%res,vres)
                call copy_array(gpar(nkref)%abres,vares)
            endif
         enddo
!
!        Rotazione frammento      
!
!        calcola il range per la rotazione random del frammento
         rangef(1) = cryst(nph)%frag(i)%rbound(2) - cryst(nph)%frag(i)%rbound(1)    
!
!        controlla se il range e' diverso da 1 e quindi non puo' essere periodico
         not_periodic(1) = abs(rangef(1) - 1.0) > 0.01
         do j=4,6
            if (cryst(nph)%frag(i)%rcod(j) > 0) then
                nkref = nkref + 1
                gpar(nkref)%str = string_rot(j-3)
                gpar(nkref)%val = XROT0(j-3)   
                cryst(nph)%frag(i)%rcod(j) = nkref
                if (not_periodic(1)) then
                    gpar(nkref)%low = max(XROT0(j-3) - rangef(1)/2,0.0)
                    gpar(nkref)%high = min(XROT0(j-3) + rangef(1)/2,1.0)
                    gpar(nkref)%bound = 0
                else
                    gpar(nkref)%low = 0.0
                    gpar(nkref)%high = 1.0
                    gpar(nkref)%bound = 1
                endif
                allocate(gpar(nkref)%pat(cryst(nph)%frag(i)%nat))
                gpar(nkref)%pat = cryst(nph)%frag(i)%pos  
                gpar(nkref)%typepar = EXT_TYPE
                gpar(nkref)%ph = nph
                call copy_array(gpar(nkref)%res,vres)
                call copy_array(gpar(nkref)%abres,vares)
            endif          
         enddo
      enddo   
!
!     Loop sulle rotazioni
      do i=1,cryst(nph)%numrot()
         if (cryst(nph)%rot(i)%rcod > 0) then
             nkref = nkref + 1         
             write(gpar(nkref)%str,'(a,i0)')'DOF(int) n.',i 
             gpar(nkref)%val = 0.0
             cryst(nph)%rot(i)%rcod = nkref
             rangef(1) = cryst(nph)%rot(i)%trbound(2) - cryst(nph)%rot(i)%trbound(1)  ! range di rotazione
             not_periodic(1) = abs(rangef(1) - 360) > 0.1                 ! valuta periodicita'
             if (not_periodic(1)) then     ! range non periodico
!
!                Converti i limiti delle torsioni in limiti degli angoli di rotazione normalizzati 
                 !write(0,*)i,'tors lim=',cryst(nph)%rot(i)%trbound(1),cryst(nph)%rot(i)%trbound(2)
                 gpar(nkref)%low = -(cryst(nph)%rot(i)%torsval - cryst(nph)%rot(i)%trbound(1))*dtor/twopi
                 gpar(nkref)%high = (cryst(nph)%rot(i)%trbound(2) - cryst(nph)%rot(i)%torsval)*dtor/twopi
                 if (cryst(nph)%rot(i)%mode == 2 .or. cryst(nph)%rot(i)%mode == 4) then  !FIX for out of range
                     gpar(nkref)%low = gpar(nkref)%low*2
                     gpar(nkref)%high = gpar(nkref)%high*2
                 elseif (cryst(nph)%rot(i)%mode == 3) then
                     gpar(nkref)%low = gpar(nkref)%low*3
                     gpar(nkref)%high = gpar(nkref)%high*3
                 endif
                 !write(0,*)'TORS=',i,twopi*gpar(nkref)%low,twopi*gpar(nkref)%high,cryst(nph)%rot(i)%torsval
                 !write(0,*)'TORS=',i,360*gpar(nkref)%low,360*gpar(nkref)%high,cryst(nph)%rot(i)%torsval
                 !write(0,*)'VALUE=',i,nint(torsion_value_for_rotat(cryst(nph)%rot(i),cryst(nph)%at,cryst(nph)%cell))
                 gpar(nkref)%bound = 0          
                 !write(6,*)i,'ang     =', -(rotation(i)%torsval - rotation(i)%trbound(1)),   &
                 !                          (rotation(i)%trbound(2) - rotation(i)%torsval)
                 !write(6,*)i,'ang norm=',gpar(nkref)%low,gpar(nkref)%high
                 out_range(1) = 0.0 < gpar(nkref)%low .or. 0.0 > gpar(nkref)%high
                 if (out_range(1)) then   ! torsione fuori range
                 !    write(0,*)'torsione fuori range'
                     newang = (gpar(nkref)%high - gpar(nkref)%low) / 2
                 endif
             else                          ! range periodico
                 gpar(nkref)%low = -0.5
                 gpar(nkref)%high = 0.5
                 gpar(nkref)%bound = 1          
             endif
             allocate(gpar(nkref)%pat(cryst(nph)%rot(i)%nat))
             gpar(nkref)%pat = cryst(nph)%rot(i)%pat
             gpar(nkref)%typepar = INT_TYPE
             gpar(nkref)%ph = nph
             call get_modified_restraints(cryst(nph),gpar(nkref)%pat,gpar(nkref)%res)
             call get_modified_abump(cryst(nph)%abres,gpar(nkref)%pat,gpar(nkref)%abres)
         endif
      enddo

!
!     Shifts and rigid bodies
      do i=1,cryst(nph)%nrigid()
         do j=1,3
            if (cryst(nph)%rigidb(i)%rcod(j) > 0) then
                nkref = nkref + 1             
                gpar(nkref)%val = 0.0
                cryst(nph)%rigidb(i)%rcod(j) = nkref
                !gpar(nkref)%low = -cryst(nph)%rigidb(i)%tra(j)/2.0
                !gpar(nkref)%high = cryst(nph)%rigidb(i)%tra(j)/2.0
                gpar(nkref)%low = -cryst(nph)%rigidb(i)%tra(j)
                gpar(nkref)%high = cryst(nph)%rigidb(i)%tra(j)
                !gpar(nkref)%low = 0.0 !-cryst(nph)%rigidb(i)%tra(j)/2.0
                !gpar(nkref)%high = 1.0 !cryst(nph)%rigidb(i)%tra(j)/2.0
                gpar(nkref)%bound = 0             
                allocate(gpar(nkref)%pat(cryst(nph)%rigidb(i)%nat))
                gpar(nkref)%pat(:) = cryst(nph)%rigidb(i)%at(:)  
                gpar(nkref)%typepar = INT_TYPE
                gpar(nkref)%ph = nph
                write(strat,'(i0,a)')gpar(nkref)%pat(1),' ('//trim(cryst(nph)%at(gpar(nkref)%pat(1))%lab)//')'  !TOFIX
                gpar(nkref)%str = trim(atomstring(j))//strat
                call get_modified_restraints(cryst(nph),gpar(nkref)%pat,gpar(nkref)%res)
                call get_modified_abump(cryst(nph)%abres,gpar(nkref)%pat,gpar(nkref)%abres)
            endif
         enddo
         do j=4,6
            if (cryst(nph)%rigidb(i)%rcod(j) > 0) then
                nkref = nkref + 1             
                gpar(nkref)%val = XROT0(j-3)
                cryst(nph)%rigidb(i)%rcod(j) = nkref
                if (j == 6) then  ! should be > 0
                    gpar(nkref)%low = XROT0(j-3)
                    gpar(nkref)%high = XROT0(j-3) + cryst(nph)%rigidb(i)%rot(j-3)/360
                else
                    !gpar(nkref)%low = XROT0(j-3) - 0.5*cryst(nph)%rigidb(i)%rot(j-3)/360
                    !gpar(nkref)%high = XROT0(j-3) + 0.5*cryst(nph)%rigidb(i)%rot(j-3)/360
                    gpar(nkref)%low = XROT0(j-3) - cryst(nph)%rigidb(i)%rot(j-3)/360
                    gpar(nkref)%high = XROT0(j-3) + cryst(nph)%rigidb(i)%rot(j-3)/360
                endif
                gpar(nkref)%bound = 0             
                allocate(gpar(nkref)%pat(cryst(nph)%rigidb(i)%nat))
                gpar(nkref)%pat(:) = cryst(nph)%rigidb(i)%at(:)  
                gpar(nkref)%typepar = INT_TYPE
                gpar(nkref)%ph = nph
                write(strat,'(i0,a)')gpar(nkref)%pat(1),' ('//trim(cryst(nph)%at(gpar(nkref)%pat(1))%lab)//')'  !TOFIX
                gpar(nkref)%str = trim(string_rot(j-3))//strat
                call get_modified_restraints(cryst(nph),gpar(nkref)%pat,gpar(nkref)%res)
                call get_modified_abump(cryst(nph)%abres,gpar(nkref)%pat,gpar(nkref)%abres)
            endif
         enddo

      enddo
!
!     Correzione per PO
      if (cryst(nph)%gcode > 0) then
          nkref = nkref + 1
          cryst(nph)%gpar = 1.0
          cryst(nph)%gcode = nkref
          gpar(nkref)%str = 'G'
          gpar(nkref)%val = 1.0
          gpar(nkref)%low = 0.0
          gpar(nkref)%high = 5.0
          gpar(nkref)%bound = 0             
          gpar(nkref)%typepar = PO_TYPE
          gpar(nkref)%ph = nph
      endif
!
!     Set  starting model for moves
      call copy_atoms(cryst(nph)%atm0,cryst(nph)%at)
   enddo
!
   end subroutine sa_prerefine 

!------------------------------------------------------------------------------------------------------

   subroutine get_modified_restraints(crystal,vat,vmod)
!
!  Array vmod contains restraints modified by move of atom vat
!
   use arrayutil
   use rrestr
   use crystal_phase
   type(crystal_phase_t), intent(inout)            :: crystal
   integer, dimension(:), intent(in)               :: vat
   integer, dimension(:), allocatable, intent(out) :: vmod
   integer                                         :: i,j,ncheck,n1
   logical                                         :: check
   integer                                         :: nzi,nzk
!   
   do i=1,crystal%numres()
      select case (crystal%res(i)%code)
         case (RESDIST)
           check = ncheck_container(vat,crystal%res(i)%na(1:2),ncheck)
           if (check) then  ! both atoms in the par
!              if symmetry are considered add res
               if (crystal%res(i)%sym(1) == 0 .or. crystal%res(i)%sym(1) == 2) then 
                   call push_back_array(vmod,i)
               endif
           else
!              add if almost one atom is modified by par
               if (ncheck >= 1) call push_back_array(vmod,i)
           endif

         case (RESANGLE)
           check = ncheck_container(vat,crystal%res(i)%na(1:3),ncheck)
           if (check) then
               if (any(crystal%res(i)%sym /= 1)) call push_back_array(vmod,i)  ! add if symmetry equivalent are considered
           else
!              add if almost one atom is modified by par
               if (ncheck >= 1) call push_back_array(vmod,i)
           endif

         case (RESBV)
           n1 = crystal%res(i)%na(1)
           if (any(vat(:) == n1)) then
               call push_back_array(vmod,i)
           else
               nzk = crystal%at(n1)%kscatt()
               do j=1,size(vat)
                  nzi = crystal%at(vat(j))%kscatt()
                  if (crystal%bvpar(nzi,nzk)%rmax > 0) then
                      call push_back_array(vmod,i)
                      exit
                  endif
               enddo
           endif

      end select
   enddo
!
   end subroutine get_modified_restraints
  
!----------------------------------------------------------------------------------------------------

   subroutine gopt_set_ncycle(info,ncy)
!
!  Compute automatically number of moves for each step of temperature
!   
   USE arrayutil
   USE rotationmod
   USE variables, only: cryst
   USE crystal_phase
   type(opti_info_type), intent(in) :: info
   integer, intent(out)             :: ncy
   real                             :: const
   real                             :: cfrag
   real                             :: cdofe
   real                             :: crot
   real                             :: ctherm
   real                             :: cshifta
   real                             :: crotc
   integer                          :: flex,ntotfrag,nph
!
   const = 10.0
   cfrag = 3.0    ! contributo per ogni frammento oltre al primo
   cdofe = 2.0    ! contributo per ogni grado di liberta esterno
   crot  = 1.0    ! contributo per ogni rotazione
   ctherm = 0.1   ! contributo per ogni fattore termico
   cshifta = 0.1  ! contributo per ogni shift
   crotc  = 0.8   ! contributo per ogni rotazione concatenata
!   
!  Compute total number of refined fragment
   ntotfrag = 0
   do nph=1,numphase(cryst)
      ntotfrag = ntotfrag + cryst(nph)%numfrag()
   enddo
!
!corr   if(cryst(1)%numfrag() == 1 .and. cryst(1)%numrot() <= 2) then 
   if(ntotfrag == 1 .and. info%nprot <= 2) then 
!           
!     Possibile formula: 20 + 10*2**numrot
      ncy = nint(20.0 + 10*2**info%nprot + const*(info%npb + info%npshift))
   else           
!
!     Molecular flexibility
      flex = 0
      do nph=1,numphase(cryst)
         flex = flex + num_connected_rotation(cryst(nph)%rot)
      enddo
! 
      ncy = nint ( const*(cfrag*(ntotfrag-1) + cdofe*info%dofe +  &
                   crot*info%nprot  + crotc*flex   +              &
                   ctherm*info%npb  + cshifta*info%npshift) )  
   endif                
!
   end subroutine gopt_set_ncycle
      
!----------------------------------------------------------------------------------------------------

   subroutine gopt_set_algo(info)
   USE variables, only: cryst
   USE crystal_phase
   type(opti_info_type), intent(inout) :: info
   integer                             :: nph,num_frag
!
   num_frag = 0
   do nph=1,numphase(cryst)
      num_frag = num_frag + cryst(nph)%numfrag()
   enddo
   if (info%nprot <= 5 .and. num_frag <= 2 .and. info%npar <= 11) then
!corr   if (info%npar <= 11) then ! e.g. 5 external DOFs and 1 fragment
       info%algo = BBBC
   else
       info%algo = SA_LOCAL
   endif
!
   end subroutine  gopt_set_algo

!----------------------------------------------------------------------------------------------------

   subroutine xpar_to_atm(x,crystal)
!
!  Convert x parameters in crystal structure parameters
!   
   USE fastfcal
   USE atom_type_util
   USE cgeom
   USE rrestr
   USE constraints
   USE trig_constants
   USE fragmentmod
   USE symmgrid
   USE crystal_phase
   USE rigid_body
   use rand_mod
!
   real, dimension(:), intent(in)             :: x       ! parametri SA
   type(crystal_phase_t), intent(inout)       :: crystal
   integer                                    :: j
   real, dimension(3)                         :: xtra
   real, dimension(3)                         :: xrot
   integer                                    :: scodt,scodr
   integer                                    :: i
   real, dimension(3)                         :: pa1,pa2
   type(atom_type), dimension(:), allocatable :: atmr 
   real                                       :: theta 
   integer                                    :: natrot
   integer                                    :: natom,rat
   real, dimension(3), parameter              :: XROT0 = (/0.5,0.5,0.0/) 
     !real :: trsval
!
   crystal%at = crystal%atm0
   natom = crystal%natoms()   
!
!  Thermal factor
   do j=1,natom
      if (crystal%at(j)%rcod(4) > 0) then      ! B
          crystal%at(j)%biso = x(crystal%atm0(j)%rcod(4))
      endif
   enddo
   do i=1,nconstraints(crystal%sconstr)
      if (crystal%sconstr(i)%ref > 0) then
          select case (crystal%sconstr(i)%code) 
             case (CONSTR_B)
                crystal%at(crystal%sconstr(i)%vat)%biso = crystal%sconstr(i)%coef*x(crystal%sconstr(i)%ref)
             case (4)
          end select
      endif
   enddo
!
!  Frammenti
   do i=1,crystal%numfrag()
      scodt = 0
      do j=1,3                           ! traslazione
         if (crystal%frag(i)%rcod(j) > 0) then
             xtra(j) = x(crystal%frag(i)%rcod(j))
             scodt = scodt + 1
         else
             xtra(j) = 0
         endif
      enddo
!      
      scodr = 0
      do j=4,6                           ! rotazione
         if (crystal%frag(i)%rcod(j) > 0) then
             xrot(j-3) = x(crystal%frag(i)%rcod(j))
             scodr = scodr + 1
         else
             xrot(j-3) = XROT0(j-3)
         endif
      enddo      
      if (scodt + scodr > 0) then
          call move_fragment(crystal%at,crystal%frag(i)%pos,xtra,xrot,scodt,scodr,crystal%cell,crystal%frag(i)%rat)
      endif          
   enddo 
!
!  Set center of rotation before moves, rotations and rigid body moves
!  can modify rotation centre and create correlation between parameters
   do i=1,crystal%nrigid()
      if (crystal%rigidb(i)%rat > 0) then
          crystal%rigidb(i)%xc = crystal%at(crystal%rigidb(i)%rat)%xc
      else
          crystal%rigidb(i)%xc = baricentro(crystal%at(crystal%rigidb(i)%at))
      endif
   enddo
!
!  Rotations. Don't move rotation after rigid body moves: rigid body can change rotation axes
   do i=1,crystal%numrot()
      if (crystal%rot(i)%rcod > 0) then
          pa1 = matmul(crystal%cell%get_ortom(),crystal%at(crystal%rot(i)%pax(1))%xc)
          pa2 = matmul(crystal%cell%get_ortom(),crystal%at(crystal%rot(i)%pax(2))%xc)
          !theta = twopi*pi*x(crystal%rot(i)%rcod)
          theta = get_theta(x(crystal%rot(i)%rcod),crystal%rot(i), &
                            gpar(crystal%rot(i)%rcod)%low,gpar(crystal%rot(i)%rcod)%high)
          natrot = crystal%rot(i)%nat
          call new_atoms(atmr,crystal%natoms())
          atmr(:natrot) = crystal%at(crystal%rot(i)%pat)
          call rotate_atoms(atmr(:natrot),pa1,direction_cos(pa1,pa2),theta,crystal%cell)
          crystal%at(crystal%rot(i)%pat) = atmr(:natrot)
!trsval = nint(torsion_value_for_rotat(crystal%rot(i),crystal%at,crystal%cell))
!if ((trsval < crystal%rot(i)%trbound(1)-1 .and. trsval > crystal%rot(i)%trbound(4)+1) .or. &
!    (trsval > crystal%rot(i)%trbound(2)+1 .and. trsval < crystal%rot(i)%trbound(3)-1)) then
!write(70,*)'VALUE=',i,trsval
!endif
      endif
   enddo
!
   do i=1,crystal%nrigid()
      scodt = 0
      do j=1,3
         if (crystal%rigidb(i)%rcod(j) > 0) then
             xtra(j) =  x(crystal%rigidb(i)%rcod(j))
             scodt = scodt + 1
         else
             xtra(j) = 0
         endif
      enddo
      scodr = 0
      do j=4,6
         if (crystal%rigidb(i)%rcod(j) > 0) then
             xrot(j-3) =  x(crystal%rigidb(i)%rcod(j))
             scodr = scodr + 1
         else
             xrot(j-3) = XROT0(j-3)
         endif
      enddo

      if (scodt + scodr > 0) then
          if (scodt > 0) then
              !xtra = getPoint(xtra)
              !xtra = xtra * crystal%rigidb(i)%tra
              xtra = matmul(crystal%cell%get_ortoi(),xtra)
          endif
                    !xx = crystal%at(1)%xc
          if (crystal%rigidb(i)%rat > 0) then
              rat = crystal%rigidb(i)%rat
          else
              rat = -1
          endif
          call move_fragment(crystal%at,crystal%rigidb(i)%at,xtra,xrot,scodt,scodr,crystal%cell,   &
                             rat,crystal%rigidb(i)%xc)
             !write(71,'(a,f10.3)')'DD=',distanzaC(xx,crystal%at(1)%xc,crystal%cell%get_g())
          !crystal%at(iat)%xc = crystal%at(iat)%xc + matmul(crystal%cell%get_ortoi(),xshift)
      endif
   enddo
!
!  Dynamical occupancy correction
   if (crystal%doc_active()) then
       call crystal%make_symmetry_cell()
       call crystal%allocate_grid(cutoff=1.0)
       if (crystal%doc_active_all()) then
!old_occ           call compute_doc_st_all(atom,crystal%spg,crystal%cell%get_g(),1.0,asymtab)
           !call compute_doc_grid_all(crystal%at,crystal%cell,1.0)
           call crystal%fill_grid()
           call crystal%compute_doc(1.0)
           call fcalc_for_tfocc(crystal%ref,crystal%at,[(i,i=1,natom)],crystal%sfact,crystal%kpscat)
       else
           !call compute_doc_atoms(atom,crystal%spg,crystal%cell%get_g(),1.0)
           !call compute_doc(atom,crystal%spg,crystal%cell%get_g(),1.0,crystal%vetdoc)  ! test also this!
           !call compute_doc_grid_vet(crystal%at,crystal%cell,1.0,crystal%vetdoc)
           call crystal%fill_grid(crystal%vetdoc)
           call crystal%compute_doc(1.0,crystal%vetdoc)
           call fcalc_for_tfocc(crystal%ref,crystal%at(crystal%vetdoc),crystal%vetdoc,crystal%sfact,crystal%kpscat)
       endif
!TODO:   call fcalc_for_tfocc is unusefull when the par is B
   endif
!
!  Orientazione preferenziale
   if (crystal%gcode > 0) then
       crystal%gpar = x(crystal%gcode)
   endif
!  
   end subroutine xpar_to_atm

!----------------------------------------------------------------------------------------------------

   real function get_theta(x, rot, low, high) result(theta)
!
!  theta is the rotation to apply to torsion in radiant
!  -1 < x, low, high < 1 
!  -180 < theta < 180
!
   use trig_constants
   use rotationmod
   real, intent(in)                :: x, low, high
   type(rotation_type), intent(in) :: rot
   real                            :: l2, h2, l3, h3
   real, parameter                 :: R120 = twopi/3
!
   select case (rot%mode)
     case (1)
      theta = twopi * x 

     case (2) ! non planar bimodal torsion
      l2 = low/2   ! negative
      h2 = high/2  ! positive
      if (x <= h2 .and. x >= l2) then
          theta = twopi * x 
      elseif (x < l2) then
!         non planar torsion: apply reflection respect to 0
!         es. -30 -> 30
          theta = -2*(dtor*rot%torsval + twopi*(x - l2))
      elseif (x > h2) then
          theta = -2*(dtor*rot%torsval + twopi*(x - h2))
      !if (kpr) write(70,'(a,5(a,f8.2),a,i4)') 'H', &
      !    'x=',360*x,' low',360*low,' high',360*high,' value=',rot%torsval,' theta=',rtod*theta
      endif

     case (4) ! planar bimodal torsion
      l2 = low/2   ! negative
      h2 = high/2  ! positive
      if (x <= h2 .and. x >= l2) then
          theta = twopi * x 
      elseif (x < l2) then
!         planar torsion - apply shift 180 but remaining in the range -180 - 180
!         es. 30 -> -150; -30 -> 150
          theta = twopi * (x - l2) + pi
          !write(70,'(a,4(a,f8.2),a,i4)') '<L', &
          !'x=',360*x,' low',360*low,' high',360*high,' theta=',rtod*theta 
      elseif (x > h2) then
          theta = twopi * (x - h2) - pi
          !write(70,'(a,4(a,f8.2),a,i4)') '>H', &
          !'x=',360*x,' low',360*low,' high',360*high,' theta=',rtod*theta
      endif

     case (3) ! trimodal torsion (apply +/-120)
      l3 = low/3   ! negative
      h3 = high/3  ! positive
      l2 = 2*l3
      h2 = 2*h3
      if (x <= h3 .and. x >= l3) then
          theta = twopi * x 
      elseif (x > h3 .and. x <= h2) then
          theta = R120 + twopi*(x - h3)
          !if (kpr) write(0,*)'H2', &
          !'x=',360*x,' low',360*low,' high',360*high,' theta=',rtod*theta
      elseif (x < l3 .and. x >= l2) then
          theta = R120 + twopi*(x - l3)
      elseif (x > h2) then
          theta = -R120 + twopi*(x - h2)
      elseif (x < l2) then
          theta = -R120 + twopi*(x - l2)
      endif

   end select
!
   end function get_theta

!----------------------------------------------------------------------------------------------------

   subroutine xpar_to_crystals(x)
   use variables, only: cryst
   use crystal_phase
   real, dimension(:), intent(in) :: x
   integer                        :: nph
   do nph=1,numphase(cryst)
      call xpar_to_atm(x,cryst(nph))
   enddo
   end subroutine xpar_to_crystals

!----------------------------------------------------------------------------------------------------

   subroutine search_overlapped_atoms(atom,gmat,spg,modifi)
   USE CGEOM
   USE spginfom, only: spaceg_type
   USE atom_type_util
   type(atom_type), dimension(:), intent(inout) :: atom   
   real, dimension(3,3), intent(in)             :: gmat
   type(spaceg_type), intent(in)                :: spg
   logical, intent(in), optional                :: modifi
   logical                                      :: modif
   integer                                      :: natc
   real, dimension(3,spg%nsymop)                :: xeq
   real                                         :: dd
   integer                                      :: i,j,k
   real, dimension(3) :: ktra
!
   natc = size(atom)
   if (present(modifi)) then
       modif = modifi
   else
       modif = .true.
   endif
!
!corr   ivetocc(:) = 1
   do i=1,natc-1
      call get_equivalent1(atom(i)%xc,spg,xeq)   !!!!!,neqa)
!corr      if (all(vet /= i)) cycle
      if (atom(i)%doc) then
          do j=i+1,natc
!corr             if (all(vet /= j)) cycle
             if (atom(j)%doc) then
                 if (atom(i)%get_nz() == atom(j)%get_nz()) then
                     do k=1,spg%nsymop
                        if (k == 1 .and. j < i) cycle  ! combinazione già testata
                        !dd = xdisteq(xeq(:,k),atom(j)%xc,xgg)
                        call xdisteqs(atom(j)%xc,xeq(:,k),gmat,dd,ktra)
                        !dd = distanzaC(atom(j)%xc,xeq(:,k),gmat)
                        if (dd < DISTDOC) then
!corr                            ivetocc(i) = ivetocc(i) + 1
!corr                            ivetocc(j) = ivetocc(j) + 1
                            if (modif) then
                                !if (dd < 0.01) then   !! this is test
                                !    atom(j)%xc = atom(i)%xc   !!!FIXME - but it is necessary reconnect j atom if k/=1
                                !else
                                    atom(j)%xc = atom(j)%xc + ktra  ! ora i 2 atomi sono nella stessa posizione
                                !endif
                        !call xdisteqs(atom(j)%xc,xeq(:,k),gmat,dd,ktra)
                            !write(0,'(a,a,a,a,f0.4,a,i0)')'vicinid=',atom(i)%lab,atom(j)%lab,'dd=',dd,' op=',k
                            !         write(0,*)'atom=',trim(atom(i)%lab),atom(i)%xc
                            !         write(0,*)'atom=',trim(atom(j)%lab),atom(j)%xc
                            endif
                            !non funziona atom(j)%xc = xeq(:,k) - ktra
                        endif
                     enddo
                 endif
             endif
          enddo
      endif
   enddo
!corr   do i=1,natc
!corr      atom(i)%och = atom(i)%och/ivetocc(i)
!corr   enddo
!
   end subroutine search_overlapped_atoms

!----------------------------------------------------------------------------------------------------

   subroutine search_overlapped_atoms_print(atom,gmat,spg)
   USE General, only:lo
   USE CGEOM
   USE strutil
   USE nr
   USE spginfom, only: spaceg_type
   USE atom_type_util
   type(atom_type), dimension(:), intent(in) :: atom   
   real, dimension(3,3), intent(in)          :: gmat
   type(spaceg_type), intent(in)             :: spg
   integer                                      :: natc
   real, dimension(3,spg%nsymop)                :: xeq 
   real                                         :: dd
   integer                                      :: i,j,k
   real, dimension(3)                           :: ktra
   integer, dimension(size(atom), size(atom))   :: tableat
   real, dimension(size(atom), size(atom))   :: tabled   !!!!FIXME  da migliorare
   integer                                      :: nrem
   integer, dimension(size(atom))               :: vrem
   integer, allocatable, dimension(:)           :: iord
!
   natc = size(atom)
   tableat(:,:) = 0
   do i=1,natc
      call get_equivalent1(atom(i)%xc,spg,xeq)   !!!!!,neqa)
      do j=1,natc
         if (i==j) cycle
         if (atom(i)%get_nz() == atom(j)%get_nz()) then
             do k=1,spg%nsymop
                if (k == 1 .and. j < i) cycle  ! combinazione già testata
                call xdisteqs(atom(j)%xc,xeq(:,k),gmat,dd,ktra)
                if (dd < DISTDOC) then
                    !write(0,'(" WARNING: Atoms ",a," and ",a," are in the same position")')   &
                    !trim(atom(i)%lab),trim(atom(j)%lab)
                    tableat(i,j) = 1
                    tableat(j,i) = 1
                else if (dd < DISTOCC + 0.3) then
                    !write(0,'(" WARNING: Critical distance between ",a," and ",a," (d=",f0.3,")")')   &
                    !trim(atom(i)%lab),trim(atom(j)%lab),dd
                    tableat(i,j) = -1
                    tableat(j,i) = -1
                    tabled(i,j) = dd
                    tabled(j,i) = dd
                endif
             enddo
         endif
      enddo
   enddo
!
   do i=1,natc-1
      do j=i+1,natc
         if (tableat(i,j) == 1) then
             write(lo,'(" WARNING: Atoms ",a," and ",a," are in the same position")')   &
             trim(atom(i)%lab),trim(atom(j)%lab)
         elseif (tableat(i,j) == -1) then
             write(lo,'(" WARNING: Critical distance between ",a," and ",a," (d=",f0.3,")")')   &
             trim(atom(i)%lab),trim(atom(j)%lab),tabled(i,j)
         endif
      enddo
   enddo
!
   nrem = 0
   do i=1,natc-1
      do j=i+1,natc
         if (abs(tableat(i,j)) > 0) then
!             
!            cosa rimuovere: i o j?
             !wej = sum(abs(tableat(i,:))) 
             !wei = sum(abs(tableat(j,:)))
             !if (wei > wej) then
             !    write(lo,*)'posso rimuovere ',trim(atom(i)%lab),wei,wej
             !    tableat(i,:) = 0
             !    tableat(:,i) = 0
             !    exit
             !else
              nrem = nrem + 1
              vrem(nrem) = j
                 !!!!write(lo,*)'posso rimuovere ',trim(atom(j)%lab)
                 tableat(j,:) = 0
                 tableat(:,j) = 0
                 cycle
             !endif
         endif
      enddo
   enddo
!
!  Stampa gli atomi da eliminare
   if (nrem > 0) then
       if (nrem > 1) then   ! ordina il vettore vrem
           allocate(iord(nrem))
           call indexx(vrem(:nrem),iord)
           vrem(:nrem) = vrem(iord)
       endif
       call write_svet(atom(vrem(:nrem))%lab,nrem,lo,' The following atoms could be removed: ')
   endif
!
   end subroutine search_overlapped_atoms_print

!----------------------------------------------------------------------------------------------------

   real function tfunk(x,hpar)
   USE variables, only:cryst
   USE anti_bump
   USE rrestr
   USE crystal_phase
!
   real, dimension(:), intent(in) :: x
   integer, intent(in)            :: hpar
   integer, parameter             :: kpr = 0
   real                           :: chid
   integer                        :: nph
!
!  Passa x in atm
   call timesa(SATIME_MOVES)%tic()
   call xpar_to_crystals(x)
   call timesa(SATIME_MOVES)%toc()
!
   call timesa(SATIME_CF)%tic()
         
   tfunk = chisquared_xdata(hpar,ginfo)
!
!  Contributo restraints
!!!!FIX for multiphase by using gpar(hpar)%ph
   do nph=1,numphase(cryst)
      if (cryst(nph)%numres() > 0) then

          !write(71,*)'CHIRES for par ',hpar

          if (hpar == 0) then
              chid=chisquared_res(cryst(nph))
              call res_save_contrib(cryst(nph)%res)
!
              !write(71,*)'SAVE=',cryst(1)%res%contrib
!
          else
              if (gpar(hpar)%ph == nph) then     ! recompute contributions for phase nph
                  chid=chisquared_res(cryst(nph),gpar(hpar)%res)
              else
                  chid=chi2tot(cryst(nph)%res)
              endif
          endif
            !write(70,*)'TF=',tfunk+chid,tfunk,chid
          tfunk = tfunk + chid
      endif
   enddo
!
!  Contributions for anti-bump
!!!!FIX for multiphase by using gpar(hpar)%ph
   do nph=1,numphase(cryst)
      if (cryst(nph)%numresab() > 0) then
          if (hpar == 0) then
!
!             Recompute all restraints
              tfunk = tfunk + antibump_function(cryst(nph)%abres,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell%get_g(),.false.)
              call res_save_contrib(cryst(nph)%abres)
          else
              if (gpar(hpar)%ph == nph) then     ! recompute contributions for phase nph
                  tfunk = tfunk + antibump_function(cryst(nph)%abres,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell%get_g(),   &
                                                    .false.,gpar(hpar)%abres)
              else                               ! add contributions of other phases
                  tfunk = tfunk + chi2tot(cryst(nph)%abres)  
              endif
          endif
      endif
   enddo
         
   call timesa(SATIME_CF)%toc()
!   
   end function tfunk

!----------------------------------------------------------------------------------------------------

   subroutine gopt_statusbar(ntot,ntemp)
   USE messagemod
   USE simulated_anneal
   USE timer
   integer, intent(in)                        :: ntot
   integer, intent(in)                        :: ntemp
   real                                       :: init_time = 0
   real                                       :: end_time
   real                                       :: range_time = 1.0 ! intervallo di tempo dopo il quale updatare la statusbar
   logical                                    :: oktime 
!
   call timesa(SATIME_GRAPH)%tic() ! calcola il tempo impiegato dalle operazioni grafiche  
!
!  Controlla se e trascorso un tempo pari a range_time
   call cpu_time(end_time)
   oktime = end_time - init_time > range_time
   if (oktime) init_time = end_time
!   
!  Aggiorna informazioni sulla statusbar
   if (oktime .or. ntemp == numcyc_calc) then
       call write_message('Moves:',inum=ntot,pos=vmess(LINE_MOVES),kproc=0)
       call write_message('Time:'//secs_to_string(timesa(SATIME_TOT)%get_time(),.true.),pos=vmess(LINE_TIME),kproc=0)
       call progress_bar_update(cval=ntemp,fval=numcyc_calc,kproc=0)
   endif
!
   call timesa(SATIME_GRAPH)%toc() ! calcola il tempo impiegato dalle operazioni grafiche  
!   
   end subroutine gopt_statusbar   
      
!----------------------------------------------------------------------------------------------------

   real function ratio_value(fopt,gopti) result(ratio)
   USE RefinecomRef, only:rinfo
   type(opti_info_type), intent(in) :: gopti     
   real, intent(in)                 :: fopt
!
   if (.not.is_profile_function(gopti%cftype)) then
       !call sa_profile(sqrt(ref%fc))
       ratio = rinfo%rwp/gopti%rwp_leb    ! migliorare forse si puo' usare sole questa funzione
   else
       ratio = fopt/gopti%rwp_leb
   endif
!
   end function ratio_value

!----------------------------------------------------------------------------------------------------

   subroutine sa_profile(square)
   USE variables, only:cryst
   USE rietv
   USE crystal_phase
   use RefinecomRef, only: rfcond

   logical, intent(in) :: square
   integer             :: nph
!
   if (square) then
       do nph=1,numphase(cryst)
          cryst(nph)%ref(:)%fc = sqrt(cryst(nph)%ref(:)%fc)
          cryst(nph)%ref(:)%pk = 10
       enddo
   else
       do nph=1,numphase(cryst)
          cryst(nph)%ref(:)%pk = 10
       enddo
   endif
   call Count0(rfcond)
!
   end subroutine sa_profile

!----------------------------------------------------------------------------------------------------
   
   subroutine gopt_draw(sdraw,pdraw,gopti)
   USE view
   USE atom_type_util
   USE connect_mod
   USE ilmviewmod
   USE fouriermap
   USE reflection_type_util
   USE variables, only: cryst,dataset
   USE fragmentmod
   USE fourier_mod
   USE counts
   USE mapinfo
   USE General, only: neutro, is_anomalous
   logical, intent(in)                              :: sdraw,pdraw
   type(opti_info_type), intent(in)                 :: gopti    
   type(atom_type), dimension(:), allocatable       :: atmsav
   type(bond_type), dimension(:), allocatable       :: legm
   type(reflection_type), dimension(:), allocatable :: refmap
   integer                                          :: nfrag,nph
   real                                             :: f000
!
   call timesa(SATIME_GRAPH)%tic ! calcola il tempo impiegato dalle operazioni grafiche  
!
!  Disegna la struttura
   if (sdraw) then
       nph = active_phase
       nfrag = cryst(nph)%numfrag()
       if (nfrag > 1 .or. cryst(nph)%doc_active()) call copy_atoms(atmsav,cryst(nph)%at)
       if (nfrag > 1) call regroup_fragment(cryst(nph)%frag,cryst(nph)%at,cryst(nph)%cell,cryst(nph)%spg)
       if (cryst(nph)%doc_active()) then
           call search_overlapped_atoms(cryst(nph)%at,cryst(nph)%cell%get_g(),cryst(nph)%spg,.true.)
       endif
       if (nfrag > 1) then 
           call copy_bonds(legm,cryst(nph)%bond)
           call connect_fragment(cryst(nph)%frag,cryst(nph)%at,legm,cryst(nph)%cell,.false.)
           call ilmview_draw(cryst(nph)%at,legm,cryst(nph)%elem,cryst(nph)%cell,cryst(nph)%spg,resize=.not.(ilmview_info(1,1) == 1))  ! no resize with visible cell
       else
           call ilmview_draw(cryst(nph)%at,cryst(nph)%bond,cryst(nph)%elem,cryst(nph)%cell,   &
                             cryst(nph)%spg,resize=.not.(ilmview_info(1,1) == 1))  ! no resize with visible cell
       endif
       if (nfrag > 1 .or. cryst(nph)%doc_active()) cryst(nph)%at(:) = atmsav(:)
       if (is_map_enabled(cryst(nph)%mapinfo)) then
           call map_set_centre(baricentro(cryst(nph)%at))
           call copy_ref(refmap,cryst(nph)%ref)
           call sfcalc(refmap,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%elem,neutro,is_anomalous)  ! compute fase
           if (cryst(nph)%mapinfo%ftype == FOUR_FO .or. cryst(nph)%mapinfo%ftype == FOUR_FC) then
               f000 = get_f000(cryst(nph)%at,cryst(nph)%elem,cryst(nph)%spg,dataset(nph)%radtype)
           else
               f000 = 0
           endif
           call jav_show_map(cryst(nph)%mapinfo%ftype,dvalue(dataset(1)%tmax,dataset(1)%wave(1)),refmap,    &
                             'SA map',sogmap=cryst(nph)%mapinfo%sog,fctype=1,scalem=.true.,f000=f000,cell=cryst(nph)%cell)
       endif
   endif
!   
!  Disegna il profilo
   if (pdraw) then
       call sa_profile(square=.true.)
       if (kzoomn) then
           call vedinew(iAct_sav,xzleft,xzright,xzleft,xzright,lsetstyle=.false.,rescale=0)
       else
           call vedinew(iAct_sav,1,gopti%nco,1,gopti%nco,lsetstyle=.false.,rescale=0)
       endif
   endif
   call timesa(SATIME_GRAPH)%toc()   
!
   end subroutine gopt_draw

!----------------------------------------------------------------------------------------------------

   real function chisquared_xdata(hpar,gopti) result(chix)
   USE fastfcal
   USE variables, only: cryst
   USE reflection_type_util
   USE crystal_phase
   integer, intent(in)              :: hpar
   type(opti_info_type), intent(in) :: gopti    
   integer                          :: i,natom,nph,hp
!
!  Calcola fattori di struttura
   if (hpar > 0) then       
       nph = gpar(hpar)%ph
       if (gpar(hpar)%typepar == THERMALF_TYPE) then 
           call fcalc_from_thermalf(cryst(nph)%ref,cryst(nph)%at(gpar(hpar)%pat),gpar(hpar)%pat,cryst(nph)%spg,cryst(nph)%sfact)
       elseif (gpar(hpar)%typepar == PO_TYPE) then 
           call calcola_correzioneOP(cryst(nph)%ref,cryst(nph)%gpar,cryst(nph)%hklpo,cryst(nph)%cell,cryst(nph)%spg)
       else
           call fcalc_from_atm(cryst(nph)%ref,cryst(nph)%at(gpar(hpar)%pat),gpar(hpar)%pat,cryst(nph)%spg,cryst(nph)%sfact)
       endif    
   else
       do nph=1,numphase(cryst)
          natom = cryst(nph)%natoms()
          if (cryst(nph)%gcode > 0) then
              call calcola_correzioneOP(cryst(nph)%ref,cryst(nph)%gpar,cryst(nph)%hklpo,cryst(nph)%cell,cryst(nph)%spg)
          endif
          do hp=1,size(gpar)
             if (gpar(hp)%ph == nph .and. gpar(hp)%typepar == THERMALF_TYPE) then
                 !call fcalc_for_tfocc(cryst(nph)%ref,cryst(nph)%at,[(i,i=1,natom)],cryst(nph)%sfact,cryst(nph)%kpscat)
                 call fcalc_for_tfocc(cryst(nph)%ref,cryst(nph)%at(gpar(hp)%pat),gpar(hp)%pat,cryst(nph)%sfact,cryst(nph)%kpscat)
                 call fcalc_save_thermalf(gpar(hp)%pat,cryst(nph)%sfact,cryst(nph)%sfactc)
             endif
          enddo
          call fcalc_from_atm(cryst(nph)%ref,cryst(nph)%at,(/(i,i=1,natom)/),cryst(nph)%spg,cryst(nph)%sfact)
          call fcalc_save_atm([(i,i=1,natom)],cryst(nph)%spg,cryst(nph)%sfact,cryst(nph)%sfactc)
       enddo
   endif    
   !call check_fcalc(cryst(1)%ref,cryst(1)%at,cryst(1)%spg,cryst(1)%elem,hpar)   ! diagnostica degli fc
   !call check_fcalc(cryst(2)%ref,cryst(2)%at,cryst(2)%spg,cryst(2)%elem,hpar)   ! diagnostica degli fc

   select case(gopti%cftype)
      case (RWPC_COSTF)
        chix = sqrt(chisquared_profile_new())/rwden

      case (RWP_COSTF)
        !chix = sqrt(chisquared_profile(gopti))/rwden
        chix = sqrt(chisquared_profile_new())/rwden

      case (CHI2FACT)
        chix = sqrt(chisquared_factor())/rwden

      case (CHI2I_SD)
        chix = sqrt(chisquared_int())/rwden

      case (CHI2I)
        chix = sqrt(chisquared_int())/rwden

      case (CHI2I_NEW)
        chix = sqrt(chisquared_int_new(gopti))/rwden

      case (CHI2_COSTF)
        chix = chisquared_profile_new()

      case (CHI2GRU)
        chix = sqrt(chisquared_int_groups())/rwden
   end select
!
   end function chisquared_xdata

!----------------------------------------------------------------------------------------------------
!corr
!corr   real function chisquared_profile(gopti) result(chi)
!corr   USE counts
!corr   !USE genlsq, only:weilsq
!corr   USE RIETV
!corr   USE variables, only: dataset,cryst
!corr   USE crystal_phase
!corr   type(opti_info_type), intent(in) :: gopti    
!corr   real                :: scal 
!corr     real :: sumyco
!corr!
!corr!  Calcola il profilo
!corr   call genera_profilo2(sumyco)
!corr   if (numphase(cryst) == 1) then
!corr       scal = sumycossb/sumyco
!corr!
!corr       call sscal(gopti%nco,scal,cryst(1)%yc,1)  
!corr!
!corr!      Calcola Rwp
!corr       !chi = sum(weilsq(:gopti%nco)*(dataset(1)%ys-cryst(1)%yc(:gopti%nco))**2)
!corr       chi = sum(dataset(1)%wei*(dataset(1)%ys-cryst(1)%yc(:gopti%nco))**2)
!corr   else
!corr   endif
!corr!   
!corr   end function chisquared_profile
!corr      
!----------------------------------------------------------------------------------------------------

   real function chisquared_profile_new() result(chi)
   USE RIETV
   USE variables, only: dataset,cryst
   USE genlsq
   USE crystal_phase
   USE type_constants, only: dp
   real :: scal 
   real :: sumyco
   integer                               :: nph,nc,ier
   real(dp), dimension(:), allocatable   :: bm
   real(dp), dimension(:,:), allocatable :: am,ainv
   real(dp), dimension(:), allocatable   :: crscal
!
   call genera_profilo2(sumyco)   !!!!!,dataset(1)%nwave)
   if (numphase(cryst) == 1) then
               !call scalefit(cryst(1)%yc,dataset(1)%ys,scal,chi,1/sqrt(dataset(1)%wei))
               !call scalefit(cryst(1)%yc,dataset(1)%ys,scal,chi,sqrt(abs(dataset(1)%ys)))
        !call scalefit(cryst(1)%yc,dataset(1)%ys,scal,chi)
       scal = sumycossb/sumyco
       nc = size(dataset(1)%ys)
!
       !call sscal(size(cryst(1)%yc),scal,cryst(1)%yc,1)  
       call sscal(nc,scal,cryst(1)%yc(:nc),1)  
!
!      Calcola Rwp
       chi = sum(dataset(1)%wei*(dataset(1)%ys-cryst(1)%yc(:nc))**2)
   else
!
!      Compute scale factor from lsq
       nc = size(dataset(1)%ys)
       allocate(am(nc,numphase(cryst)),bm(nc))
       allocate(crscal(numphase(cryst)),ainv(numphase(cryst),numphase(cryst)))
       do nph=1,numphase(cryst)
          am(:,nph) = sqrt(dataset(1)%wei)*cryst(nph)%yc(:nc)
       enddo
       bm(:) = sqrt(dataset(1)%wei)*dataset(1)%ys
       call solve_linlsq(am,bm,crscal,ier,linv=.true.,ainv=ainv)
       cryst%scal = crscal
       dataset(1)%yc(:nc) = cryst(1)%scal*cryst(1)%yc(:nc)
       do nph=2,numphase(cryst)
          dataset(1)%yc(:nc) = dataset(1)%yc(:nc) + cryst(nph)%scal*cryst(nph)%yc(:nc)
       enddo
       chi = sum(dataset(1)%wei*(dataset(1)%ys-dataset(1)%yc(:nc))**2)
           !write(0,'(a,5f20.8)')'SCAL=',crscal,sum(cryst(1)%yc),sum(cryst(2)%yc),chi
   endif
!   
   end function chisquared_profile_new

!----------------------------------------------------------------------------------------------------

   real function chisquared_factor() result(chi)
   USE variables, only: cryst
   USE crystal_phase
   real    :: scal
   integer :: nph
!
   chi = 0.0
   do nph=1,numphase(cryst)
      cryst(nph)%ref%fc = sqrt(cryst(nph)%ref%fc)
      scal = sum(cryst(nph)%ref%fo) / sum(cryst(nph)%ref%fc)
      chi = sum((cryst(nph)%ref%fo - scal*cryst(nph)%ref%fc)**2)    
   enddo
!   
   end function chisquared_factor

 !----------------------------------------------------------------------------------------------------

   real function chisquared_int_groups() result(chi)
   USE reflection_type_util
   USE variables, only: cryst
   USE crystal_phase
   real    :: scal
   integer :: nph
!
   chi = 0.0
   do nph=1,numphase(cryst)
      cryst(nph)%ref%fc = sqrt(cryst(nph)%ref%fc)              ! TOFIX
      call update_int_refg(cryst(nph)%ref,cryst(nph)%gref)     ! TOFIX
      scal = sum(cryst(nph)%gref(:)%Io) / sum(cryst(nph)%gref(:)%Ic)
      chi = chi + sum((cryst(nph)%gref(:)%Io - scal*cryst(nph)%gref(:)%Ic)**2)
   enddo
!   
   end function chisquared_int_groups  

!----------------------------------------------------------------------------------------------------

   real function chisquared_int() result(chi)
   USE variables, only: cryst
   USE crystal_phase
   real    :: scal
   integer :: nph
!
   chi = 0.0
   do nph=1,numphase(cryst)
      !intcal = cryst(1)%ref%m*cryst(1)%ref%fc 
      !intcal = cryst(1)%ref%lp*cryst(1)%ref%fc 
      cfuncsa(nph)%intcal = cryst(nph)%ref%po*cryst(nph)%ref%fc  ! fc is squared
      scal = cfuncsa(nph)%sumoss / sum(cfuncsa(nph)%intcal)
      chi = chi + sum((cfuncsa(nph)%intoss - scal*cfuncsa(nph)%intcal)**2)    
   enddo
   !     rwden = sqrt(sumintoss)/100.0
   !chi = sum(abs(intoss - scal*intcal)) 
!   chi = sum(dataset(1)%wei(:)*abs(intoss - scal*intcal)) 
!   
   end function chisquared_int
               
!----------------------------------------------------------------------------------------------------

   real function chisquared_int_new(gopti) result(chi)
   USE variables, only: cryst,dataset
   USE crystal_phase
   type(opti_info_type), intent(in) :: gopti    
   integer                          :: i,kal
   real                             :: scal
   integer                          :: nnn
   real                             :: opT
   real                             :: ffx
   real                             :: sumyco
   integer                          :: nph

   sumyco = 0.0
   opT = 1
   dataset(1)%yc(:gopti%nco) = 0.0
   kal = 1
   do nph=1,numphase(cryst)
      do i=1,size(cryst(nph)%ref)
         ffx = cryst(nph)%ref(i)%m*cryst(nph)%ref(i)%fc*opT*cryst(nph)%ref(i)%lp(1)
         do nnn=cryst(nph)%profi(i,kal)%iniz,cryst(nph)%profi(i,kal)%ifin
            dataset(1)%yc(nnn) = dataset(1)%yc(nnn) + ffx*cryst(nph)%profi(i,kal)%func(nnn-cryst(nph)%profi(i,kal)%iniz+1)
         enddo 
      enddo 
   enddo
   chi = 0.0
   do nph=1,numphase(cryst)
      do i=1,cryst(nph)%numref()
         cfuncsa(nph)%intcal(i) = 0
         do nnn=cryst(nph)%profi(i,kal)%iniz,cryst(nph)%profi(i,kal)%ifin
            cfuncsa(nph)%intcal(i) = cfuncsa(nph)%intcal(i) + dataset(1)%yc(nnn) 
         enddo 
         sumyco = sumyco + cfuncsa(nph)%intcal(i)
      enddo 
      scal = cfuncsa(nph)%sumoss/sumyco
      chi = chi + sum((cfuncsa(nph)%intoss - scal*cfuncsa(nph)%intcal)**2)    
   enddo
!   
   end function chisquared_int_new
               
!----------------------------------------------------------------------------------------------------
   subroutine check_fcalc(ref1,atmi,spg,elem,pos)
!
!  Controlla che il calcolo degli fc sia corretto comparandoli 
!  con quelli calcolati usando la routine standard fcalcan
!
   USE reflection_type_util
   USE atom_type_util
   USE spginfom
   USE elements
   USE mpi_prog
   USE General, only: neutro, is_anomalous
   type(reflection_type), dimension(:), intent(in) :: ref1
   type(atom_type), dimension(:), intent(in)       :: atmi
   type(spaceg_type), intent(in)                   :: spg
   type(element_type), dimension(:), allocatable, intent(in) :: elem
   integer, intent(in)                             :: pos
   type(reflection_type), dimension(size(ref1))    :: ref2
   real                                            :: absf,fcal
   integer                                         :: i
!   
   ref2 = ref1
   call sfcalc(ref2,atmi,spg,elem,neutro,is_anomalous)
   do i=1,size(ref1)
      if (spg%symcent == 1) then
          fcal = 2.0*sqrt(ref1(i)%fc)
      else
          fcal = sqrt(ref1(i)%fc)
      endif
      absf = abs(fcal-ref2(i)%fc)
      !if (absf > 0.005) then
      if (absf > 1.0) then
!
!         Scrivi solo il primo fc diverso      
          write(70+mpi_prog_rank+1,'(i10,2x,i4,3f10.3)')pos,i,fcal,ref2(i)%fc,absf
          exit
      endif    
   enddo           
!   
   end subroutine check_fcalc
   
!----------------------------------------------------------------------------------------------------

   real function chisquared_bond(kpr,gmat) result(chi)
!
!  Calcola chi2 sulle distanze di legame
!   
   USE CGEOM
   USE variables, only: cryst
   USE connect_mod
!
   integer, intent(in) :: kpr
   real, dimension(3,3), intent(in) :: gmat
   real                :: d, diffd
   real                :: delta = 0.02
   integer             :: i   
!
   chi = 0.0
   do i=1,numbonds(cryst(1)%bond)
      d = distanzaC(cryst(1)%at(cryst(1)%bond(i)%n1)%xc,cryst(1)%at(cryst(1)%bond(i)%n2)%xc,gmat)
      diffd = d-cryst(1)%bond(i)%dist
      if (diffd >= delta) then
          chi = chi + ((diffd-delta)/cryst(1)%bond(i)%sigma)**2
      elseif (diffd <= -delta) then
          chi = chi + ((diffd+delta)/cryst(1)%bond(i)%sigma)**2
      endif 
      if (kpr > 0) write(6,'(a,i0,a,i0,a,f0.4,a,f0.4)')        &
       ' at=',cryst(1)%bond(i)%n1,'-',cryst(1)%bond(i)%n2,' dcalc=',d,' datteso=',cryst(1)%bond(i)%dist
   enddo
!   
   end function chisquared_bond

!----------------------------------------------------------------------------------------------------

   real function chisquared_res(crystal,vres)
   USE CGEOM
   USE rrestr
   USE trig_constants
   USE crystal_phase
   USE bond_valence
   USE symmgrid
           USE arrayutil
   type(crystal_phase_t), intent(inout)  :: crystal
   integer, dimension(:), allocatable, intent(in), optional :: vres
   integer                               :: i,k
   real, dimension(3,crystal%spg%nsymop) :: xeq 
   real                                  :: dd,dmin
   real, dimension(3)                    :: ktra,ktramin
   integer                               :: n1,n2,n3
!corr   integer                               :: kloc
   real, dimension(3)                    :: x1eq, x3eq
!corr   integer                               :: kst
   real                                  :: d12, d23, val
   real, dimension(3,3)                  :: gmat
   logical                               :: kpr = .false.
   integer                               :: ks,nvres
!corr   integer, dimension(:), allocatable :: vress
!corr        real :: dminn
   logical computesym
!
   crystal%chi2resa = 0
   crystal%chi2resd = 0
   crystal%chi2resbv = 0
   gmat = crystal%cell%get_g()
!
   nvres = -1
   if (present(vres)) then
       nvres = size_array(vres)
   endif
!
   if (crystal%numresbv() > 0) then
       computesym = .true.
       if (nvres > 0) then
           computesym = any(crystal%res(vres)%code == RESBV)
       endif
       if (computesym) then
!
!          Make grid for BVS restraints
           if (.not.crystal%doc_active()) call crystal%make_symmetry_cell()
!corr           call crystal%allocate_grid(maxval(crystal%bvpar(:,:)%rmax))
           call crystal%allocate_grid(crystal%bvs_cutoff())
           call crystal%fill_grid()
       endif
   endif
!
!!!!!!!!!!!test-res
              !corr write(71,'(a)')'========================================'
              !corr if (present(vres)) then
              !corr     if (size_array(vres) > 0) then
              !corr         write(71,*)'RES=,compute only for res ',vres,computesym
              !corr     else
              !corr         write(71,*)'Compute not necessary!'
              !corr     endif
              !corr     write(71,*)'CHI2CONTR_INI=',crystal%res%contrib
              !corr endif
!!!!!!!!!!!test-res
!corr   if (present(vres)) then
!corr       do i=1, crystal%numres()
!corr          if (any(vres == i)) cycle
!corr          crystal%res(i)%code = -crystal%res(i)%code  ! temporary disable restraints
!corr       enddo
!corr   endif
!test   if (present(vres)) then
!test       call copy_array(vress,vres)
!test   else
!test       allocate(vress(crystal%numres()),source=[(i,i=1,crystal%numres())])
!test       crystal%res%contrib = 0.0
!test   endif
!test   do iv=1, size_array(vress)   !crystal%numres()
   do i=1, crystal%numres()
      if (nvres == 0) cycle
      if (nvres > 0) then
          if (all(vres /= i)) cycle  ! jump restraints non included in array vres
      endif
      if (crystal%res(i)%code < 0) cycle

      select case (crystal%res(i)%code)
         case (RESDIST)   ! distance
!
!          atomi coinvolti nel restraint distanza
           n1 = crystal%res(i)%na(1)
           n2 = crystal%res(i)%na(2)
           select case(crystal%res(i)%sym(1))
              case (0)   ! All atoms are considered
                call get_equivalent1(crystal%at(n2)%xc,crystal%spg,xeq)  
                dmin = huge(1.0)
                do k=1,crystal%spg%nsymop
                   call xdisteqs(crystal%at(n1)%xc,xeq(:,k),gmat,dd,ktra)
                   if (dd < dmin) dmin = dd
                enddo
                crystal%res(i)%val = dmin

              case (1)   ! Intramolecular restraint
                crystal%res(i)%val = distanzaC(crystal%at(n1)%xc,crystal%at(n2)%xc,gmat)

              case (2)   ! Intermolecular restraint
                call get_equivalent1(crystal%at(n2)%xc,crystal%spg,xeq)  
                call xdisteqs_intra(crystal%at(n1)%xc,xeq(:,1),gmat,dmin,ktra)
!test
                !if (all(abs((crystal%at(n1)%xc + ktra) - xeq(:,1)) <= epsilon(1.0))) then
                !    write(70,*)'DMIN_CYCLE=',dmin,abs((crystal%at(n1)%xc + ktra) - xeq(:,k))
                !endif
!test
                do k=2,crystal%spg%nsymop
                   call xdisteqs(crystal%at(n1)%xc,xeq(:,k),gmat,dd,ktra)
                   if (dd < dmin) dmin = dd
                enddo

!corr                do k=1,crystal%spg%nsymop
!corr                   call xdisteqs(crystal%at(n1)%xc,xeq(:,k),gmat,dd,ktra)
!corr                   if (k == 1) then
!corr!
!corr!                      check if cell translation is 0: ktra - (xeq(:,k)-atom(n2)%xc) == 0
!corr                       ! write(71,*)'KTRA=',ktra
!corr                       ! write(71,*)'XEQ=',xeq(:,k)
!corr                       ! write(71,*)'XC=',crystal%at(n1)%xc,all((crystal%at(n1)%xc + ktra) - xeq(:,k) <= epsilon(1.0))
!corr                       ! write(71,*)'DIFF=',(crystal%at(n1)%xc + ktra) - xeq(:,k)
!corr                       if (all(abs((crystal%at(n1)%xc + ktra) - xeq(:,k)) <= epsilon(1.0))) then
!corr                           !write(70,*)'DMIN_CYCLE=',dd,dminn,dd == dminn,(crystal%at(n1)%xc + ktra) - xeq(:,k)
!corr                           cycle  ! if true n1 is not an equivalent for translation
!corr                       endif
!corr                       !write(70,*)'DMIN=',dd,dminn,dd == dminn
!corr                   endif
!corr                   if (dd < dmin) dmin = dd
!corr                enddo
                crystal%res(i)%val = dmin

           end select 
!
!          Valuta contributo alla funzione costo
!!!!!test-res
!test-res           if (.not.present(vres)) then
!test-res               crystal%res(i)%contrib = chi2res_value(crystal%res(i))
!test-res           else
!test-res               if (any(vres == i)) then
!test-res                   crystal%res(i)%contrib = chi2res_value(crystal%res(i))
!test-res                   write(71,*)'Update cntrib for res ',i,crystal%res(i)%contrib
!test-res               endif
!test-res               write(71,*)'STD=',i,chi2res_value(crystal%res(i))
!test-res           endif
!!!!!test-res
           crystal%res(i)%contrib = chi2res_value(crystal%res(i))
           !write(70,'(i5,3(a,f0.3))')i,'chi2resd=',crystal%res(i)%contrib,' dist=',crystal%res(i)%val,' expected=',crystal%res(i)%targ
           !crystal%chi2resd =  crystal%chi2resd + chi2res_value(crystal%res(i))   !!!!!,deltad)
           if (kpr) write(6,'(i5,3(a,f0.3))')i,'chi2resd=',crystal%res(i)%contrib,  &
                    ' dist=',crystal%res(i)%val,' expected=',crystal%res(i)%targ

         case (RESANGLE)   ! angle
           n1 = crystal%res(i)%na(1)
           n2 = crystal%res(i)%na(2)
           n3 = crystal%res(i)%na(3)
           if (all(crystal%res(i)%sym == 1)) then
!
!              intramolecular angle
               crystal%res(i)%val = rtod*angleC(crystal%at(n1)%xc,crystal%at(n2)%xc,crystal%at(n3)%xc,gmat)
           else
!
!              include equivalent atoms for atom n1 and/or n2
               if (crystal%res(i)%sym(1) /= 1) then
                   call get_equivalent1(crystal%at(n1)%xc,crystal%spg,xeq)  
                   if (crystal%res(i)%sym(1) == 2) then
                       call xdisteqs_intra(crystal%at(n2)%xc,xeq(:,1),gmat,dmin,ktramin)
                       ks = 2
                   else 
                       dmin = huge(1.0)
                       ks = 1
                   endif
                   do k=ks,crystal%spg%nsymop
                      call xdisteqs(crystal%at(n2)%xc,xeq(:,k),gmat,d12,ktra)
                      if (d12 < dmin) then
                          dmin = d12
                          ktramin = ktra
                      endif
                   enddo
                   x1eq = crystal%at(n2)%xc + ktramin
                   !write(70,*)'DIST=',distanzaC(x1eq,crystal%at(n2)%xc,gmat),dmin
               else
                   x1eq = crystal%at(n1)%xc
               endif
               if (crystal%res(i)%sym(3) /= 1) then
                   call get_equivalent1(crystal%at(n3)%xc,crystal%spg,xeq)  
                   if (crystal%res(i)%sym(3) == 2) then
                       call xdisteqs_intra(crystal%at(n2)%xc,xeq(:,1),gmat,dmin,ktramin)
                       ks = 2
                   else
                       dmin = huge(1.0)
                       ks = 1
                   endif
                   do k=ks,crystal%spg%nsymop
                      call xdisteqs(crystal%at(n2)%xc,xeq(:,k),gmat,d23,ktra)
                      if (d23 < dmin) then
                          dmin = d23
                          ktramin = ktra
                      endif
                   enddo
                   x3eq = crystal%at(n2)%xc + ktramin
                   !write(70,*)'DIST=',distanzaC(x3eq,crystal%at(n2)%xc,gmat),dmin
               else
                   x3eq = crystal%at(n3)%xc
               endif
               crystal%res(i)%val = rtod*angleC(x1eq,crystal%at(n2)%xc,x3eq,gmat)
           endif
!
!          Evaluate contribution to cost function
           !crystal%chi2resa = crystal%chi2resa + chi2res_value(crystal%res(i))   !!!!,deltaa)
           crystal%res(i)%contrib = chi2res_value(crystal%res(i))
           if (kpr) write(6,'(4(a,f0.3))')    &
               'chi2a=',crystal%chi2resa,' angl=',crystal%res(i)%val,' expected=',crystal%res(i)%targ,' dim=',dmin
           !write(66,'(3(a,f0.3))')'chi2resa=',crystal%chi2resa,' angl=',crystal%res(i)%val,' expected=',crystal%res(i)%targ

         case (RESBV)   ! bond valence
           n1 = crystal%res(i)%na(1)
           call crystal%compute_bvs(n1,val)
           crystal%res(i)%val = val
           !crystal%chi2resbv = crystal%chi2resbv + chi2res_value(crystal%res(i))
           !if (present(vres)) then
           !corr if (all(vres /= i)) then
           !corr     write(71,*)'Calcolo non necessario!',(crystal%res(i)%contrib == chi2res_value(crystal%res(i)))
           !corr endif
           !endif
           crystal%res(i)%contrib = chi2res_value(crystal%res(i))
!!!!!test
           !write(71,*)'VAL(',crystal%at(n1)%glab()//')=',crystal%res(i)%val,crystal%res(i)%contrib
!!!!!test

      end select
   enddo
!test-res
!test-res   write(71,*)'CHI2TOT=',chi2tot(crystal%res),crystal%chi2resd
!test-res   do i=1,crystal%numres()
!test-res      write(71,*)'CONTR_STD_NEW:',crystal%res(i)%contrib,chi2res_value(crystal%res(i)),abs(crystal%res(i)%contrib-chi2res_value(crystal%res(i)))
!test-res      if (abs(crystal%res(i)%contrib-chi2res_value(crystal%res(i))) > 0.1) write(71,*)'WARNING'
!test-res   enddo
!test-res   write(71,'(a)')'========================================'
!test-res
   !chisquared_res = crystal%chi2resa + crystal%chi2resd + crystal%chi2resbv
   !write(70,*)'CONTR=',crystal%res%contrib
!   if (present(vres)) then
!       crystal%res(vres)%code = -crystal%res(vres)%code  ! rienable all restraints
!   endif
   chisquared_res = chi2tot(crystal%res)
   crystal%chi2res = chisquared_res
!
   end function chisquared_res
        
!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine xdisteqs_intra(xb2,xeq,gm,dist,ddelta)
!corr!
!corr!  Subroutine used to compute the distance between an atom and its equivalent
!corr!  xeq = xb2 + ddelta
!corr!  Cell translation to apply to xeq is: ddelta - (xeq - xb2)
!corr!
!corr   real, dimension(3), intent(in)   :: xb2,xeq
!corr   real, dimension(3,3), intent(in) :: gm
!corr   real, intent(out)                :: dist
!corr   real, dimension(3), intent(out)  :: ddelta
!corr   real, dimension(7,3)             :: delta
!corr   real, dimension(3)               :: dp,ds
!corr   real, dimension(7)               :: dd
!corr   real                             :: d
!corr   integer, dimension(1)            :: lmin
!corr   integer                          :: i
!corr!
!corr   do i=1,3
!corr      d=xeq(i)-xb2(i)
!corr      dp(i)=d-int(d)
!corr      if (dp(i) <= 0.0) then
!corr          ds(i)=dp(i)+1.0
!corr      else
!corr          ds(i)=dp(i)-1.0
!corr      endif
!corr   enddo
!corr!corr   delta(1,:) = dp
!corr   delta(1,:) = (/dp(1),dp(2),ds(3)/)
!corr   delta(2,:) = (/dp(1),ds(2),dp(3)/)
!corr   delta(3,:) = (/dp(1),ds(2),ds(3)/)
!corr   delta(4,:) = (/ds(1),dp(2),dp(3)/)
!corr   delta(5,:) = (/ds(1),dp(2),ds(3)/)
!corr   delta(6,:) = (/ds(1),ds(2),dp(3)/)
!corr   delta(7,:) = ds
!corr!
!corr      !write(70,*)0,'DD=',sqrt(dot_product(dp,matmul(gm,dp))),xb2 + dp - xeq
!corr   do i=1,7
!corr      dd(i) = dot_product(delta(i,:),matmul(gm,delta(i,:)))
!corr      !!write(70,*)i,'DD=',sqrt(dd(i)),xb2 + delta(i,:) - xeq
!corr   enddo
!corr   lmin = minloc(dd)
!corr   dist = sqrt(dd(lmin(1)))
!corr   ddelta = delta(lmin(1),:3)
!corr!
!corr   end subroutine xdisteqs_intra
!corr
!-------------------------------------------------------------------------

   subroutine distance_equivalent1(x1,x2,xg,spg,dmin,ks)
!
!  Find minimum distance between x1 and all equivalent of x2. ks=2 for only equivalent atoms
!
   USE spginfom
   real, dimension(3), intent(in)  :: x1,x2
   real, dimension(3,3),intent(in) :: xg
   type(spaceg_type), intent(in)   :: spg
   real, intent(out)               :: dmin
   integer, intent(in)             :: ks
   real, dimension(3)              :: xeq, ktramin,xeq1,dx
   integer                         :: k
   real                            :: d2
   integer :: k1,k2,k3
   integer, dimension(3) :: diff,ktra
!
   dmin = huge(1.0)
   do k=ks,spg%nsymop
      xeq = matmul(spg%symop(k)%rot,x2) + spg%symop(k)%trn
      diff = nint(xeq-x1)
      do k1=-1,1
          do k2=-1,1
             do k3=-1,1
                ktra = (/k1,k2,k3/) + diff
                xeq1 = xeq-ktra
                dx = xeq1-x1
                d2 = DOT_PRODUCT(dx,MATMUL(xg,dx))
                if (d2 < dmin) then
                    dmin = d2
                    ktramin = ktra
                endif
             enddo
          enddo
      enddo
   enddo
   dmin = sqrt(dmin)
!
   end subroutine distance_equivalent1

!----------------------------------------------------------------------------------------------------
!old-res
!old-res   real function chisquared_res(crystal,vres)
!old-res   USE CGEOM
!old-res   USE rrestr
!old-res   USE trig_constants
!old-res   USE crystal_phase
!old-res   USE bond_valence
!old-res   USE symmgrid
!old-res           USE arrayutil
!old-res   type(crystal_phase_t), intent(inout)  :: crystal
!old-res   integer, dimension(:), allocatable, intent(in), optional :: vres
!old-res   integer                               :: i,k
!old-res   real, dimension(3,crystal%spg%nsymop) :: xeq 
!old-res   real                                  :: dd,dmin
!old-res   real, dimension(3)                    :: ktra
!old-res   integer                               :: n1,n2,n3
!old-res   integer                               :: kloc
!old-res   real, dimension(3)                    :: x3, x1eq, x3eq
!old-res!corr   integer                               :: kst
!old-res   real                                  :: d12, d23, val
!old-res   real, dimension(3,3)                  :: gmat
!old-res   logical                               :: kpr = .false.
!old-res   integer                               :: ks
!old-res!
!old-res   crystal%chi2resa = 0
!old-res   crystal%chi2resd = 0
!old-res   crystal%chi2resbv = 0
!old-res   gmat = crystal%cell%get_g()
!old-res!
!old-res   if (crystal%numresbv() > 0) then
!old-res!
!old-res!      Make grid for BVS restraints
!old-res       if (.not.crystal%doc_active()) call crystal%make_symmetry_cell()
!old-res       call crystal%allocate_grid(maxval(crystal%bvpar(:,:)%rmax))
!old-res       call crystal%fill_grid()
!old-res   endif
!old-res!
!old-res!!!!!!!!!!!test-res
!old-res!test-res               write(71,'(a)')'========================================'
!old-res!test-res               if (present(vres)) then
!old-res!test-res                   if (size_array(vres) > 0) then
!old-res!test-res                       write(71,*)'RES=,compute only for res ',vres
!old-res!test-res                   else
!old-res!test-res                       write(71,*)'Compute not necessary!'
!old-res!test-res                   endif
!old-res!test-res                   write(71,*)'CHI2CONTR_INI=',crystal%res%contrib
!old-res!test-res               endif
!old-res!!!!!!!!!!!test-res
!old-res   do i=1, crystal%numres()
!old-res      select case (crystal%res(i)%code)
!old-res         case (RESDIST)   ! distance
!old-res!
!old-res!          atomi coinvolti nel restraint distanza
!old-res           n1 = crystal%res(i)%na(1)
!old-res           n2 = crystal%res(i)%na(2)
!old-res           select case(crystal%res(i)%sym(1))
!old-res              case (0)   ! All atoms are considered
!old-res                call get_equivalent1(crystal%at(n2)%xc,crystal%spg,xeq)  
!old-res                dmin = huge(1.0) !!!!distanzaC(atom(n2)%xc,atom(n1)%xc,gmat)
!old-res                do k=1,crystal%spg%nsymop
!old-res                   call xdisteqs(crystal%at(n1)%xc,xeq(:,k),gmat,dd,ktra)
!old-res                   if (dd < dmin) dmin = dd
!old-res                enddo
!old-res                crystal%res(i)%val = dmin
!old-res
!old-res              case (1)   ! Intramolecular restraint
!old-res                crystal%res(i)%val = distanzaC(crystal%at(n1)%xc,crystal%at(n2)%xc,gmat)
!old-res
!old-res           end select 
!old-res#if 0           
!old-res!
!old-res!          Genera equivalenti di n2
!old-res           call get_equivalent1(crystal%at(n2)%xc,crystal%spg,xeq)  
!old-res!
!old-res!          Cerca la distanza minima tra n1 e equivalenti di n2
!old-res           !if (crystal%res(i)%sym(1) == 0) then
!old-res           !    kst = 2                  ! restraint between the same atoms (n1=n2) or n1-n2 are in bond: jump operator 1
!old-res           !    if (crystal%spg%nsymop == 1) cycle  ! not possible!
!old-res           !else
!old-res           !    kst = 1
!old-res           !endif
!old-res           !call xdisteqs(crystal%at(n1)%xc,xeq(:,kst),gmat,dmin,ktra)
!old-res           !do k=kst+1,crystal%spg%nsymop
!old-res           !   call xdisteqs(crystal%at(n1)%xc,xeq(:,k),gmat,dd,ktra)
!old-res           !   if (dd < dmin) dmin = dd
!old-res           !enddo
!old-res           !crystal%res(i)%val = dmin
!old-res!!!!!new
!old-res           dmin = huge(1.0) !!!!distanzaC(atom(n2)%xc,atom(n1)%xc,gmat)
!old-res           do k=1,crystal%spg%nsymop
!old-res              call xdisteqs(crystal%at(n1)%xc,xeq(:,k),gmat,dd,ktra)
!old-res              if (crystal%res(i)%sym(1) == 0 .and. k == 1) then
!old-res!
!old-res!                 check if cell translation is 0: ktra - (xeq(:,k)-atom(n2)%xc) == 0
!old-res                  ! write(71,*)'KTRA=',ktra
!old-res                  ! write(71,*)'XEQ=',xeq(:,k)
!old-res                  ! write(71,*)'XC=',crystal%at(n1)%xc,all((crystal%at(n1)%xc + ktra) - xeq(:,k) <= epsilon(1.0))
!old-res                  ! write(71,*)'DIFF=',(crystal%at(n1)%xc + ktra) - xeq(:,k)
!old-res                  if (all((crystal%at(n1)%xc + ktra) - xeq(:,k) <= epsilon(1.0))) cycle  ! if true n1 is not an equivalent for translation
!old-res              endif
!old-res              !if ((k == 1) .or. (k > 2 .and. dd <= 3)) then
!old-res              !    crystal%res(i)%val = dd
!old-res              !    crystal%chi2resd =  crystal%chi2resd + chi2res_value(crystal%res(i))   !!!!!,deltad)
!old-res              !endif
!old-res              !write(6,'(i5,3(a,f0.3))')i,'chi2resd=',crystal%chi2resd,' dist=',dd,' expected=',crystal%res(i)%targ
!old-res              if (dd < dmin) dmin = dd
!old-res           enddo
!old-res           crystal%res(i)%val = dmin
!old-res#endif
!old-res!!!!!new
!old-res!
!old-res!          Valuta contributo alla funzione costo
!old-res!!!!!test-res
!old-res!test-res           if (.not.present(vres)) then
!old-res!test-res               crystal%res(i)%contrib = chi2res_value(crystal%res(i))
!old-res!test-res           else
!old-res!test-res               if (any(vres == i)) then
!old-res!test-res                   crystal%res(i)%contrib = chi2res_value(crystal%res(i))
!old-res!test-res                   write(71,*)'Update cntrib for res ',i,crystal%res(i)%contrib
!old-res!test-res               endif
!old-res!test-res               write(71,*)'STD=',i,chi2res_value(crystal%res(i))
!old-res!test-res           endif
!old-res!!!!!test-res
!old-res           crystal%chi2resd =  crystal%chi2resd + chi2res_value(crystal%res(i))   !!!!!,deltad)
!old-res           if (kpr) write(6,'(i5,3(a,f0.3))')i,'chi2resd=',crystal%chi2resd,' dist=',dmin,' expected=',crystal%res(i)%targ
!old-res
!old-res         case (RESANGLE)   ! angle
!old-res           n1 = crystal%res(i)%na(1)
!old-res           n2 = crystal%res(i)%na(2)
!old-res           n3 = crystal%res(i)%na(3)
!old-res           if (n1 == n3) then
!old-res               ks = 2   ! e.g., Si1-O1 ... Si1
!old-res           else
!old-res               ks = 1
!old-res           endif
!old-res!
!old-res!          generate equivalent if sym = 1
!old-res           if (crystal%res(i)%sym(1) == 1) then
!old-res               call distance_equivalent(crystal%at(n1)%xc,crystal%at(n2)%xc,gmat,crystal%spg,d12,x1eq,ks)
!old-res           else
!old-res               d12 = distanzaC(crystal%at(n1)%xc,crystal%at(n2)%xc,gmat)
!old-res               x1eq = crystal%at(n1)%xc
!old-res           endif
!old-res           if (crystal%res(i)%sym(3) == 1) then
!old-res               call distance_equivalent(crystal%at(n3)%xc,crystal%at(n2)%xc,gmat,crystal%spg,d23,x3eq,ks)
!old-res           else
!old-res               d23 = distanzaC(crystal%at(n3)%xc,crystal%at(n2)%xc,gmat)
!old-res               x3eq = crystal%at(n3)%xc
!old-res           endif
!old-res           crystal%res(i)%val = rtod*angleC(x1eq,crystal%at(n2)%xc,x3eq,gmat)
!old-res!
!old-res!          Evaluate contribution to cost function
!old-res           crystal%chi2resa = crystal%chi2resa + chi2res_value(crystal%res(i))   !!!!,deltaa)
!old-res           if (kpr) write(6,'(4(a,f0.3))')    &
!old-res               'chi2a=',crystal%chi2resa,' angl=',crystal%res(i)%val,' expected=',crystal%res(i)%targ,' dim=',dmin
!old-res           !write(66,'(3(a,f0.3))')'chi2resa=',crystal%chi2resa,' angl=',crystal%res(i)%val,' expected=',crystal%res(i)%targ
!old-res
!old-res         case (RESBV)   ! bond valence
!old-res           n1 = crystal%res(i)%na(1)
!old-res           call crystal%compute_bvs(n1,val)
!old-res           crystal%res(i)%val = val
!old-res           crystal%chi2resbv = crystal%chi2resbv + chi2res_value(crystal%res(i))
!old-res           !write(70,*)'VAL(',crystal%at(n1)%glab()//')=',crystal%res(i)%val,crystal%chi2resbv
!old-res
!old-res         case (10000)   ! angle
!old-res!
!old-res!          atomi coinvolti nel restraint angolo
!old-res           n1 = crystal%res(i)%na(1)
!old-res           n2 = crystal%res(i)%na(2)
!old-res           n3 = crystal%res(i)%na(3)
!old-res!
!old-res!          Genera equivalenti di n3
!old-res           call get_equivalent1(crystal%at(n3)%xc,crystal%spg,xeq)  !!!!!,neqa)
!old-res!
!old-res!          Cerca l'equivalente a distanza minima tra n2-n3
!old-res           call xdisteqs(crystal%at(n2)%xc,xeq(:,1),gmat,dmin,ktra)
!old-res           !call xdisteqs(xeq(:,1),atom(n2)%xc,gmat,dmin,ktra)
!old-res           kloc = 1
!old-res           do k=2,crystal%spg%nsymop
!old-res              call xdisteqs(crystal%at(n2)%xc,xeq(:,k),gmat,dd,ktra)
!old-res              !call xdisteqs(xeq(:,k),atom(n2)%xc,gmat,dd,ktra)
!old-res              if (dd < dmin) then
!old-res                  dmin = dd
!old-res                  kloc = k
!old-res              endif
!old-res           enddo
!old-res!
!old-res!         valuta angolo n1-n2-n3
!old-res          !xeq(:,kloc) = xeq(:,kloc) + ktra(:)
!old-res          x3(:) = crystal%at(n2)%xc + ktra(:)
!old-res            !write(6,*)'distanza=',distanzaC(atom(n2)%xc,x3,gmat),dmin
!old-res          !rest(i)%val = rtod*angleC(atom(n1)%xc,atom(n2)%xc,xeq(:,kloc),gmat)
!old-res          crystal%res(i)%val = rtod*angleC(crystal%at(n1)%xc,crystal%at(n2)%xc,x3,gmat)
!old-res!
!old-res!         Valuta contributo alla funzione costo
!old-res          crystal%chi2resa = crystal%chi2resa + chi2res_value(crystal%res(i))   !!!!,deltaa)
!old-res          if (kpr) write(6,'(4(a,f0.3))')    &
!old-res              'chi2a=',crystal%chi2resa,' angl=',crystal%res(i)%val,' expected=',crystal%res(i)%targ,' dim=',dmin
!old-res           !write(66,'(3(a,f0.3))')'chi2resa=',crystal%chi2resa,' angl=',rest(i)%val,' expected=',rest(i)%targ
!old-res      end select
!old-res   enddo
!old-res!test-res
!old-res!test-res   write(71,*)'CHI2TOT=',chi2tot(crystal%res),crystal%chi2resd
!old-res!test-res   do i=1,crystal%numres()
!old-res!test-res      write(71,*)'CONTR_STD_NEW:',crystal%res(i)%contrib,chi2res_value(crystal%res(i)),abs(crystal%res(i)%contrib-chi2res_value(crystal%res(i)))
!old-res!test-res      if (abs(crystal%res(i)%contrib-chi2res_value(crystal%res(i))) > 0.1) write(71,*)'WARNING'
!old-res!test-res   enddo
!old-res!test-res   write(71,'(a)')'========================================'
!old-res!test-res
!old-res   chisquared_res = crystal%chi2resa + crystal%chi2resd + crystal%chi2resbv
!old-res   crystal%chi2res = chisquared_res
!old-res!
!old-res   end function chisquared_res

!corr!----------------------------------------------------------------------------------------------------        
!corr
!corr   real function chi2res_value(rest)
!corr   type(restraint_type), intent(in) :: rest
!corr   real                             :: rplus,rless
!corr!
!corr   rplus = rest%targ + rest%sigma
!corr   rless = rest%targ - rest%sigma
!corr   if (rest%val <= rless) then
!corr       chi2res_value = rest%wei*((rest%val-rless))**2
!corr   elseif (rest%val >= rplus) then
!corr       chi2res_value = rest%wei*((rest%val-rplus))**2
!corr   else
!corr       chi2res_value = 0
!corr   endif
!corr!
!corr   end function chi2res_value
!corr
!corr!----------------------------------------------------------------------------------------------------        

   subroutine chi2res_print(crystal,kpr)
   USE crystal_phase
   USE rrestr
   type(crystal_phase_t), intent(inout) :: crystal
   integer, intent(in)                  :: kpr
   write(kpr,'(2x,a,f15.4)')    'Chi-squared (restraints contribution): ',crystal%chi2res
   if (crystal%numres(RESDIST) > 0 .and. crystal%numres(RESDIST) /= crystal%numres()) then
       write(kpr,'(2x,a,f15.4)')'                            Distances: ',crystal%chi2resd
   endif
   if (crystal%numres(RESANGLE) > 0 .and. crystal%numres(RESANGLE) /= crystal%numres()) then
       write(kpr,'(2x,a,f15.4)')'                               Angles: ',crystal%chi2resa
   endif
   if (crystal%numres(RESBV) > 0 .and. crystal%numres(RESBV) /= crystal%numres()) then
       write(kpr,'(2x,a,f15.4)')'                         Bond valence: ',crystal%chi2resbv
   endif
   end subroutine chi2res_print

!----------------------------------------------------------------------------------------------------        
   
   subroutine gopt_print_info(info,sc,unt)
!
!  Print general information
!   
   USE mpi_prog
   USE variables, only: cryst
   USE reflection_type_util
   USE crystal_phase
   USE strutil
   use simulated_anneal, only: SimAnn_Conditions_type
   type(opti_info_type), intent(in)         :: info
   type(SimAnn_Conditions_type), intent(in) :: sc                     
   integer, intent(in)                      :: unt
   integer                                  :: nph
!
   if (mpi_prog_rank /= mpi_prog_master) return
!
   write(unt,'(/1x,a,i0)')         'Number of CPU-cores:                 ',mpi_prog_numprocs
   write(unt,'(1x,a,i0)')          'Number of parameters:                ',info%npar
   write(unt,'(3x,a,i0)')            'DOF(ext):                          ',info%dofe
   write(unt,'(3x,a,i0)')            'DOF(int):                          ',info%dofi
   if (info%npb > 0) then
       write(unt,'(3x,a,i0)')        'Thermal factors:                   ',info%npb
   endif
   if (info%npshift > 0) then
       write(unt,'(3x,a,i0)')        'Shifts on atoms:                   ',info%npshift
   endif
   write(unt,'(1x,a,f0.2,a,f0.2)') 'Used data range:                     ',info%ttmin,'-',info%ttmax
   if (numphase(cryst) > 1) then
       do nph=1,numphase(cryst)
          write(unt,'(1x,a,i0,a)')'Settings for phase ',nph,' ('//trim(cryst(nph)%cr_name)//')'
          write(unt,'(1x,a,i0,a,i0)') 'Number of used reflections:          ',cryst(nph)%numref(),'/',numrefl(cryst(nph)%refc)
          call cryst(nph)%print_doc(unt)
          write(unt,*)
       enddo
   else
       write(unt,'(1x,a,i0,a,i0)') 'Number of used reflections:          ',cryst(1)%numref(),'/',numrefl(cryst(1)%refc)
       call cryst(1)%print_doc(unt)
   endif
   write(unt,'(1x,a,a)')           'Resolution:                          ',r_to_s(info%res,3)//' ang'
   write(unt,'(1x,a)')             'Cost function:                       '//trim(info%cost_function_name)
   if (cryst(1)%gcode > 0) then
       write(unt,'(1x,a,3(1x,i0))')'P.O. correction is active for plane: ',cryst(1)%hklpo(:)
   endif
   if (info%algo == SA_LOCAL .and. mpi_prog_numprocs > 1 .and. sc%cooperate) then
       write(unt,'(1x,a)')         'Algorithm:                           '//'Parallel '//algoname_long(info%algo)
   else
       write(unt,'(1x,a)')         'Algorithm:                           '//algoname_long(info%algo)
   endif
!   
   end subroutine gopt_print_info
   
!----------------------------------------------------------------------------------------------------        
   
   subroutine gopt_print_parameters(xopt,par,unt,title)
!
!  Stampa lista dei parametri affinati
!   
   USE strutil, only:centra_str
   USE atom_type_util, only:norm_torsion
   USE rrestr
   USE constraints
   USE trig_constants
   USE variables, only: cryst
   USE crystal_phase
   USE rigid_body
   use rotationmod
   real, dimension(:), intent(in)             :: xopt
   type(param_type), dimension(:), intent(in) :: par
   integer, intent(in)                        :: unt
   integer                                    :: i,j,nph
   integer                                    :: scod,icod   
   character(len=*), intent(in), optional     :: title
   character(len=3)                           :: adv
!
   if (present(title)) then
       write(unt,'(/a)')centra_str(title,80)
   else        
       write(unt,*)    
   endif        
   write(unt,'(20x,a)')'        Value        Step length          Bounds'
!
   do nph=1,numphase(cryst)
!
!     Stampa per traslazione e rotazione frammenti   
      do i=1,cryst(nph)%numfrag()
         scod = count(cryst(nph)%frag(i)%rcod > 0)   ! numero di paramtri affinati per il frammento i
         if (scod > 0) then
             write(unt,'(1x,a,i0)')'External DOF for fragment n.',i
             do j=1,6
                icod = cryst(nph)%frag(i)%rcod(j)
                if (icod > 0) then             
                    write(unt,'(3x,a,t20,f15.4,f15.4,2f13.2)')trim(par(icod)%str),xopt(icod),   &
                          par(icod)%stp,par(icod)%low,par(icod)%high
                endif
             enddo
         endif
      enddo
!
!     Stampa per rotazioni interne ai frammenti
      if (cryst(nph)%numrot() > 0) then
          scod = count(cryst(nph)%rot%rcod > 0)
          if (scod > 0) then
              write(unt,'(1x,a)')'Internal DOF'
              do i=1,cryst(nph)%numrot()
                 icod = cryst(nph)%rot(i)%rcod
                 if (icod > 0) then
                     if (cryst(nph)%rot(i)%mode > 1) then
                         adv = 'no'
                     else
                         adv = 'yes'
                     endif
                     write(unt,'(3x,a,t20,f15.4,f15.4,2f13.2)',advance=adv)trim(par(icod)%str),              &
                           torsion_value_for_rotat(cryst(nph)%rot(i),cryst(nph)%at,cryst(nph)%cell),         &
!!                          norm_torsion(rtod*twopi*xopt(icod)+cryst(nph)%rot(i)%torsval),par(icod)%stp,     &
                           par(icod)%stp,cryst(nph)%rot(i)%trbound(1),cryst(nph)%rot(i)%trbound(2)
                     if (cryst(nph)%rot(i)%mode == 2 .or. cryst(nph)%rot(i)%mode == 4) then
                         write(unt,'(2f13.2)') cryst(nph)%rot(i)%trbound(3),cryst(nph)%rot(i)%trbound(4)   
                     elseif (cryst(nph)%rot(i)%mode == 3) then
                         write(unt,'(4f13.2)') cryst(nph)%rot(i)%trbound(3),cryst(nph)%rot(i)%trbound(4), &
                                                            cryst(nph)%rot(i)%trbound(5),cryst(nph)%rot(i)%trbound(6)
                     endif
                 endif
              enddo
          endif
      endif
!
!     Stampa per shift sugli atomi
      scod = 0
      do i=1,cryst(nph)%nrigid()
         if (count(cryst(nph)%rigidb(i)%rcod > 0) > 0) then
             scod = 1
             exit
         endif 
      enddo
      if (scod > 0) then
          write(unt,'(1x,a)')'Shift on atoms'
          do i=1,cryst(nph)%nrigid()
             do j=1,6
                icod = cryst(nph)%rigidb(i)%rcod(j)
                if (icod > 0) then
                    write(unt,'(3x,a,t20,f15.4,f15.4,2f13.2)')trim(par(icod)%str),xopt(icod),   &
                    par(icod)%stp,par(icod)%low,par(icod)%high
                    write(unt,'(a,*(i4))')'PAT=',par(icod)%pat
                endif
             enddo
          enddo        
      endif
   enddo
!
   do nph=1,numphase(cryst)
!
!     Stampa per parametri termici
      if (count(cryst(nph)%at%rcod(4) > 0) > 0) then
          write(unt,'(1x,a)')'Thermal parameters'
          do j=1,cryst(nph)%natoms()
             icod = cryst(nph)%at(j)%rcod(4)
             if (icod > 0) then      
                 write(unt,'(3x,a,t20,f15.4,f15.4,2f13.2)')trim(par(icod)%str),xopt(icod),par(icod)%stp,par(icod)%low,par(icod)%high
             endif
          enddo       
      endif
      do i=1,nconstraints(cryst(nph)%sconstr)
         icod = cryst(nph)%sconstr(i)%ref
         if (icod > 0) then
             select case (cryst(nph)%sconstr(i)%code) 
                case (CONSTR_B)
                  write(unt,'(3x,a,t20,f15.4,f15.4,2f13.2)')trim(par(icod)%str),xopt(icod),  &
                        par(icod)%stp,par(icod)%low,par(icod)%high
             end select
         endif
      enddo
!
!     Orientazione Preferenziale
      if (cryst(nph)%gcode > 0) then
          write(unt,'(1x,a)')'Preferred Orientation'
          icod = cryst(nph)%gcode
          write(unt,'(3x,a,t20,f15.4,f15.4,2f13.2)')trim(par(icod)%str),cryst(nph)%gpar,par(icod)%stp,par(icod)%low,par(icod)%high
      endif
   enddo
!   
   end subroutine gopt_print_parameters
      
!----------------------------------------------------------------------------------------------------        

   subroutine simann(gopti,sc,bb_cond,extrac)
!
!  Routine principale: esegue multirun di SA, stampa parametri iniziali e finale, gestisce la lista dei modelli
!   
   USE simulated_anneal
   USE variables, only:cryst
   USE General, only:fname,OUT_FILE,neutro,is_anomalous
   USE compare_struct
   USE atom_type_util, only: print_atoms, calcola_occ, make_symmetry_table
   USE RefinecomRef, only:rinfo
   USE strutil, only:s_s_delete, f_to_c
   USE Molcom
   USE bb_bc
   USE rand_mod
   USE messagemod
   USE model_util
   USE anti_bump, only: antibump_function, print_antibump
   USE rrestr, only: nrestraints,print_restraints,print_restraints_all,RESBV
   USE Conteggi, only: npunti
   USE connect_mod
   USE fragmentmod, only: regroup_fragment, connect_fragment
   USE constraints, only: print_constraints
   USE timer
   USE reflection_type_util
   USE mpi_prog
   USE cryutil
   USE crystal_phase
   USE rigid_body
   USE contacts
   USE plot2
   USE enable_amb, only: STOP_EVENT
   USE interop_cpp_for, only: dosingleiteration
#ifdef MPI
   USE mpi
#endif
   type(opti_info_type), intent(inout)         :: gopti    
   type(SimAnn_Conditions_type), intent(inout) :: sc
   type(bb_bc_condition), intent(in)           :: bb_cond                     
   logical, intent(in)                         :: extrac
   real, dimension(gopti%npar)                 :: xopt
   real                                        :: fopt
   real                                        :: tfin
   integer                                     :: nfcnev
   integer                                     :: nacc, nobds 
   integer                                     :: i,j
   type(param_type), dimension(size(gpar))     :: gparini
   type(atom_type), dimension(:), allocatable  :: atbest
   real, dimension(NINFOMODEL)                 :: infomodel 
   real                                        :: timerun
   integer                                     :: kevent
   integer                                     :: nsol      
   real                                        :: rmsnsol
   integer, dimension(0:6)                     :: vcolor 
!   integer                                     :: pcol
   type(atom_type), dimension(:), allocatable  :: atmr
   logical                                     :: is_profile
   integer                                     :: nplotr
   real                                        :: rappo
   real                                        :: resval
!corr   integer                                     :: xdim,ydim
   type(bb_bc_condition)                       :: bcond
   type(bond_type), dimension(:), allocatable  :: legm
   real                                        :: frun_min, frun_max, frun_ave
   integer                                     :: sum_eval
   integer                                     :: neval
   integer                                     :: nsol35,natfound
   integer, dimension(3)                       :: ncont
   real                                        :: rmsd
   real(DP)                                    :: timetot
   integer                                     :: nsamodel,nrun,nph,numrefu,numreftot
   logical                                     :: restart_par
#ifdef MPI
   integer                                     :: myrank,mpi_err
   real, dimension(:,:), allocatable           :: mpi_sol,mpi_rec
#endif
!
   gparini = gpar
   if (jscreen > 0) then
       if(npunti == 0) call s_s_delete(gopti%popt,'P')
   else
      gopti%popt = 'SE'
   endif
   if (gopti%algo == BBBC) then
       call set_bb_bc_condition(gopti,bcond,sc)
       numcyc_calc = sc%num_temps*sc%ns*sc%nt*gopti%npar*bb_cond%niter*bb_cond%nlocmax  ! calculated for progress bar
   else
       numcyc_calc = nint(sc%t_ini*sc%ns*sc%nt*gopti%npar)  ! for progress bar
   endif
!
   ginfo = gopti
!
!  Stampa condizioni iniziali
   !if (mpi_prog_rank == mpi_prog_master) then
   if (is_go_print()) then
       if (gopti%algo == BBBC) then
           call bb_bc_print_conditions(bb_cond,go_unit)
           call sa_print_conditions(sc,go_unit,'Local search conditions')
       else
           call sa_print_conditions(sc,go_unit,'Simulated Annealing conditions')
       endif
       write(go_unit,'(1x,a,f0.3)')     'Initial cost function value:  ',tfunk(gpar%val,0)
       numrefu = 0
       numreftot = 0
       do nph=1,numphase(cryst)
          numrefu = numrefu + cryst(nph)%numref()
          numreftot = numreftot + numrefl(cryst(nph)%refc)
          if (cryst(nph)%numresab() > 0)   &
              write(go_unit,'(1x,a,f0.3)') 'Anti-bump cost function:      ',  &
!corr              cryst(nph)%abset%weight*antibump_function(cryst(nph)%abres,cryst(nph)%at,   &
              antibump_function(cryst(nph)%abres,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell%get_g(),.false.)
          if (cryst(nph)%numres() > 0) then
              write(go_unit,'(1x,a,f0.3)') 'Restraints cost function:     ',chisquared_res(cryst(nph))
              call print_restraints_all(go_unit,cryst(nph)%res,cryst(nph)%at,cryst(nph)%cell,.true.)
          endif
          call print_antibump(go_unit,cryst(nph)%abres,cryst(nph)%at,cryst(nph)%cell)
          call print_constraints(cryst(nph)%sconstr,cryst(nph)%at,go_unit,.true.,.false.)
          call print_rigidb(cryst(nph)%rigidb,cryst(nph)%at,go_unit)
       enddo
       call gopt_print_parameters(gpar%val,gpar,go_unit,'Starting refined parameters')
!      
       if (print_tb) then
           write(go_unit,'(4(a,i0),a,f0.2)')STRTAG//' DOFext=',gopti%dofe,' DOFint=',gopti%dofi,   &
                                       ' NREF=',numrefu,'/',numreftot,' RES=',gopti%res
           if (any(cryst%is_known())) then
               write(go_unit,'(a)')STRTAG//' NRUN    found     RMSD       CF       n.moves          time'   
           else
               write(go_unit,'(a)')STRTAG//' NRUN      CF       n.moves          time'   
           endif
       endif
   endif
!
!  Operazione preliminari per grafico della CF
   is_profile = is_profile_function(gopti%cftype)
   if (index(gopti%popt,'V') > 0) then  
       vcolor = (/color_BROWN,color_LIGHTBLUE,color_LIGHTRED,color_LIGHTGREEN,     &
                  color_LIGHTCYAN,color_LIGHTMAGENTA,color_YELLOW/)
       if (is_profile) then
           nplotr = gopti%nrun + 1
       else
           nplotr = gopti%nrun
       endif
       !call init_plot2(c_char_'moves'//c_null_char,c_char_'Rwp'//c_null_char);
       call init_plot2(c_char_'moves'//c_null_char,f_to_c(COSTF_NAME_SHORT(gopti%cftype)));
   endif
!   
!  gestione messaggi
   if (.not.cryst(1)%is_known()) call clear_messages(8) ! se la struttura e' incognita la riga 8 non viene usata  
   call write_message(trim(algoname_long(gopti%algo)), pos=1)
!
   gopti%fmin = huge(1.0)
   gopti%rmin = 1
   neval = 0
   sum_eval = 0
   restart_par = .true.   ! set false to use the parameters of the previous run
#ifdef MPI
   myrank = mpi_prog_rank + 1
   if (sc%cooperate) then
       nrun = gopti%nrun
   else
       nrun = mpi_prog_ntrials(gopti%nrun)
       gopti%nrun = max(gopti%nrun,mpi_prog_numprocs)       ! increase total runs if more cpus are available
   endif
   if (mpi_prog_rank == 0 .and. mpi_prog_numprocs > 1) then
       write(0,'(2x,a,i0,a)')'Direct space method will run on ',mpi_prog_numprocs,' cpu-cores ...'
   endif
   allocate(mpi_rec(gopti%npar+2,gopti%nrun),source=0.0)
   allocate(mpi_sol(gopti%npar+2,nrun),source=0.0)   ! save solutions for each cpu 
#else
   nrun = gopti%nrun
#endif
   do nph=1,numphase(cryst)
      call new_models(cryst(nph)%model,nrun)
   enddo
   nsamodel = 0
   loop_run: do i=1,nrun          ! inizio loop sui run di SA
      call timesa(SATIME_RUN)%tic()
      !if (mpi_prog_rank == mpi_prog_master) then
      if (is_go_print()) then
          write(go_unit,'(/2x,80("=")/,25x,"** ",a," Run n.",i0," **")')trim(algoname_long(gopti%algo)),i
      endif
      if (restart_par .or. i == 1) then
          gpar(:)%val = gparini(:)%val
          gpar(:)%stp = gparini(:)%stp
          gopti%nrunc = i         ! memorizza a che run sono
      endif
!
!     Initialize random number generator
      if (gopti%seed == 0) then
          call Init_Ran(gopti%seed,20)  ! seed from system clock
      else
#ifdef MPI
          call Init_Ran(gopti%seed+gopti%nrun*mpi_prog_rank+i,20)
#else
          call Init_Ran(gopti%seed+i-1,20)
#endif
      endif
!
!     Applica ottimizzazione dei parametri con SA      
      if (gopti%algo == SA_LOCAL) then
          call sa(gpar, sc, gopti, tfin, xopt, fopt, nacc, nfcnev, nobds, kevent, tfunk,safunc1,-1.0)   
          !call sa_adp(gpar, sc, gopti, tfin, xopt, fopt, nacc, nfcnev, nobds, kevent, tfunk,safunc1,-1.0)   
      elseif (gopti%algo == SA_GLOBAL) then
          !call sa_glob(gpar, sc, gopti, tfin,fopt, nfcnev, kevent,safunc,safuncg)
          call sa_glob_old(gpar, sc, gopti, tfin,fopt, nfcnev, kevent,safunc,safunc1)
      elseif (gopti%algo == HARMONYS) then
          !call hm_search(gpar,fopt,tfunk,safunc1,sc)
      elseif (gopti%algo == BBBC) then
          call bb_bc_algo(gpar,bb_cond,gopti,fopt,tfunk,safunc1,sc,nfcnev,kevent)
      endif
      call DoSingleIteration()
      !write(mpi_prog_rank+70,*)'END OF RUN:',i
!      
!     Stampa risultati alla fine del run
      !if (mpi_prog_rank == mpi_prog_master) then
      if (is_go_print()) then
          write(go_unit, '(/, 25x,"** Results at the end of ",a," Run n.",i0," **")')trim(algoname_short(gopti%algo)),i
          if (gopti%algo /= BBBC) call sa_print_final(fopt,nfcnev,nacc,nobds,tfin,6)      
          call gopt_print_parameters(gpar%val,gpar,go_unit)      
          do nph=1,numphase(cryst)
             call print_atoms(cryst(nph)%at,write_occ=.true.,kpr=go_unit)
             if(cryst(nph)%numres() > 0) then
                call print_restraints_all(go_unit,cryst(nph)%res,cryst(nph)%at,cryst(nph)%cell,.false.)
                call chi2res_print(cryst(nph),go_unit)
             endif
             if(cryst(nph)%numresab() > 0)then
                resval = antibump_function(cryst(nph)%abres,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell%get_g(),.false.)
                call print_antibump(go_unit,cryst(nph)%abres,cryst(nph)%at,cryst(nph)%cell)
                write(go_unit,'(a,f10.3)')' A-bumpCF:',resval
             endif
             if (cryst(nph)%is_known()) then
                write(go_unit,*)
                call cryst(nph)%sfactor(neutro,is_anomalous)
                call to_coord_cry(cryst(nph),go_unit,1)
             endif
          enddo
      endif
!
!     Calcola tempo per ogni run      
      call timesa(SATIME_RUN)%toc()
      timerun = timesa(SATIME_RUN)%get_interval()
#ifdef MPI
!
!     save solutions
      mpi_sol(:gopti%npar,i) = gpar%val
      mpi_sol(gopti%npar+1,i) = timerun
      mpi_sol(gopti%npar+2,i) = nfcnev
#else
!     Calcola rapporto Rwp(calc)/Rwp(Le Bail)
      if (npunti > 0 .and. extrac) then
          if (is_profile) then
              rappo = fopt/gopti%rwp_leb
          else
              call sa_profile(square=.false.)
              rappo = rinfo%rwp/gopti%rwp_leb
          endif
      else
          rappo = 0  ! da migliorare
      endif
!
!     Ragruppa e calcola occupanza
      nsamodel = nsamodel + 1    !!!!!!FIX THIS FOR MULTI
      do nph=1,numphase(cryst)
         call cryst(nph)%get_atoms(atmr)
         if (cryst(nph)%numfrag() > 1) then      ! prima raggruppa i frammenti 
             call regroup_fragment(cryst(nph)%frag,atmr,cryst(nph)%cell,cryst(nph)%spg)
         endif
         if (cryst(nph)%doc_active()) then
             call calcola_occ(atmr,cryst(nph)%spg,cryst(nph)%cell,dist=DISTOCC)  
             call search_overlapped_atoms(atmr,cryst(nph)%cell%get_g(),cryst(nph)%spg)
             call search_overlapped_atoms_print(atmr,cryst(nph)%cell%get_g(),cryst(nph)%spg)
         else
             call calcola_occ(atmr,cryst(nph)%spg,cryst(nph)%cell)  ! dist=default (0.6)
         endif
         call bond_distance_update(atmr,cryst(nph)%bond,cryst(nph)%cell%get_g())  ! update bond for distance change
!
!        Aggiungi modello alla lista       
         !infomodel = (/fopt,real(cryst(nph)%natfound),cryst(nph)%rmsd,timerun,real(i),rappo,cryst(nph)%rmsd/)
         infomodel = [fopt,real(cryst(nph)%natfound),cryst(nph)%rmsd,timerun,real(i),rappo]
         call model_set(cryst(nph)%model(nsamodel),atmr,fom=infomodel)
!
!        esporta i modelli ordinati per fom
         call model_sort(cryst(nph)%model(1:nsamodel),1)
         call overlay_models_list(nph,nsamodel)
         if (index(gopti%popt,'E') > 0) then
             do j=1,min(gopti%nexport,nsamodel)
                call export_model(j,nph,gopti%algo)
             enddo
         endif
      enddo
#endif
!
!     Aggiorna info sul minimo su tutti i run
      if (fopt < gopti%fmin) then
          gopti%fmin = fopt
          gopti%rmin = i
      endif
#ifndef MPI
!      
!     Write per tabella test
      if(print_tb)then
         call print_model_test(i,fopt,nfcnev,timerun,go_unit)
      endif
#endif
      if (kevent == STOP_EVENT) exit loop_run  ! evento stop dell'utente
!
!     recompute the number of evaluation using the previous run
      neval = neval + 1
      sum_eval = sum_eval + nfcnev
      numcyc_calc = sum_eval / neval
   enddo loop_run                    ! fine loop sui run di SA
!
   do nph=1,numphase(cryst)
      call resize_models(cryst(nph)%model,nsamodel)
   enddo
#ifdef MPI
!
   if (sc%cooperate) then
       if (mpi_prog_rank  == mpi_prog_master) then
           write(go_unit,*)'SIZE=',size(mpi_rec,1),size(mpi_rec,2),size(mpi_sol,1),size(mpi_sol,2)
           mpi_rec(:,:) = mpi_sol(:,:)
       endif
   else
       call MPI_Barrier(MPI_COMM_WORLD,mpi_err)
       call model_communicate(nrun,mpi_sol,mpi_rec)
       call MPI_Barrier(MPI_COMM_WORLD,mpi_err)
   endif
   if (mpi_prog_rank  == mpi_prog_master) then
       do nph=1,numphase(cryst)
          call new_models(cryst(nph)%model,gopti%nrun)
       enddo
       nsamodel = 0
       do i=1,gopti%nrun
          fopt = tfunk(mpi_rec(:gopti%npar,i),0)   ! compute CF and fill atm
          timerun = mpi_rec(gopti%npar+1,i)
          nfcnev = mpi_rec(gopti%npar+2,i)
          write(go_unit,'(a,i0,a,f10.3,a,*(f10.3))') 'Run ',i,' CF:',fopt,' Par: ',mpi_rec(:gopti%npar,i)
          if (npunti > 0 .and. extrac) then
              if (is_profile) then
                  rappo = fopt/gopti%rwp_leb
              else
                  call sa_profile(square=.false.)
                  rappo = rinfo%rwp/gopti%rwp_leb
              endif
          else
              rappo = 0  ! da migliorare
          endif
          nsamodel = nsamodel + 1
          do nph=1,numphase(cryst)
             if (cryst(nph)%numfrag() > 1) then      ! prima raggruppa i frammenti 
                 call regroup_fragment(cryst(nph)%frag,cryst(nph)%at,cryst(nph)%cell,cryst(nph)%spg)
             endif
             if (cryst(nph)%doc_active()) then
                 call calcola_occ(cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell,dist=DISTOCC)  
                 call search_overlapped_atoms(cryst(nph)%at,cryst(nph)%cell%get_g(),cryst(nph)%spg)
             else
                 call calcola_occ(cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell)  ! dist=default (0.6)
             endif

             call print_atoms(cryst(nph)%at,title='Structure solution',write_occ=.true.,kpr=go_unit)
             if (cryst(nph)%doc_active()) then
                 call search_overlapped_atoms_print(cryst(nph)%at,cryst(nph)%cell%get_g(),cryst(nph)%spg)
             endif
             if (cryst(nph)%is_known()) then
                 write(go_unit,*)
                 call cryst(nph)%sfactor(neutro,is_anomalous)
                 call to_coord_cry(cryst(nph),go_unit,1)
             else
                 cryst(nph)%ref%fc = sqrt(cryst(nph)%ref%fc) ! rimetti fo in ref%fc
             endif
             infomodel = [fopt,real(cryst(nph)%natfound),cryst(nph)%rmsd,timerun,real(i),rappo]
             call model_set(cryst(nph)%model(i),cryst(nph)%at,fom=infomodel)
          enddo
!      
!         Write per tabella test
          if(print_tb)then
             call print_model_test(i,fopt,nfcnev,timerun,go_unit)
          endif
       enddo
!
!      Ragruppa e calcola occupanza
!!!!!!!!!!!!CHECK THIS PART
       !test do j=1,nsamodel
       !test    call model_get(cryst(1)%model(j),atmr)
       !test    if (cryst(1)%numfrag() > 1) then      ! prima raggruppa i frammenti 
       !test        call regroup_fragment(cryst(1)%frag,atmr,cryst(1)%cell,cryst(1)%spg)
       !test    endif
       !test    if (cryst(1)%doc_active()) then
       !test        call calcola_occ(atmr,cryst(1)%spg,cryst(1)%cell,dist=DISTOCC)  
       !test        call search_overlapped_atoms(atmr,cryst(1)%cell%get_g(),cryst(1)%spg)
       !test        call search_overlapped_atoms_print(atmr,cryst(1)%cell%get_g(),cryst(1)%spg)
       !test    else
       !test        call calcola_occ(atmr,cryst(1)%spg,cryst(1)%cell)  ! dist=default (0.6)
       !test    endif
       !test    call bond_distance_update(atmr,cryst(1)%bond,cryst(1)%cell%get_g())  ! update bond for distance change
       !test    call model_set(cryst(1)%model(j),atmr)
       !test enddo
!
!      Overlay and export sorted model
       do nph=1,numphase(cryst)
          call model_sort(cryst(nph)%model,1)
          call overlay_models_list(nph,cryst(nph)%nmodels())
          if (index(gopti%popt,'E') > 0) then
              do j=1,min(gopti%nexport,nsamodel)
                 call export_model(j,nph,gopti%algo)
              enddo
          endif
       enddo
   else
       return
   endif
#endif
!
!  Ordina per CF e stampa la lista
   call print_listsa(go_unit,gopti%nexport)
!
!  Associa connettivita' ai modelli
   do nph=1,numphase(cryst)
      if (cryst(nph)%numfrag() > 1) then  ! riconnetti i frammenti fra loro
          do i=1,cryst(nph)%nmodels()
             call model_get(cryst(nph)%model(i),atmr)
             call copy_bonds(legm,cryst(nph)%bond)
             call connect_fragment(cryst(nph)%frag,atmr,legm,cryst(nph)%cell,.false.)
             call model_set(cryst(nph)%model(i),leg=legm)
          enddo
      else
          do i=1,cryst(nph)%nmodels()
             call model_set(cryst(nph)%model(i),leg=cryst(nph)%bond)
          enddo
      endif
!!!!!!!! ANALYZE CONTACTS
      do i=1,cryst(nph)%nmodels()
         call analyze_contacts(cryst(nph)%model(i)%at,cryst(nph)%model(i)%leg,  &
                               cryst(nph)%cell,cryst(nph)%spg,cutoff=0.0,nc=ncont,kpr=go_unit)
         !if (sum(ncont) > 0)  &
         if (i == 1) then
             write(go_unit,'(a,i2,"(",i2,")",3i4,f8.2,1x,a)')  &
             '@@Contacts for model n. ', i,int(cryst(nph)%model(i)%info(5)),ncont,  &
             cryst(nph)%model(i)%info(1),trim(cryst(nph)%cr_name)
         else
             write(go_unit,'(a,i2,"(",i2,")",3i4)')'@@Contacts for model n. ', i,int(cryst(nph)%model(i)%info(5)),ncont
         endif
      enddo
!
!     Select the first/best model
      call model_get(cryst(nph)%model(1),atbest,fom=infomodel)
      fopt = infomodel(1)
      natfound = nint(infomodel(2))
      rmsd = infomodel(3)
      timerun = infomodel(4)   
      rappo = infomodel(6)
!      
      cryst(nph)%at(:) = atbest(:)         ! update coordinates
      call bond_distance_update(cryst(nph)%at,cryst(nph)%bond,cryst(nph)%cell%get_g())  ! update bond for distance change
!   
      timetot = timesa(SATIME_TOT)%get_time()
      if(print_tb)then
         nsol = 0
         rmsnsol = 0
         do i=1,cryst(nph)%nmodels()
            if (cryst(nph)%model(i)%info(2) == cryst(nph)%natpub()) then
                nsol = nsol + 1
                rmsnsol = rmsnsol + cryst(nph)%model(i)%info(3)
            endif
         enddo
         frun_min = minval(cryst(nph)%model(:)%info(1))
         frun_max = maxval(cryst(nph)%model(:)%info(1))
         frun_ave = sum(cryst(nph)%model(:)%info(1)) / cryst(nph)%nmodels()
         if (cryst(nph)%is_known()) then
             nsol35 =  count(cryst(nph)%model(:)%info(3) <= 0.35)   ! number of solutions with rmsd <= 0.08
             write(go_unit,"(a,'best ',i6,'/',i0,f10.4,f12.4,1x,a,' NSOL=',i0,' RMSDm=',f0.3)")   &
             STRTAG,natfound,cryst(nph)%natpub(),rmsd,                    &
             fopt,secs_to_string(timetot),nsol,rmsnsol/nsol 
             write(go_unit,"(a,'info: ','CF=',f0.3,'-',f0.3,' CF(ave)=',f0.3,' Nsol_0.35=',i0,' Nsol_0.35 per min=',f0.3)")  &
             STRTAG,frun_min,frun_max,frun_ave,nsol35,60*nsol35/timetot
         else
             write(go_unit,"(a,'best ',f12.4,1x,a)")STRTAG,fopt,secs_to_string(timetot)
             write(go_unit,"(a,'info: ','CF=',f0.3,'-',f0.3,' CF(ave)=',f0.3)")  &
             STRTAG,frun_min,frun_max,frun_ave
         endif
      endif
      write(go_unit,'(/a,i0,1x,a)')' The structure of run n.',int(infomodel(5)),' was selected:'
      call print_atoms(cryst(nph)%at,title='Structure solution',write_occ=.true.,kpr=go_unit)
      if (cryst(nph)%numres() > 0) then
          resval = chisquared_res(cryst(nph))
          call print_restraints_all(go_unit,cryst(nph)%res,cryst(nph)%at,cryst(nph)%cell,.false.)
          call chi2res_print(cryst(nph),go_unit)
      endif
      if(cryst(nph)%numresab() > 0)then
         resval = antibump_function(cryst(nph)%abres,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell%get_g(),.false.)
         call print_antibump(go_unit,cryst(nph)%abres,cryst(nph)%at,cryst(nph)%cell)
         write(go_unit,'(a,f10.3)')' A-bumpCF:',resval
      endif
!
      call write_message('Best structure found at run:',inum=int(infomodel(5)),pos=2)
      call write_message('Saved in file:'//trim(cryst(nph)%cr_name)//'_best1.cif',pos=3)
      call write_message('CF:',rnum=fopt,pos=4)
      if (is_profile)call write_message('Ratio:',rnum=rappo,pos=0)
      call write_message('Time:'//secs_to_string(timetot),pos=0)
      if (cryst(nph)%doc_active()) &
          call search_overlapped_atoms_print(cryst(nph)%at,cryst(nph)%cell%get_g(),cryst(nph)%spg)
!
      if (cryst(nph)%is_known()) then
          call cryst(nph)%sfactor(neutro,is_anomalous)
          call to_coord_cry(cryst(nph),go_unit,0)
          call write_message_found1(0,cryst(nph)%natfound,cryst(nph)%natpub(),cryst(nph)%rmsd)
      endif
!qt      call write_message("Click on 'Solutions' to explore the other structures",pos=0)
      call PrintMess("Click on 'Solutions' to explore the other structures")
!
!     Write per tabella riepilogativa       
      if (print_tb) &
      write(go_unit,100)  &
            '"'//trim(fname(OUT_FILE)(scan(fname(OUT_FILE),'/\',back=.true.)+1:))//'",', &
            gopti%dofe,gopti%dofi,                                           &
            '"',gopti%ttmin,gopti%ttmax,'"',cryst(nph)%numref(),gopti%res,            &
            sc%nt,'"',nsol,gopti%nrun,'"',                                   &
            '"',cryst(nph)%natfound,cryst(nph)%natpub(),'"',cryst(nph)%dmean,cryst(nph)%rmsd,   &
            fopt,frun_min,frun_max,frun_ave,rappo, ',"'//trim(secs_to_string(timetot))//'"'
 100  format('TB2',a,i0,',',i0,',',a,f0.2,'-',f0.2,a,',',i0,',',f0.2,',',  &
      i0,',',a, i0,'/',i0,a,',',a,i0,'/',i0,a,7(',',f0.3),a)
   enddo
!
   flush(go_unit)
!   
   end subroutine simann

!----------------------------------------------------------------------------------------------------
#ifdef MPI
   subroutine model_communicate(nrun,mpi_sol,mpi_rec)
!
!  Send all models mpi_sol to root process (cpu 0) in mpi_rec
!
   USE mpi
   USE mpi_prog
   implicit none
   integer, intent(in)               :: nrun
   real, dimension(:,:), intent(in)  :: mpi_sol
   real, dimension(:,:), intent(out) :: mpi_rec
   integer                           :: mpi_err
   integer                           :: tag
   integer                           :: i,j
   integer                           :: status(MPI_STATUS_SIZE)
   integer                           :: nruntot
   integer                           :: nrun_proc
!
   if (mpi_prog_rank > 0) then     ! is a slave process
!
!      mpi_prog_rank send number of models to master process
       tag = 1
       call MPI_Ssend(nrun, 1, MPI_INTEGER, 0, tag, MPI_COMM_WORLD,mpi_err)   
!
!      mpi_prog_rank send models to master process
       tag = 2
       call MPI_Ssend(mpi_sol, size(mpi_sol), MPI_REAL, 0, tag, MPI_COMM_WORLD,mpi_err) 
       !write(70+mpi_prog_rank,'(a,i0,a,i0,a)')'Cpu ',mpi_prog_rank,' sends ',nrun,' solutions to cpu 1'
       !do j=1,nrun
       !   write(70+mpi_prog_rank,'(a,i0,a,*(f10.3))') 'Run ',j,' Par: ',mpi_sol(:,j)
       !enddo
   else                            ! master process
!
!      Add model of root to array mpi_rec
       mpi_rec(:,:nrun) = mpi_sol(:,:nrun)
       nruntot = nrun
!
!      Now add models of other cpu
       do i=1,mpi_prog_numprocs-1
          tag = 1
          call MPI_Recv(nrun_proc, 1, MPI_INTEGER, i, tag, MPI_COMM_WORLD,status,mpi_err)
          tag = 2
          call MPI_Recv(mpi_sol, size(mpi_sol), MPI_REAL, i, tag, MPI_COMM_WORLD,status,mpi_err)
          write(6,'(a,i0,a,i0,a,i0)')'Cpu ',mpi_prog_rank,' recieve ',nrun_proc,' solutions from cpu ',i
          do j=1,nrun_proc
             write(6,'(a,i0,a,*(f10.3))') 'Run ',j,' Par: ',mpi_sol(:,j)
          enddo
          do j=1,nrun_proc
             nruntot = nruntot + 1
             mpi_rec(:,nruntot) = mpi_sol(:,j)
          enddo
       enddo
   endif
!
   end subroutine model_communicate
#endif
!----------------------------------------------------------------------------------------------------

   subroutine print_model_test(irun,fopt,nfcnev,timerun,kpr)
   use variables, only: cryst
   use crystal_phase
   use timer
   integer, intent(in) :: irun
   real, intent(in)    :: fopt
   integer, intent(in) :: nfcnev
   real, intent(in)    :: timerun
   integer, intent(in) :: kpr
   integer             :: nph

   if (any(cryst%is_known())) then
       do nph=1,numphase(cryst)
          if (cryst(nph)%is_known()) then
              write(kpr,"(a,i5,i6,'/',i0,f10.4,f12.4,1x,i15,1x,a15)")   &
              STRTAG,irun,cryst(nph)%natfound,cryst(nph)%natpub(),   &
              cryst(nph)%rmsd,fopt,nfcnev,secs_to_string(timerun,.true.)
          else
              write(kpr,"(a,i5,f12.4,1x,i15,1x,a15)")   &
              STRTAG,irun,fopt,nfcnev,secs_to_string(timerun,.true.)
          endif
       enddo
   else
       write(kpr,"(a,i5,f12.4,1x,i15,1x,a15)")   &
       STRTAG,irun,fopt,nfcnev,secs_to_string(timerun,.true.)
   endif

   end subroutine print_model_test

!----------------------------------------------------------------------------------------------------

   subroutine set_bb_bc_condition(gopti,cond,sc)
   USE bb_bc
   USE variables, only: cryst
   USE crystal_phase
   use simulated_anneal, only: SimAnn_Conditions_type
   type(opti_info_type), intent(in)                      :: gopti    
   type(bb_bc_condition), intent(out)                    :: cond
   type(SimAnn_Conditions_type), intent(inout), optional :: sc
   integer                                               :: nph,ntotfrag
!
   if (cond%auto) then
!corr       select case(cryst(1)%numrot())
       select case(gopti%nprot)
          case (0)
            cond%niter = 5
          case (1:4)
            cond%niter = 20
          case (5:7)
            cond%niter = 30
          case (8:9)
            cond%niter = 40
          case (10:)
            cond%niter = 80
      
       end select
       ntotfrag = 0
       do nph=1,numphase(cryst)
          ntotfrag = ntotfrag + cryst(nph)%numfrag()
       enddo
       !corr cond%niter = cond%niter + nint(0.5*(cryst(1)%numfrag() - 1))*cond%niter
       cond%niter = cond%niter + nint(0.5*(ntotfrag - 1))*cond%niter
   endif
!
!  condizioni per SA all'interno del BB-BC
   if (present(sc)) then
       sc%num_temps = 5                 
       sc%nt = sc%nt / 10
       sc%temp_fact = sc%temp_fact / 4 
       sc%t_ini = 0.40*sc%t_ini 
       sc%eps = 10                    
       sc%ns = sc%ns
   endif
!
   end subroutine set_bb_bc_condition

!----------------------------------------------------------------------------------------------------

   subroutine safunc(xpar,h,hpar)
   implicit none
   real , intent(in)   :: xpar(:)
   real , intent(out)  :: h
   integer, intent(in) :: hpar
!   
   h = tfunk(xpar,hpar)
!
   end subroutine safunc

!----------------------------------------------------------------------------------------------------

   subroutine safunc_temp(xpar,h,hpar,gopti)
   implicit none
   real , intent(in)   :: xpar(:)
   real , intent(out)  :: h
   integer, intent(in) :: hpar
   type(opti_info_type), intent(in) :: gopti    
!
   call xpar_to_crystals(xpar)
   h = chisquared_xdata(hpar,gopti) !!!!,cryst(1)%at,cryst(1)%cell,cryst(1)%ref)
!
   end subroutine safunc_temp

!----------------------------------------------------------------------------------------------------        

   real function safuncg(x)
   real, dimension(:), intent(in) :: x
   safuncg = tfunk(x,0)
   end function safuncg

!----------------------------------------------------------------------------------------------------        
  
   subroutine safunc1(xp,fval,grad,icod)
   USE lsqalgo
!
!  Funzione che per  icod = 1  valuta la funzione costo
!                    icod = 2  valuta il gradiente
   real, dimension(:), intent(in)  :: xp
   real, intent(out)               :: fval
   real, dimension(:), intent(out), optional :: grad
   integer, intent(in)             :: icod
!
   select case(icod)
      case (1)
        fval = tfunk(xp,0)

      case (2)
        call calc_gradient(grad,xp,0.0,safuncg,icod=2)

   end select

!   CONTAINS
! 
!   real function funcgrad(x)
!   real, dimension(:), intent(in) :: x
!   funcgrad = tfunk(x,0)
!   end function funcgrad
!!
   end subroutine safunc1

!----------------------------------------------------------------------------------------------------        

   logical function is_profile_function(cftype) result(lval)
   integer, intent(in) :: cftype
!
   select case(cftype)

      case(RWPC_COSTF,RWP_COSTF,CHI2_COSTF)
           lval = .true.
      case(CHI2FACT,CHI2I,CHI2I_SD,CHI2I_NEW)
           lval = .false.
      case default 
           lval = .false.
   end select
!
   end function is_profile_function

!----------------------------------------------------------------------------------------------------        
   subroutine print_listsa(unitp,ncif)
!
!  Stampa riepilogativa della lista dei modelli
!   
   USE variables, only: cryst
   USE timer
   USE crystal_phase
   integer, intent(in)           :: unitp
   integer, intent(in)           :: ncif
   integer                       :: i
   character(len=10)             :: snum
   integer                       :: nmodel,nph
   character(len=:), allocatable :: cifname
!
   nmodel = cryst(1)%nmodels()  ! nmodel is the same for all phase
   write(unitp,'(/3x,90("-"),/a)')'           List of structural models sorted according to cost function '
   if (any(cryst%is_known())) then
       write(unitp,'(a/,3x,90("-"))') &
   '      RUN          CF             Ratio       found        RMSD    time           saved on'
       do i=1,nmodel
          write(snum,'(i0)')i
          do nph=1,numphase(cryst)
             if (i <= ncif) then
                 cifname = trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
             else
                 cifname = ' '
             endif
             if (nph == 1) then
                 write(unitp,'(2x,i5,2x,f15.6,f15.3,5x,i5,"/",i0,t55,f10.4,2x,a,t84,a)')   &
                 int(cryst(nph)%model(i)%info(5)),cryst(nph)%model(i)%info(1),cryst(nph)%model(i)%info(6),   &
                 int(cryst(nph)%model(i)%info(2)),cryst(nph)%natpub(),  &
                 cryst(nph)%model(i)%info(3),secs_to_string(cryst(nph)%model(i)%info(4)),    &
                 cifname
!                 trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
             else
                 write(unitp,'(t45,i5,"/",i0,t55,f10.4,t84,a)')   &
                 int(cryst(nph)%model(i)%info(2)),cryst(nph)%natpub(),  &
                 cifname
!                 cryst(nph)%model(i)%info(3),trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
             endif
          enddo
       enddo 
   else
       write(unitp,'(a/,3x,90("-"))') '      RUN        CF                Ratio         time         saved on'
       do i=1,nmodel
          write(snum,'(i0)')i
          do nph=1,numphase(cryst)
             if (i <= ncif) then
                 cifname = trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
             else
                 cifname = ' '
             endif
             if (nph == 1) then
                 write(unitp,'(2x,i5,2x,f15.6,f15.3,8x,a,t66,a)')   &
                 int(cryst(nph)%model(i)%info(5)),cryst(nph)%model(i)%info(1),cryst(nph)%model(i)%info(6),    &
                 secs_to_string(cryst(nph)%model(i)%info(4)),cifname
!                 secs_to_string(cryst(nph)%model(i)%info(4)),trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
             else
                 write(unitp,'(t66,a)')cifname
!                 write(unitp,'(t66,a)')trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
             endif
          enddo
       enddo 
   endif
!
   end subroutine print_listsa
   
!----------------------------------------------------------------------------------------------------

   subroutine overlay_models_list(nph,nmodel)
!
!  Fa in modo che i modelli possano essere sovrapponibili
!
   USE compare_struct
   USE PhaseModule, only:shiftexam
   USE model_util
   USE reflection_type_util
   USE crystal_phase
   USE variables, only: cryst
   USE sinterface
   use connect_mod
   use General, only: neutro, is_anomalous
   integer, intent(in)                               :: nph
   integer, intent(in)                               :: nmodel
   integer                                           :: i
   type(atom_type), allocatable, dimension(:)        :: atm1,atm2
   type(reflection_type), dimension(:), allocatable  :: ref1,ref2
   type(bond_type), dimension(:), allocatable        :: bonds
!
   if (nmodel <= 1) return
   allocate(ref1(cryst(nph)%numref()),source=cryst(nph)%ref)
   allocate(ref2(cryst(nph)%numref()),source=cryst(nph)%ref)
   if (numphase(cryst) > 1) call set_spaceg_new(cryst(nph)%spg)  ! set spg for multiphase system
!
!  Prendi il primo modello della lista
   call model_get(cryst(nph)%model(1),atm1)
   call sfcalc(ref1,atm1,cryst(nph)%spg,cryst(nph)%elem,neutro,is_anomalous)
!
!  Sovrapponi gli altri modelli al primo
   do i=2,nmodel
      call model_get(cryst(nph)%model(i),atm2)
      call sfcalc(ref2,atm2,cryst(nph)%spg,cryst(nph)%elem,neutro,is_anomalous)
      call examing(ref1,ref2)
      call copy_bonds(bonds,cryst(nph)%bond)  ! avoid changes in bonds
      call drawup_models(atm1,cryst(nph)%bond,atm2,bonds,cryst(nph)%cell,cryst(nph)%spg,   &
           shiftexam,kfix=.true.,kenant=.false.,kinorg=.false.)
      call model_set(cryst(nph)%model(i),atm2)
   enddo
!
   end subroutine overlay_models_list

!----------------------------------------------------------------------------------------------------

   integer function numref_from_res(resval,refl)   result(numr)
!   
!  Calcola il numero di riflessi associati fino risoluzione resval
!
   USE reflection_type_util
   real, intent(in)                                :: resval
   type(reflection_type), dimension(:), intent(in) :: refl
   integer                                         :: i
!   
   !numr = count(refl%dval() >= resval) ! this crash in ifort - release version
   numr = 0
   do i=1,size(refl)
       if (refl(i)%dval() >= resval) numr=numr+1
   enddo
!
   end function numref_from_res

!----------------------------------------------------------------------------------------------------

   integer function numref_from_res_old(resval,refl)   result(numr)
!   
!  Calcola il numero di riflessi associati fino risoluzione resval
!
   USE counts, only: thvalue
   USE reflection_type_util
   USE variables, only: dataset
   real, intent(in)                                :: resval !,wave
   type(reflection_type), dimension(:), intent(in) :: refl
   real                                            :: tthrval, wave
!   
   wave = dataset(1)%wave(1)
   tthrval = thvalue(resval,wave)
   numr = count(refl%tthd(1) <= tthrval)
     !write(0,*)'NUMR: ',numr, count(refl%dval() >= resval)
!
   end function numref_from_res_old

!----------------------------------------------------------------------------------------------------
   
   subroutine export_model(codefile,nph,algo)
   USE gen_frm, only: CIF_FILE, export_structure
   USE variables, only: cryst,dataset
   USE model_util
!
   integer, intent(in)                        :: codefile
   integer, intent(in)                        :: nph
   integer                                    :: nrun
   real                                       :: cfval
   integer, intent(in)                        :: algo
   character(len=250)                         :: filename
   character(len=30)                          :: snum,snum1
   character(len=200)                         :: commline
   type(atom_type), dimension(:), allocatable :: atmr
   real, dimension(7)                         :: infomodel 
!
   write(snum,'(i0)')codefile
   filename = trim(cryst(nph)%cr_name)//'_best'//trim(snum)// '.cif'
   call model_get(cryst(nph)%model(codefile),atmr,fom=infomodel)
   nrun = int(infomodel(5))
   cfval = infomodel(1)
   write(snum,'(f15.3)')cfval
   write(snum1,'(i0)')nrun
   commline = trim(algoname_short(algo))//' at run '//trim(snum1)//' CF = '//trim(adjustl(snum))
   call export_structure(filename,atmr,cryst(nph)%cell,cryst(nph)%spg,cryst(nph)%elem,CIF_FILE,   &
                         cryst(nph)%cr_name,dataset(1)%wave(1),comm=commline)
!
   end subroutine export_model

!----------------------------------------------------------------------------------------------------

   subroutine save_sa(unitp,gopt)
   integer, intent(in)              :: unitp
   type(opti_info_type), intent(in) :: gopt   
   write(unitp)gopt
   end subroutine save_sa

!----------------------------------------------------------------------------------------------------

   subroutine read_sa(unitp,gopt,err)
   USE errormod
   integer, intent(in)               :: unitp
   type(opti_info_type), intent(out) :: gopt   
   type(error_type), intent(out)     :: err
   integer                           :: ier
   read(unitp, iostat = ier)gopt
   if (ier /= 0) call err%set("Error on reading global optimization conditions")
   end subroutine read_sa

!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine set_doc(doc,atcode)
!corr   USE arrayutil
!corr   type(doc_type), intent(inout)     :: doc
!corr   integer, dimension(:), intent(in) :: atcode
!corr   integer                           :: ndoc
!corr   integer                           :: nat
!corr   integer                           :: i
!corr   ndoc = count(atcode == 1)
!corr   nat = size(atcode)
!corr   if (ndoc == 0) then             ! DOC non active
!corr       doc%dtype = 0
!corr       !call reallocate(doc%vet,0)
!corr       call delete_array(doc%vet)
!corr   else
!corr       !call reallocate(doc%vet,ndoc)
!corr       call resize_array(doc%vet,ndoc)
!corr       if (ndoc == nat) then  ! DOC active on all atoms
!corr           doc%dtype = 1
!corr           doc%vet = [(i,i=1,nat)]
!corr       else                        ! DOC active on selected atoms
!corr           doc%dtype = 2
!corr           doc%vet = pack([(i,i=1,nat)],mask=atcode == 1)
!corr       endif
!corr   endif
!corr   end subroutine set_doc
!corr
!corr!----------------------------------------------------------------------------------------------------
!corr
!corr   logical function doc_active(doc) 
!corr   class(doc_type), intent(in) :: doc
!corr   doc_active = (doc%dtype > 0) 
!corr   end function doc_active
!corr
!corr!----------------------------------------------------------------------------------------------------
!corr
!corr   logical function doc_all_atoms(doc)
!corr   class(doc_type), intent(in) :: doc
!corr   doc_all_atoms = (doc%dtype == 1)
!corr   end function doc_all_atoms
!corr
!corr!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine doc_print(doc,kpr,atom)
!corr   USE atom_basic
!corr   class(doc_type), intent(in)               :: doc
!corr   integer, intent(in)                       :: kpr
!corr   type(atom_type), dimension(:), intent(in) :: atom
!corr   integer, parameter                        :: MAXLENDOC = 60
!corr   integer                                   :: natdoc
!corr   if (.not.doc%active()) return
!corr   if (doc%all_atoms()) then
!corr       write(kpr,'(1x,a)')'Dynamical Occupancy Correction is active on all atoms'
!corr   else
!corr       natdoc = size(doc%vet)
!corr       if (natdoc == 1) then
!corr           write(kpr,'(1x,a,i0,a)')'Dynamical Occupancy Correction is active on selected atom: '//  &
!corr           slabvet(doc%vet,atom%lab)
!corr       else
!corr           write(kpr,'(1x,a,i0,a)')'Dynamical Occupancy Correction is active on ',size(doc%vet),' selected atoms: '//  &
!corr           slabvet(doc%vet,atom%lab,maxlen=MAXLENDOC)
!corr       endif
!corr   endif
!corr   end subroutine doc_print
!----------------------------------------------------------------------------------------------------

   subroutine check_resp(crystal,resval)
!   
!  Check if resolution is comparable with the number of parameters
!
!corr   USE variables, only: dataset
!corr   USE counts, only: dvalue
   USE crystal_phase, only: crystal_phase_t
!corr   implicit none
   type(crystal_phase_t), intent(in) :: crystal
   real, intent(inout)               :: resval
   integer                           :: nrefpar
   integer                           :: nrefval
   integer, parameter                :: OBS_FOR_PAR = 7
!   
!corr   nrefval = numref_from_res(resval,crystal%refc,dataset(1)%wave(1)) ! number of reflections up to resval
   nrefval = numref_from_res(resval,crystal%refc) ! number of reflections up to resval
   if (nrefval < crystal%numref()) then
!
!      Numero minimo di riflessi in base al numero di paramatri (edof+idof)
       !nrefpar = nparam * OBS_FOR_PAR
       nrefpar = (crystal%edof() + crystal%idof()) * OBS_FOR_PAR
       if (nrefpar > crystal%numref()) nrefpar = crystal%numref()
!   
       if (nrefval < nrefpar) then
           !resval = dvalue(crystal%refc(nrefpar)%tthd(1),dataset(1)%wave(1))    
           resval = crystal%refc(nrefpar)%dval()
       endif           
   endif
!
   end subroutine check_resp

!----------------------------------------------------------------------------------------------------   

   function found_string(nat,natpub,rmsd)  result(str)
   use strutil
   integer, intent(in) :: nat,natpub
   real, intent(in)    :: rmsd
   character(len=:), allocatable :: str
   !character(len=:), allocatable :: strnat
!
   !allocate(character(len=ndigits(natpub)::strnat)
   str = 'found: '//i_to_s1(nat,ndigits(natpub))//'/'//i_to_s(natpub)//' RMSD:'//r_to_s(rmsd,4)
   !write(str,'(a,i0,a,i0,a)')'found: ',nat,'/',natpub,' RMSD:'//r_to_s(rmsd,4)
!
   end function found_string

!----------------------------------------------------------------------------------------------------

   subroutine evaluate_res(res,numref,tthmaxr,ier)
!
!  Data la risoluzione calcola il numero di riflessi e il 2theta massimo
!
   USE counts, only: thvalue
   USE variables, only: dataset,cryst
   USE reflection_type_util
!
   implicit none
   real, intent(in)     :: res      ! risoluzione
   integer, intent(out) :: numref   ! numero di riflessi
   real, intent(out)    :: tthmaxr  ! 2theta massimo
   integer              :: ier      ! >0 se res e' inferiore al valore minimo
   real                 :: resmin
!
!  Calcola il valore minimo consentito per res
   !resmin = dvalue(dataset(1)%tmax,dataset(1)%wave(1))
   resmin = dataset(1)%resolution()
!
   if (res < resmin) then
       tthmaxr = dataset(1)%tmax
       numref = numrefl(cryst(1)%refc)
       ier = 1   ! segnala che res e' fuori range
   else
       tthmaxr = thvalue(res,dataset(1)%wave(1))
       !numref = count(cryst(1)%refc%tthd(1) <= tthmaxr .and. cryst(1)%refc%tthd(1) >= dataset(1)%tmin)   
       numref = count(cryst(1)%refc%dval() >= res)
       ier = 0
   endif
!
   end subroutine evaluate_res

!----------------------------------------------------------------------------------------------------

   subroutine randomize_parameters() bind(C,name="randomize_parameters")
!
!  'Randomize button'
!
   USE fragmentmod
   USE variables, only: cryst
   USE ilmviewmod
   USE General, only:lo,neutro,is_anomalous
   USE reflection_type_util
   USE rotationmod
   USE cryutil
   implicit none
   integer :: i
!
   call randomize_molecule(cryst(1)%at,cryst(1)%bond,cryst(1)%cell,cryst(1)%frag,cryst(1)%rot)  ! randomize atm
   call set_atm_in_frag(cryst(1)%frag,cryst(1)%at)                  ! update fragment info
   do i=1,cryst(1)%numrot()                                         ! update torsion value
      cryst(1)%rot(i)%torsval = torsion_value_for_rotat(cryst(1)%rot(i),cryst(1)%at,cryst(1)%cell)
   enddo
   call ilmview_draw(cryst(1)%at,cryst(1)%bond,cryst(1)%elem,cryst(1)%cell,cryst(1)%spg)                          ! update graphic view
   if (cryst(1)%is_known()) then
       call cryst(1)%sfactor(neutro,is_anomalous)
       call to_coord_cry(cryst(1),lo,-1)
       call write_message_found1(5,cryst(1)%natfound,cryst(1)%natpub(),cryst(1)%rmsd)
   endif
!
   end subroutine randomize_parameters

!----------------------------------------------------------------------------------------------------

   subroutine showSolution(n) bind(C,name="showSolution")
!qt   USE sannel, only: ginfo, sa_profile, NINFOMODEL
   USE variables, only: cryst,dataset
   USE ilmviewmod
   USE view
   USE messagemod
   USE strutil
   USE model_util
   USE fouriermap
   USE atom_type_util
   USE datasetmod
   USE counts
   USE mapinfo
   USE fourier_mod
   USE General, only: neutro, is_anomalous
   USE iso_c_binding
!qt   implicit none
   integer(c_int), intent(in), value          :: n
   real, dimension(NINFOMODEL)                :: infomodel
   type(bond_type), dimension(:), allocatable :: legm
   integer                                    :: nph
   logical                                    :: sfac 
   real                                       :: f000
   integer                                    :: nrow
!
   nph = active_phase
!   
!  Seleziona il modello n dalla lista
   call model_get(cryst(nph)%model(n),cryst(nph)%at,leg=legm,fom=infomodel)
!
!  Disegna il modello
   if (.not.jav_opened()) call ilmview_open()
   call ilmview_draw(cryst(nph)%at,legm,cryst(nph)%elem,cryst(nph)%cell,cryst(nph)%spg)
!
!  Disegna il profilo
   sfac = .false.
   if (dataset(1)%datatype == POW_DATA) then
       call cryst(nph)%sfactor(neutro,is_anomalous)
       sfac = .true.
       call sa_profile(square=.false.) 
       call vedinew(8,1,ginfo%nco,rescale=0)
   endif
!
   if (is_map_enabled(cryst(nph)%mapinfo)) then
       call map_set_centre(baricentro(cryst(nph)%at))
       if (.not.sfac) call cryst(nph)%sfactor(neutro,is_anomalous)
       if (cryst(nph)%mapinfo%ftype == FOUR_FO .or. cryst(nph)%mapinfo%ftype == FOUR_FC) then
           f000 = get_f000(cryst(nph)%at,cryst(nph)%elem,cryst(nph)%spg,dataset(nph)%radtype)
       else
           f000 = 0
       endif
       call jav_show_map(cryst(nph)%mapinfo%ftype,dvalue(dataset(1)%tmax,dataset(1)%wave(1)),cryst(nph)%ref,   &
                         'SA Map',sogmap=cryst(nph)%mapinfo%sog,fctype=1,scalem=.true.,f000=f000,cell=cryst(nph)%cell)
   endif
!
!qt   call clear_messages(ipos=2)
   nrow = 4 ! title + trial + file + CF
   nrow = nrow + count(cryst%is_known())
   if (dataset(1)%datatype == POW_DATA) nrow = nrow + 1
   call SetRowCountMesg(nrow,1)
   if (n == 1) then
       call write_message('Best structure, run:'//trim(i_to_s(int(infomodel(5)))),pos=2)
   else
       call write_message('Selected run:'//trim(i_to_s(int(infomodel(5)))),pos=2)
   endif
   if (n <= ginfo%nexport) then
       call write_message('Saved in file:'//trim(cryst(nph)%cr_name)//'_best'//trim(i_to_s(n))//'.cif',pos=3)
   else
       call write_message('Saved in file:not saved',pos=3)
   endif
   call write_message('CF:',rnum=infomodel(1),pos=4)
   if (dataset(1)%datatype == POW_DATA) call write_message('Ratio:',rnum=infomodel(6),pos=0)
   if (cryst(nph)%is_known()) then
       cryst(nph)%natfound = nint(infomodel(2))        ! found atoms
       cryst(nph)%rmsd = infomodel(3)
       call write_message_found1(0,cryst(nph)%natfound,cryst(nph)%natpub(),cryst(nph)%rmsd)
   endif
!   
   end subroutine ShowSolution

!----------------------------------------------------------------------------------------------------

   subroutine saveSolution(n) bind(C,name="saveSolution")
   USE variables, only: cryst
   USE atom_type_util
!qt   USE sannel, only: search_overlapped_atoms_print
   USE ilmviewmod, only: active_phase
   USE iso_c_binding
!
!qt   implicit none
   integer(c_int), intent(in), value :: n
   integer                           :: nph
!
   nph = active_phase
   call SetToggleVal('SymmOn'//char(0), 0)
   write(6,'("Solution n. ",i0," was selected using Solutions menu")')n
   call print_atoms(cryst(nph)%at,title='Structure solution',write_occ=.true.,kpr=go_unit)
   call search_overlapped_atoms_print(cryst(nph)%at,cryst(nph)%cell%get_g(),cryst(nph)%spg)
   end subroutine SaveSolution

!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine CancelSolution(n)
!corr!qt   implicit none
!corr   integer       ::n
!corr   call SetToggleVal('SymmOn'//char(0), 0)
!corr   call ShowSolution(n)
!corr   end subroutine CancelSolution

!----------------------------------------------------------------------------------------------------

   subroutine exitsanneal()
!
!  Routine chiamata dalla grafica alla pressione del tasto 'Exit'
!
   USE Molcom, only:jscreen
   !USE sannel, only: gpar
   USE General, only: cell
   USE genlsq
   !USE counts
   USE RIETV
   USE variables, only:cryst, atm
   USE exportfiles
   USE ilmviewmod
!qt   USE grexpomod
   USE messagemod
   USE enable_amb
   USE profile_function
   !USE arrayutil
   USE sinterface, only: fillatom
   USE crystal_phase
   USE reflection_type_util
   use progtype, only: DEALLOCA
   implicit none
   integer :: nph
!
!  dealloca parametri
   if (allocated(gpar)) deallocate(gpar)
!
!  Ripristina riflessi
   do nph=1,numphase(cryst)
      call cryst(nph)%set_used_ref(numrefl(cryst(nph)%refc))
      cryst(nph)%ref(:)%pk = 20
   enddo
!
!  dealloca riflessi e conteggi
   call Allocafuncls(DEALLOCA)
!
!  dealloca i plot
   if (jscreen > 0) then
!qt       call clear_plot(grplot,1)
!qt       call gr_close_viewer()
!!FIX this call in cpp at the end of run procedure
       call jav_set_modify(1)
!qt       call abilita_tasti('structure+pattern')
   endif
   call cryst(1)%get_atoms(atm)
   call fillatom(cryst(1)%at)
   cell(1:6) = cryst(1)%get_cellpar()
   call clear_messages()
!
   call export_files()
!
   end subroutine exitsanneal

!----------------------------------------------------------------------------------------------------

   subroutine make_test(gopti,scond,test_type)
   USE simulated_anneal, only: SimAnn_Conditions_type
   type(opti_info_type), intent(inout)         :: gopti    
   type(SimAnn_Conditions_type), intent(inout) :: scond                     
   integer, intent(in)                         :: test_type
!
   gopti%nrun = 20
!
   select case (test_type)
      case(1)   ! test for parallel SA
       scond%cooperate = .true.
!       scond%nrare = mpi_prog_numprocs / 5
!       if (scond%nrare == 0) scond%nrare = 1
       gopti%algo = SA_LOCAL
       gopti%autoalgo = .false.

      case (2)  ! test for antibump restraints
!      call cryst(1)%set_doc(.false.) ! disable doc for test
!      call set_antibump_auto(bump_sa,bumptab,atm,cellt,fragment,bumpset%scaledist)
!      if (nrestraints(bump_sa) == 0) stop

   end select

   end subroutine make_test

END MODULE sannel
   
!----------------------------------------------------------------------------------------------------

   subroutine ges_atom_parameters_win(code,atomw,shifta,shiftval)
   USE sannel, only: rigidb_copy, shiftatomval
   USE variables, only: cryst
   USE atom_type_util
   USE rigid_body
   implicit none
   integer, intent(in)                       :: code
   real, save                                :: shiftatomval_copy
   type(atom_type), dimension(*), intent(in) :: atomw
   integer, dimension(*), intent(inout)      :: shifta
   real, intent(in)                          :: shiftval
   integer                                   :: i,natom,nph
!
   nph = 1
   select case (code)
      case (0)        ! salva i parametri all'apertura della finestra
        call save_atoms(cryst(1)%at)
        call copy_rigidb(cryst(nph)%rigidb,rigidb_copy)
        shiftatomval_copy = shiftatomval

      case (1)        ! applica modifiche su 'OK'
!
!       aggiorna i parametri
        shiftatomval = shiftval
        call clear_rigidb(cryst(nph)%rigidb)
        do i=1,cryst(1)%natoms()
           if (shifta(i) > 0) then
               call add_rigidb(cryst(nph)%rigidb,[i],1,.true.,.false.,shiftatomval,0.0,0)
           endif
           cryst(1)%at(i) = atomw(i)
        enddo
        call cryst(1)%set_doc(1,atomw(1)%doc)  ! update doc settings
!
!       disegna il nuovo modello
        call clear_saved_atoms()
        call clear_rigidb(rigidb_copy)
        
      case (2)        ! rispristina parametri su 'Cancel'
        call restore_atoms(cryst(1)%at,natom)
        call cryst(1)%set_doc(1,atomw(1)%doc)  ! update doc settings
        call clear_saved_atoms()
        call copy_rigidb(rigidb_copy,cryst(nph)%rigidb)
        call clear_rigidb(rigidb_copy)
        shiftatomval = shiftatomval_copy
        do i=1,cryst(1)%natoms()
           shifta(i) = 0
        enddo
        do i=1,cryst(nph)%nrigid()
           if (count(cryst(nph)%rigidb(i)%rcod > 0) > 0) then
               shifta(cryst(nph)%rigidb(i)%at(1)) = 1
           endif
        enddo
!
   end select
!
   end subroutine ges_atom_parameters_win

!----------------------------------------------------------------------------------------------------
!qt   subroutine get_info_for_gtklist(nrow,nrcol,cfcol,strcol) bind(C,name="get_info_for_gtklist")
!qt   USE model_util
!qt   USE variables, only: cryst
!qt   use iso_c_binding
!qt   implicit none
!qt   integer(c_int), intent(out)                                     :: nrow    ! numero di righe della gtklist
!qt   integer(c_int), dimension(*), intent(out)                       :: nrcol   ! colonna 'Run nr.'
!qt   real(c_float), dimension(*), intent(out)                          :: cfcol   ! colonna 'Cost Function'
!qt   character(len=*) , intent(out)                           :: strcol  ! nome file
!qt   integer                                                  :: i
!qt   real, dimension(5)                                       :: infolist
!qt!   
!qt   nrow = cryst(1)%nmodels() 
!qt!   
!qt   strcol = trim(cryst(1)%cr_name)//char(0)
!qt   do i=1,nrow  
!qt      call model_get(cryst(1)%model(i),fom=infolist)
!qt      nrcol(i) = nint(infolist(5))
!qt      cfcol(i) = infolist(1)
!qt   enddo
!qt!
!qt   end subroutine get_info_for_gtklist
!qt


!----------------------------------------------------------------------------------------------------

   subroutine fcnprint(gopti,popt,xopt,fopt,temp,nftot,nmtemp)
!
!  Stampa informazioni alla fine di ogni step di temperatura
!  Opzioni di output: S = stampa parametri
!                     P = draw pattern su viewer principale
!                     M = draw molecule su viewer principale
!                     V = draw on secondary viewer
!                     B = manage status and progress bar
   USE variables, only:cryst   !fcnprint
   USE sannel, only: timesa, SATIME_TOT, SATIME_GRAPH, is_profile_function, chisquared_res, go_unit, &
             is_go_print, SA_LOCAL, ratio_value, gopt_statusbar, gopt_draw, xpar_to_crystals,        &
             found_string, numcyc_calc
   USE simulated_anneal
   USE General, only:neutro,is_anomalous
   USE TestModel
   USE reflection_type_util
   USE compare_struct
   USE Conteggi, only:npunti
   USE Molcom
   USE messagemod
   USE strutil
   USE timer
   USE unit_cell
   USE anti_bump
   USE mpi_prog
   USE rrestr
   USE crystal_phase
   USE interop_cpp_for, only: add_point_to_plot2, dosingleiteration
   USE type_constants, only: DP
!                
   implicit none
   type(opti_info_type), intent(in)           :: gopti    
   character(len=*), intent(in)               :: popt     ! stringa opzioni di stampa
   real, dimension(:), intent(in)             :: xopt     ! parametri minimo
   real, intent(in)                           :: fopt     ! cf per parametri minimo
   real, intent(in)                           :: temp     ! temperatura
   integer, intent(in)                        :: nftot    ! num. totale di movimenti
   integer, intent(in)                        :: nmtemp   ! num. di movimenti alla temperatura temp
   integer                                    :: nph
   real, save                                 :: ratiopt
   character(len=300)                         :: stringa
   type(reflection_type), dimension(:), allocatable :: refc
   logical                                    :: is_anti_bump, is_restraints
   real                                       :: totaltime
   real                                       :: fbump,fres
   character(len=3) :: adv
!corr   type(timer_type),save :: t_draw
   logical :: sdraw
   integer :: nlen
   character(len=:), allocatable :: sform
   real(DP)                                   :: init_time = 0_dp
   real(DP)                                   :: end_time
   real(DP)                                   :: range_time = 1.0_dp ! intervallo di tempo dopo il quale updatare la statusbar
   logical                                    :: oktime 
!
   if (.not.is_go_print()) return
!
   call cpu_time(end_time)
   oktime = end_time - init_time > range_time
   if (oktime) init_time = end_time
   !if (.not.oktime) return
!
!  Aggiorna la status e progress bar
   if (index(popt,'B') > 0) then
       if (oktime .or. nmtemp == numcyc_calc) then
           call write_message('Moves:',inum=nftot,pos=vmess(LINE_MOVES),kproc=0)
           call write_message('Time:'//secs_to_string(timesa(SATIME_TOT)%get_time(),.true.),pos=vmess(LINE_TIME),kproc=0)
           !call progress_bar_update(cval=nmtemp,fval=numcyc_calc,kproc=0)
           call progress_bar_update(fval=numcyc_calc,kproc=0)
       !write(0,*)'2nmtemp=',nmtemp,' numcyc_calc=',numcyc_calc
           !call go_progress_bar_update(nftot,nmtemp)
       endif
   endif
!
!  Stampa informazioni alla fine di uno step di temperatura
   if (index(popt,'S') > 0) then    !S=brevi informazioni
!
!      aggiorna ratio perche un nuovo minimo e stato trovato
       if (npunti > 0) then     ! da migliorare
           ratiopt = ratio_value(fopt,gopti)
       else
           ratiopt = 0.0
       endif
       call write_message('Ratio:',rnum=ratiopt,pos=vmess(LINE_RATIO),kproc=0)
!
       adv = 'no'
       write(go_unit,'(a,i0)',advance=adv)'Run:',gopti%nrunc
       nlen = 4+ndigits(gopti%nrunc)
       if (gopti%algo == SA_LOCAL) then
           write(go_unit,'(a,f8.2)',advance=adv)'  Temp:',temp
           nlen = nlen + 7 + 8
       endif
       is_anti_bump = any(cryst(:)%numresab() > 0)
       is_restraints = cryst(1)%numres() > 0
       if (is_anti_bump .or. is_restraints) then
           write(go_unit,'(1x,a,f10.3)',advance=adv)'CF:',fopt
           nlen = nlen + 1 + 3 + 10
       else
           write(go_unit,'(1x,a,f10.3)',advance=adv)trim(gopti%cost_function_name_short)//':',fopt
           nlen = nlen + 1 + len_trim(gopti%cost_function_name_short) + 1 + 10
       endif
!!!!FIX: should be better saving CFs of xopt-rewrite func with array of CFs in output
       if (is_anti_bump .or. is_restraints .or. cryst(1)%is_known()) then
           call xpar_to_crystals(xopt)
       endif
       if (is_anti_bump) then
           fbump = 0
           do nph=1,numphase(cryst)
              !corr fbump = fbump + cryst(nph)%abset%weight*    &
              fbump = fbump + antibump_function(cryst(nph)%abres,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%cell%get_g(),.false.)
           enddo
           write(go_unit,'(a,f10.3)',advance=adv)' A-bumpCF:',fbump
           nlen = nlen + 10 + 10
       endif
       if (is_restraints) then
           fres = chisquared_res(cryst(1))
           write(go_unit,'(a,f10.3)',advance=adv)' CF-res:',fres
           nlen = nlen + 10
       endif
       totaltime = real(timesa(SATIME_TOT)%get_time())
       if (any(cryst%is_known())) then
!   
!          calcola fasi senza ricoprire ref
           write(go_unit,'(a,a13)',advance='no')'  Time:',secs_to_string(totaltime)
           nlen = nlen + 7 + 13
           do nph=1,numphase(cryst)
              call copy_ref(refc,cryst(nph)%ref)
              call sfcalc(refc,cryst(nph)%at,cryst(nph)%spg,cryst(nph)%elem,neutro,is_anomalous)
              call to_coord_sir(-1,cryst(nph)%at,refc,size(refc),cryst(nph)%cell,cryst(nph)%spg,go_unit,cryst(nph)%atpub)
              if (nph == 1) then
                  stringa = ' (found: '//trim(i_to_s(int(vett_test(3))))//'/'//trim(i_to_s(cryst(nph)%natpub()))//   &
                  ' RMSD: '//trim(r_to_s(rmsd,3))//')'
                  call write_message( 'CF:'//trim(r_to_s(fopt,3))//trim(stringa),pos=vmess(LINE_CF),kproc=0)
              endif
              if (nph <= 2) then
                  adv = 'no'
              else
                  if (nph == numphase(cryst) .or. mod(nph,2) == 0) then
                      adv = 'yes'
                  else
                      adv = 'no'
                  endif
              endif
              if (mod(nph,2) == 0) then
                  sform = '(1x,a)'
              else
                  if (nph == 1) then
                      sform = '(1x,a)'
                  else
                      sform = '('//i_to_s(nlen+1)//'x,a)'
                  endif
              endif
              write(go_unit,sform,advance=adv)found_string(int(vett_test(3)),cryst(nph)%natpub(),rmsd)
              if (numphase(cryst) == 1 .or. nph == 2) write(go_unit,'(a,i0)')'  Moves:',nftot 
           enddo
           if (gopti%fmin < fopt) then
               stringa = '(found: '//trim(i_to_s(int(cryst(1)%model(1)%info(2))))//'/'//trim(i_to_s(cryst(1)%natpub()))//   &
               ' RMSD: '//trim(r_to_s(cryst(1)%model(1)%info(3),3))//')'
               call write_message('CFmin:'//trim(r_to_s(gopti%fmin,3))//' at run.'    &
                    //trim(i_to_s(gopti%rmin))//' '//trim(stringa),pos=vmess(LINE_CFMIN),kproc=0)
           else
               call write_message('CFmin:'//trim(r_to_s(fopt,3))//' at run.'//trim(i_to_s(gopti%nrunc)),  &
                    pos=vmess(LINE_CFMIN),kproc=0)
           endif
       else
           call write_message('CF:',rnum=fopt,pos=vmess(LINE_CF),kproc=0)
           write(go_unit,'(a,a10,4x,a,i0)')'  Time:',secs_to_string(totaltime),'Moves:',nftot
           if (gopti%fmin < fopt) then
               call write_message('CFmin:'//trim(r_to_s(gopti%fmin,3))//' at run.'  &
                    //trim(i_to_s(gopti%rmin)),pos=vmess(LINE_CFMIN),kproc=0)
           else
               call write_message('CFmin:'//trim(r_to_s(fopt,3))//' at run.'//trim(i_to_s(gopti%nrunc)), &
                    pos=vmess(LINE_CFMIN),kproc=0)
           endif
       endif   
       if (is_anti_bump) call write_message('Anti-bump CF:',rnum=fbump,pos=vmess(LINE_CFBUMP),kproc=0)
       if (is_restraints) call write_message('Restraints CF:',rnum=fres,pos=vmess(LINE_CFRES),kproc=0)
   endif
!   
!  Output grafico su xopt
   if (index(popt,'M') > 0) then
       !sdraw = diff_time(t_draw,0.5)
       sdraw = oktime
   else
       sdraw = .false.
   endif
   call gopt_draw(sdraw,index(popt,'P')>0,gopti)
!
!  Output grafico su viewer secondario
   if (index(popt,'V') > 0) then  
       call timesa(SATIME_GRAPH)%tic() ! calcola il tempo impiegato dalle operazioni grafiche  
!           
!      Aggiorna plot CF
       call add_point_to_plot2(gopti%nrunc,real(nftot),fopt)
!
       call timesa(SATIME_GRAPH)%toc() ! calcola il tempo impiegato dalle operazioni grafiche  
   endif
   if (oktime .and. jscreen > 0) then
       call DoSingleIteration()
   endif

   flush(go_unit)
!
   end subroutine fcnprint         

!----------------------------------------------------------------------------------------------------   

   subroutine go_progress_bar_update(nftot,nmtemp)
   use sannel, only: timesa, SATIME_TOT, numcyc_calc
   use messagemod
   use timer
   use simulated_anneal, only: vmess, LINE_MOVES, LINE_TIME
   implicit none
   integer, intent(in) :: nftot,nmtemp
   real(DP) :: init_time = 0_dp
   real(DP) :: end_time
   real(DP) :: range_time = 1.0_dp
   logical  :: oktime 

   call cpu_time(end_time)
   oktime = end_time - init_time > range_time
   if (oktime) init_time = end_time

   if (oktime .or. nmtemp == numcyc_calc) then
       call write_message('Moves:',inum=nftot,pos=vmess(LINE_MOVES),kproc=0)
       call write_message('Time:'//secs_to_string(timesa(SATIME_TOT)%get_time(),.true.),pos=vmess(LINE_TIME),kproc=0)
       call progress_bar_update(cval=nmtemp,fval=numcyc_calc,kproc=1)
   endif

   end subroutine go_progress_bar_update

!----------------------------------------------------------------------------------------------------   
   
   subroutine sa_restorepar(hpar,xpar)
!
!  Ripristina i parametri se il movimento non e' accettato
!   
   USE fastfcal
   USE sannel, only: gpar, PO_TYPE, THERMALF_TYPE
   USE variables, only:cryst
   USE reflection_type_util
   USE rrestr
   USE arrayutil
   implicit none
   integer, intent(in) :: hpar
   real, intent(in)    :: xpar
   integer             :: nph
!
   nph = gpar(hpar)%ph
   if (gpar(hpar)%typepar == THERMALF_TYPE) then 
       call fcalc_ripr_thermalf(gpar(hpar)%pat,cryst(nph)%sfact,cryst(nph)%sfactc)
   elseif (gpar(hpar)%typepar == PO_TYPE) then
       cryst(nph)%gpar = xpar !!!!gpo_sav
       call calcola_correzioneOP(cryst(nph)%ref,cryst(nph)%gpar,cryst(nph)%hklpo,cryst(nph)%cell,cryst(nph)%spg)
   else
       call fcalc_ripr_atm(gpar(hpar)%pat,cryst(nph)%spg,cryst(nph)%sfact,cryst(nph)%sfactc)
       if (cryst(nph)%doc_active()) then
           call fcalc_ripr_thermalf(cryst(nph)%vetdoc,cryst(nph)%sfact,cryst(nph)%sfactc)
       endif
       if (size_array(gpar(hpar)%res) > 0) then
           call res_restore_contrib(cryst(nph)%res)
       endif
       if (size_array(gpar(hpar)%abres) > 0) then
           call res_restore_contrib(cryst(nph)%abres)
       endif
!test-res
!test-res           write(71,*)'RESTORE=',cryst(nph)%res%contrib
!test-res
   endif
!   
   end subroutine sa_restorepar

!----------------------------------------------------------------------------------------------------      
   subroutine sa_savepar(hpar)
!
!  Salva i parametri se il movimento e' accettato
!   
   USE fastfcal
   USE sannel, only: gpar, THERMALF_TYPE, PO_TYPE
   USE variables, only: cryst
   USE rrestr
   USE arrayutil
   implicit none
   integer, intent(in) :: hpar
   integer             :: i,nph
!
   nph = gpar(hpar)%ph
   if (gpar(hpar)%typepar == THERMALF_TYPE) then 
       call fcalc_save_thermalf(gpar(hpar)%pat,cryst(nph)%sfact,cryst(nph)%sfactc)
!corr   elseif (gpar(hpar)%typepar == PO_TYPE) then
!corr       gpo_sav = cryst(nph)%gpar
   else
       call fcalc_save_atm(gpar(hpar)%pat,cryst(nph)%spg,cryst(nph)%sfact,cryst(nph)%sfactc)
       if (cryst(nph)%doc_active()) then
           call fcalc_save_thermalf((/(i,i=1,cryst(nph)%natoms())/),cryst(nph)%sfact,cryst(nph)%sfactc)
       endif
       if (size_array(gpar(hpar)%res) > 0) then
           call res_save_contrib(cryst(nph)%res)
       endif
       if (size_array(gpar(hpar)%abres) > 0) then
           call res_save_contrib(cryst(nph)%abres)
       endif
!test-res
!test-res           write(71,*)'SAVE=',cryst(nph)%res%contrib
!test-res
   endif    
!   
   end subroutine sa_savepar

!corr!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine sa_savepar_all()
!corr   USE fastfcal
!corr   USE variables, only: cryst
!corr   implicit none
!corr   integer :: i
!corr!
!corr   call fcalc_save_atm((/(i,i=1,cryst(1)%natoms())/),cryst(1)%spg,cryst(1)%sfact,cryst(1)%sfactc)
!corr!
!corr   end subroutine sa_savepar_all
!corr
!----------------------------------------------------------------------------------------------------

   subroutine sannelexe()   
!
!  Questa routine e' chiamata quando e' attivo il comando %sannel
!
   USE sannel, only: goptim,sa_prelim,sa_run,exitsanneal
   USE General, only:ifAutomatic
   USE molcom, only:jscreen
   USE errormod
   USE bb_bc
   USE enable_amb
   USE variables, only: cryst
   use simulated_anneal, only: SimAnn_Conditions_type
   use mpi_prog
   use interop_cpp_for, only: c_fine_lavoro, c_fortran_al_lavoro
   use prog_constants, only: ERR_WINDOW
   use sannel, only: is_go_print
   implicit none
   interface
      subroutine aprisannelwin() bind(C,name="aprisannelwin")
      end subroutine aprisannelwin
!corr      subroutine open_import_structure() bind(C,name="open_import_structure")
!corr      end subroutine open_import_structure
   end interface
   type(SimAnn_Conditions_type) :: sc                     
   type(bb_bc_condition)        :: bcond                     
   type(error_type)             :: saerr
   integer                      :: exitCode
!
!corr   if (cryst(1)%natoms() == 0 .and. jscreen > 0) then
!corr       !call open_file_selector('Import starting model for Simulated Annealing' //char(0))
!corr       call open_import_structure()
!corr   endif
!
   if (ifAutomatic == 1 .or. jscreen == 0) then
       if (cryst(1)%natoms() == 0) then
           if (is_go_print()) then
               call MsgWinErr('ERROR','No starting model available to run simulated annealing.',  &
                              ERR_WINDOW,exitCode)
           endif
           return
       endif
!
!      Inizializza e legge le direttive associate al comando %sannel 
       call c_fortran_al_lavoro(0)
       call abilita_tasti(SaveAction)
       call abilita_tasti(DialogOpenAction)
       call sa_prelim(goptim,sc,bcond,.true.,saerr=saerr)
       if (saerr%signal) then
           if (mpi_prog_rank == 0) call saerr%print()
       else
           call sa_run(goptim,sc,bcond,.true.)
           call exitsanneal()
       endif
       call c_fine_lavoro(0)
       call abilita_tasti(RestoreAction)
   else
       call aprisannelwin()
   endif
!
   end subroutine sannelexe

!----------------------------------------------------------------------------------------------------
!corr #if 0   
!corr    subroutine sannelexewin(ierr, resmin, resmax, info, sc, bcond, ifButton, bumpsetC)   
!corr !
!corr !  Questa routine e' chiamata alla pressione del tasto 'Simulated Annealing'
!corr !
!corr    USE sannel, only: bumptab,goptim,sa_prelim,shiftatomval, &
!corr                      set_allowed_cfunction
!corr    USE anti_bump, only: bump_settings
!corr    USE simulated_anneal, only: SimAnn_Conditions_type
!corr    USE variables, only:cryst,dataset
!corr    USE rotationmod
!corr    USE sinterface
!corr    USE bb_bc
!corr    USE enable_amb
!corr    USE ilmviewmod
!corr    USE rrestr
!corr    USE errormod
!corr    USE counts, only: dvalue
!corr    USE rinterface
!corr    USE progtype, only: opti_info_type
!corr    implicit none
!corr    interface
!corr       subroutine open_import_structure() bind(C,name="open_import_structure")
!corr       end subroutine open_import_structure
!corr    end interface
!corr    type(SimAnn_Conditions_type)        :: sc                     
!corr    type(bb_bc_condition)               :: bcond                     
!corr    real*8                              :: resmin,resmax
!corr    type(opti_info_type), intent(inout) :: info
!corr    integer                             :: nfrag
!corr    integer, intent(in)                 :: ifButton    ! (=1 se arrivo qui da menubar)
!corr    integer, intent(out)                :: ierr
!corr    integer                             :: i
!corr    logical                             :: modif
!corr    real                                :: bump_sc
!corr    integer                             :: nrl
!corr    type(bump_settings)                 :: bumpsetC
!corr    type(error_type)                    :: saerr
!corr    integer, dimension(:), allocatable  :: cfvet
!corr    integer                             :: numph, nph
!corr    integer, dimension(:), allocatable  :: natoms,nelem
!corr 
!corr !
!corr    if (.not.cryst(1)%has_symmetry()) then
!corr        call set_cell_and_spg(cryst(1),modif)
!corr        if (modif .and. jav_opened()) then
!corr            call ilmview_draw(cryst(1)%at,cryst(1)%bond,cryst(1)%elem,cryst(1)%cell,cryst(1)%spg)
!corr        endif
!corr    endif
!corr !
!corr    if (cryst(1)%natoms() == 0) then
!corr        !call open_file_selector('Import starting model for Simulated Annealing' //char(0))
!corr        call open_import_structure()
!corr        if (cryst(1)%natoms() == 0) then
!corr            ierr = -1
!corr            return
!corr        endif
!corr    endif
!corr !
!corr    if (ifButton == 0) then
!corr !
!corr !      Sto chiamando SA da comando %sannel nell'input: mi leggo le direttive
!corr        call sa_prelim(goptim,sc,bcond,.true.,saerr=saerr)
!corr    else
!corr !
!corr !      Sto chiamando SA da menubar: non mi leggo le direttive
!corr        call sa_prelim(goptim,sc,bcond,.false.,saerr=saerr)
!corr    endif
!corr !
!corr    if (saerr%signal) then
!corr        call saerr%print()
!corr        ierr = -1
!corr    else
!corr !
!corr !qt       call abilita_tasti('window')
!corr !
!corr !      Calcola risoluzione minima e massima per la grafica
!corr        !resmin = dvalue(dataset(1)%tmax,dataset(1)%wave(1))
!corr        resmin = dataset(1)%resolution()
!corr        !resmax = dvalue(cryst(1)%refc(2)%tthd(1),dataset(1)%wave(1))   ! non va bene se i riflessi non sono ordinati (ncount=0)
!corr        resmax = cryst(1)%refc(2)%dval()   ! non va bene se i riflessi non sono ordinati (ncount=0)
!corr !
!corr        call set_allowed_cfunction(1,cfvet)
!corr        call passa_cfvect(size(cfvet), cfvet)
!corr !       
!corr        call get_info_crystal(numph,natoms,nelem)
!corr        call make_crystal_t()
!corr        call put_info_crystal(numph,natoms,nelem,cryph)
!corr        do nph=1, numph
!corr           call put_info_ph(cryst(nph)%elem, cryst(nph)%at,nph)
!corr           call put_info_po(cryst(nph)%kopcorr,cryst(nph)%hklpo,cryst(nph)%gcode, cryst(nph)%gpar,nph)
!corr        enddo
!corr !
!corr        bump_sc = cryst(1)%abset%weight
!corr        bumpsetC = cryst(1)%abset
!corr        nrl = nrestraints(bumptab)
!corr !
!corr        call passa_abres(nrl)
!corr        do i=1, nrl
!corr           call put_info_abres(i-1,bumptab(i)%code,bumptab(i)%targ, &
!corr                size(bumptab(i)%na),bumptab(i)%na)
!corr        end do
!corr !
!corr !      Apre finestra della grafica
!corr        ierr = cryst(1)%natoms()
!corr        info = goptim
!corr        nfrag = cryst(1)%numfrag()
!corr        call passa_info(cryst(1)%at, cryst(1)%frag%nat, nfrag, cryst(1)%numrot(),cryst(1)%natoms(), shiftatomval)
!corr        do i = 1, nfrag
!corr           call passa_frag(cryst(1)%frag(i)%pos, cryst(1)%frag(i)%rcod, cryst(1)%frag(i)%tlow, &
!corr              cryst(1)%frag(i)%thigh, cryst(1)%frag(i)%rbound, cryst(1)%frag(i)%rat, i-1)
!corr        end do
!corr    endif
!corr !
!corr    end subroutine sannelexewin
!corr #endif
!----------------------------------------------------------------------------------------------------

   subroutine passa_rcod(i, c_rcod, c_tlow, c_thigh, c_rbound, c_rat)
   USE variables, only: cryst
   implicit none
                                                                                                                            
   integer      :: i
   integer      :: c_rcod(6)
   integer      :: c_rat
   real         :: c_tlow(3)
   real         :: c_thigh(3)
   real         :: c_rbound(2)
!
   cryst(1)%frag(i)%rcod = c_rcod
   cryst(1)%frag(i)%tlow = c_tlow
   cryst(1)%frag(i)%thigh = c_thigh
   cryst(1)%frag(i)%rbound = c_rbound
   cryst(1)%frag(i)%rat = c_rat
!                
   end subroutine passa_rcod

!----------------------------------------------------------------------------------------------------

   integer function get_srcod(i, j)
   USE variables, only: cryst
   USE rigid_body
   implicit none
   integer                            :: i
   integer                            :: j
   integer, dimension(:), allocatable :: vet
   integer                            :: nb,nph
!
   nph = 1
   allocate(vet(1))
   nb = find_rigidb(cryst(nph)%rigidb,i,vet)
   if (nb > 0) then
       get_srcod = cryst(nph)%rigidb(vet(1))%rcod(j)
   else
       get_srcod = 0
   endif
!
   end function get_srcod

!----------------------------------------------------------------------------------------------------

   integer function get_rrcod(i)
   USE variables, only: cryst
   implicit none
                                                                                                                
   integer      :: i
!
   get_rrcod = cryst(1)%rot(i)%rcod
!
   end function get_rrcod

!----------------------------------------------------------------------------------------------------

   subroutine  get_rators(i, vettore, cod, val, bounds)
   USE variables, only: cryst
   implicit none
   integer      :: i, j, cod
   integer      :: vettore(4)
   real         :: val, bounds(2)
!
   do j=1, 4
      vettore(j) = cryst(1)%rot(i)%ators(j)
   end do
   do j=1, 2
      bounds(j) = cryst(1)%rot(i)%trbound(j)
   end do
   cod = cryst(1)%rot(i)%rcod
   val = cryst(1)%rot(i)%torsval
!
   end subroutine get_rators

!----------------------------------------------------------------------------------------------------

   subroutine  put_rators(i, cod, val, bounds)
   USE variables, only: cryst
   implicit none
                                                                                                                
   integer      :: i, j, cod
   real         :: val, bounds(2)
!
   do j=1, 2
      cryst(1)%rot(i)%trbound(j) = bounds(j)
   end do
   cryst(1)%rot(i)%rcod = cod
   cryst(1)%rot(i)%torsval = val
!
   end subroutine put_rators

!----------------------------------------------------------------------------------------------------
!corr
!corr   subroutine  put_srcod(i, cod)
!corr!corr   USE sannel, only: shift_atom
!corr   implicit none
!corr                                                                                                                
!corr   integer      :: i, cod(3), j
!corr!
!corr!!!!!!FIX
!corr      write(0,*)'FIX put_srcod'
!corr!corr   do j=1, 3
!corr!corr      shift_atom(i)%rcod(j) = cod(j)
!corr!corr   end do
!corr!
!corr   end subroutine put_srcod
!corr
!----------------------------------------------------------------------------------------------------
!qt
!qt   subroutine finishsannel(info, sc, bcond, shiftatomv, bumpsetC)
!qt!
!qt!  Routine chiamata dalla grafica alla pressione del tasto 'Execute'
!qt!
!qt   !USE sannel, only: goptim,shiftatomval,bumptab,sa_run
!qt   USE sannel, only: goptim,shiftatomval,sa_run
!qt   USE rotationmod, only: torsion_value_for_rotat
!qt   USE variables, only:cryst
!qt   USE bb_bc, only: bb_bc_condition
!qt   USE ilmviewmod, only: jav_set_modify
!qt   USE anti_bump, only: bump_settings, bump_from_table
!qt   USE rrestr, only: delete_restraints
!qt   USE progtype, only: opti_info_type
!qt   use simulated_anneal, only: SimAnn_Conditions_type
!qt   implicit none
!qt   type(opti_info_type)         :: info
!qt   type(SimAnn_Conditions_type) :: sc                     
!qt   type(bb_bc_condition)        :: bcond                     
!qt   integer                      :: i
!qt   real                         :: shiftatomv
!qt   type(bump_settings)          :: bumpsetC    
!qt!
!qt   call jav_set_modify(0)
!qt   goptim = info
!qt   shiftatomval = shiftatomv
!qt!corr   !cryst(1)%abset = bumpsetC  ! not working in optimized ifort
!qt!corr   cryst(1)%abset = bump_settings(bumpsetC%baction,bumpsetC%weight,bumpsetC%scaledist)
!qt!corr   if (cryst(1)%abset%baction > 0) then
!qt!corr       call bump_from_table(bumptab,cryst(1)%abres,cryst(1)%at,cryst(1)%bond,cryst(1)%cell,   &
!qt!corr                            cryst(1)%frag,cryst(1)%rot,cryst(1)%abset)
!qt!corr   else    ! 'None' was selected
!qt!corr       call delete_restraints(cryst(1)%abres)
!qt!corr   endif
!qt!corr      !write(0,*)'ABSET:',cryst(1)%abset%baction,cryst(1)%numresab()
!qt   call sa_run(goptim,sc,bcond)
!qt   info = goptim                           !sa_run modifica alcuni campi di gopti
!qt!
!qt!  Calcola nuovi valori per angoli di torsione per la grafica
!qt   do i=1,cryst(1)%numrot()   
!qt      cryst(1)%rot(i)%torsval = torsion_value_for_rotat(cryst(1)%rot(i),cryst(1)%at,cryst(1)%cell)
!qt   enddo
!qt   call jav_set_modify(3)
!qt!
!qt   end subroutine finishsannel
!qt
!qt
!----------------------------------------------------------------------------------------------------
!corr   subroutine test_calcg()
!corr   USE lsqalgo
!corr   implicit none
!corr   real, dimension(2) :: xgrad,xval
!corr   real               :: h,fx
!corr   integer         :: i
!corr     interface
!corr       real function funcprova(x)
!corr       real, dimension(:), intent(in) :: x
!corr       end function funcprova
!corr     end interface
!corr!
!corr   xval(1) = 8.0
!corr   xval(2) = 1.0
!corr   fx = funcprova(xval)
!corr   write(0,*)'eps=',epsilon(1.0)
!corr   do i=1,10
!corr      h=10.0**(-i)
!corr      call calc_gradient(xgrad,xval,fx,funcprova,h,icod=1)
!corr      write(0,*)'1 h=',h,' xgrad=',xgrad(:)
!corr      call calc_gradient(xgrad,xval,fx,funcprova,h,icod=2)
!corr      write(0,*)'2 h=',h,' xgrad=',xgrad(:)
!corr   enddo
!corr   call calc_gradient(xgrad,xval,fx,funcprova,icod=1)
!corr   write(0,*)'1 xgrad=',xgrad(:)
!corr   call calc_gradient(xgrad,xval,fx,funcprova,icod=2)
!corr   write(0,*)'2 xgrad=',xgrad(:)
!corr!
!corr   end subroutine test_calcg
!corr
!corr   real function funcprova(x)
!corr   real, dimension(:), intent(in) :: x
!corr!
!corr   funcprova = 2*x(1)*x(2)**2 + 3*x(1)*x(2)
!corr!
!corr   end function funcprova

!----------------------------------------------------------------------
!qt   subroutine SetValuesForPattern(cgpo, chklpo, ckopcorr)
!qt!
!qt   USE progtype
!qt   USE variables, only: cryst
!qt!corr   USE RefinecomRef, only: gpo
!qt!
!qt   implicit none
!qt   type(rparam_type) :: cgpo
!qt   integer           :: i, ckopcorr, chklpo(3)
!qt!
!qt!qt   cryst(1)%kopcorr = ckopcorr
!qt   cryst(1)%gpar = cgpo%val
!qt   cryst(1)%gcode = cgpo%rcod
!qt   do i=1, 3
!qt      cryst(1)%hklpo(i) = chklpo(i)
!qt   end do
!qt!
!qt   end subroutine SetValuesForPattern
!----------------------------------------------------------------------
!qt
!qt   subroutine GetValuesForPattern(cgpo, chklpo, ckopcorr)
!qt!
!qt   USE progtype
!qt   USE variables, only: cryst
!qt!corr   USE RefinecomRef, only: gpo
!qt!
!qt   implicit none
!qt   type(rparam_type) :: cgpo
!qt   integer           :: i, ckopcorr, chklpo(3)
!qt!
!qt!qt   ckopcorr = cryst(1)%kopcorr
!qt!corr   cgpo = gpo
!qt   cgpo%val  = cryst(1)%gpar
!qt   cgpo%rcod = cryst(1)%gcode
!qt   do i=1, 3
!qt      chklpo(i) = cryst(1)%hklpo(i)
!qt   end do
!qt!
!qt   end subroutine GetValuesForPattern
!----------------------------------------------------------------------

   subroutine Passa_Info_Bumptab(i,targ,code)
   USE sannel, only: bumptab
   implicit none
   integer, intent(in)              :: code
   integer, intent(in)              :: i
   real, intent(in)                 :: targ
!
   bumptab(i)%code  = code
   bumptab(i)%targ  = targ
!
   end subroutine Passa_Info_Bumptab


